/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <fcntl.h>
#include <sys/un.h>
#include "STAFConnectionProvider.h"
#include "STAFSocket.h"
#include "STAFUtil.h"
#include "STAFString.h"
#include "STAFEventSem.h"
#include "STAFTrace.h"
#include "STAFInternalUtil.h"
#include "STAFThreadManager.h"
#include <set>
#include <stdlib.h>

// XXX: Might want to look at adding someing like kSTAFConnectionError, instead
//      of using kSTAFBaseOSError in this code.  Same goes for Local IPC.

#define CATCH_STANDARD(functionString) \
catch (STAFException &e)\
{\
    if (errorBuffer)\
    {\
        *errorBuffer = getExceptionString(e, "STAFLocalIPCConnProvider.cpp: "\
                                          functionString)\
                       .adoptImpl();\
    }\
}\
catch (...)\
{\
    if (errorBuffer)\
    {\
        STAFString error("STAFLocalIPCConnProvider.cpp: Caught unknown "\
                         "exception in " functionString "()");\
        *errorBuffer = error.adoptImpl();\
    }\
}

#define CATCH_STANDARD_TRACE(functionString) \
catch (STAFException &e)\
{\
    STAFTrace::trace(kSTAFTraceError,\
                     getExceptionString(e, "STAFLocalIPCConnProvider.cpp: "\
                                        functionString));\
}\
catch (...)\
{\
    STAFTrace::trace(kSTAFTraceError,\
                     "STAFLocalIPCConnProvider.cpp: Caught unknown exception "\
                     "in " functionString "()");\
}

#ifdef STAF_OS_NAME_HPUX
extern "C"
{
    void _main();
}
#endif

// Prototypes

STAFRC_t STAFConnectionProviderConstruct(STAFConnectionProvider_t *provider,
                                         void *constructInfo,
                                         unsigned int constructInfoLevel,
                                         STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderStart(STAFConnectionProvider_t provider,
                                     void *startInfo,
                                     unsigned int startInfoLevel,
                                     STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderStop(STAFConnectionProvider_t provider,
                                    void *stopInfo,
                                    unsigned int stopInfoLevel,
                                    STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderDestruct(STAFConnectionProvider_t *provider,
                                        void *destructInfo,
                                        unsigned int destructInfoLevel,
                                        STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderConnect(STAFConnectionProvider_t provider,
                                       STAFConnection_t *connection,
                                       void *connectInfo,
                                       unsigned int connectInfoLevel,
                                       STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderGetMyNetworkIDs(STAFConnectionProvider_t provider,
                                               STAFStringConst_t *logicalID,
                                               STAFStringConst_t *physicalD,
                                               STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderGetOptions(
    STAFConnectionProvider_t provider,
    STAFObject_t *options,
    STAFString_t *errorBuffer);

STAFRC_t STAFConnectionProviderGetProperty(
    STAFConnectionProvider_t provider,
    STAFConnectionProviderProperty_t property, STAFStringConst_t *value,
    STAFString_t *errorBuffer);

STAFRC_t STAFConnectionRead(STAFConnection_t connection, void *buffer,
                            unsigned int readLength, STAFString_t *errorBuffer);

STAFRC_t STAFConnectionReadUInt(STAFConnection_t connection, unsigned int *uint,
                                STAFString_t *errorBuffer);

STAFRC_t STAFConnectionReadSTAFString(STAFConnection_t connection,
                                      STAFString_t *stafString,
                                      STAFString_t *errorBuffer);

STAFRC_t STAFConnectionWrite(STAFConnection_t connection, void *buffer,
                             unsigned int writeLength,
                             STAFString_t *errorBuffer);

STAFRC_t STAFConnectionWriteUInt(STAFConnection_t connection, unsigned int uint,
                                 STAFString_t *errorBuffer);

STAFRC_t STAFConnectionWriteSTAFString(STAFConnection_t connection,
                                       STAFStringConst_t stafString,
                                       STAFString_t *errorBuffer);

STAFRC_t STAFConnectionGetPeerNetworkIDs(STAFConnection_t connection,
                                         STAFStringConst_t *logicalID,
                                         STAFStringConst_t *physicalD,
                                         STAFString_t *errorBuffer);

STAFRC_t STAFConnectionDestruct(STAFConnection_t *connection,
                                STAFString_t *errorBuffer);


static STAFConnectionProviderFunctionTable gFuncTable =
{
    STAFConnectionProviderConstruct,
    STAFConnectionProviderStart,
    STAFConnectionProviderStop,
    STAFConnectionProviderDestruct,
    STAFConnectionProviderConnect,
    STAFConnectionProviderGetMyNetworkIDs,
    STAFConnectionProviderGetOptions,
    STAFConnectionProviderGetProperty,
    STAFConnectionRead,
    STAFConnectionReadUInt,
    STAFConnectionReadSTAFString,
    STAFConnectionWrite,
    STAFConnectionWriteUInt,
    STAFConnectionWriteSTAFString,
    STAFConnectionGetPeerNetworkIDs,
    STAFConnectionDestruct
};

struct STAFConnectionProviderImpl
{
    STAFConnectionProviderMode_t mode;
    void *data;

    STAFString_t ipcName;
    STAFSocket_t serverSocket;
    
    STAFStringConst_t logicalNetworkID;
    STAFStringConst_t physicalNetworkID;

    STAFObjectPtr options;

    STAFString_t portProperty;
    STAFString_t isSecureProperty;

    STAFConnectionProviderNewConnectionFunc_t connFunc;

    STAFEventSemPtr syncSem;
    STAFConnectionProviderState_t state;
    STAFThreadManagerPtr threadManager;
};

struct STAFConnectionImpl
{
    STAFSocket_t clientSocket;

    // XXX: HACK HACK
    //      Due to an unknown bug on Linux, these strings need to be here
    //      to prevent corruption of the corresponding tcp/ip connection
    //      provider.  It appears that some form of copy
    //      assignment/construction operator from this structure replaces
    //      one for the tcp/ip version and the buffer below zeros out the
    //      strings in the corresponding spot in the tcp/ip provider.

    STAFString dummyLogicalNetworkID;
    STAFString dummyPhysicalNetworkID;

    // ???: Look at moving this buffer into the read/write functions.  It will
    //      make the read/write functions thread safe.  Just need to make sure
    //      it doesn't hurt performance.

    char buffer[4096];
};

struct ConnectionData
{
    STAFConnectionProviderNewConnectionFunc_t connFunc;
    STAFConnectionProvider_t provider;
    STAFConnection_t connection;
};

typedef std::set<STAFConnectionProvider_t> ProviderSet;

static STAFMutexSem sActiveProvidersSem;
static ProviderSet sActiveProviders;

void atExit()
{
    ProviderSet activeProvidersCopy;

    {
        STAFMutexSemLock lock(sActiveProvidersSem);
        activeProvidersCopy = sActiveProviders;
    }

    for (ProviderSet::iterator iter = activeProvidersCopy.begin();
         iter != activeProvidersCopy.end();  ++iter)
    {
        STAFConnectionProvider_t thisProvider = *iter;

        // XXX: Clean up.  I.e., kill threads, etc.
    }
}


unsigned int STAFLocalIPCConnectionThread(void *data)
{
    try
    {
        ConnectionData *connData = reinterpret_cast<ConnectionData *>(data);

        connData->connFunc(connData->provider, connData->connection,
                           &gFuncTable, connData->provider->data);

        delete connData;
    }
    CATCH_STANDARD_TRACE("STAFLocalIPCConnectionThread");

    return 0;
}


unsigned int STAFTCPRunThread(void *providerImpl)
{
    STAFConnectionProviderImpl *provider =
        reinterpret_cast<STAFConnectionProviderImpl *>(providerImpl);

    try
    {
        provider->syncSem->post();

        while (provider->state == kSTAFConnectionProviderActive)
        {
            STAFConnectionImpl connImpl = { 0 };
            struct sockaddr_in clientAddress = { 0 };
            STAFSocketLen_t clientAddressLength = sizeof(clientAddress);

            do
            {
                connImpl.clientSocket = accept(
                    provider->serverSocket,
                    reinterpret_cast<struct sockaddr *>(&clientAddress),
                    &clientAddressLength);
            } while (!STAFUtilIsValidSocket(connImpl.clientSocket) &&
                     (STAFSocketGetLastError() == SOCEINTR) &&
                     (provider->state != kSTAFConnectionProviderStopped));

            if (provider->state == kSTAFConnectionProviderStopped)
                break;

            if (!STAFUtilIsValidSocket(connImpl.clientSocket))
            {
                STAFTrace::trace(kSTAFTraceError,
                                 STAFString("Error accepting on server socket"
                                            ", socket RC: ") +
                                 STAFString(STAFSocketGetLastError()));
                continue;
            }

            // Set the socket to be non-inheritable

            unsigned int osRC = 0;
            STAFSocket_t newSocket;

            if (STAFUtilGetNonInheritableSocket(connImpl.clientSocket,
                                                &newSocket, &osRC))
            {
                STAFTrace::trace(
                    kSTAFTraceError,
                    STAFString("Error getting non-inheritable socket, "
                               "STAFUtilGetNonInheritableSocket(), OS RC: ") +
                    STAFString(osRC));
                continue;
            }

            connImpl.clientSocket = newSocket;

            // Ok, let's perform the callback now

            ConnectionData connData;

            connData.connFunc   = provider->connFunc;
            connData.provider   = provider;
            connData.connection = new STAFConnectionImpl(connImpl);

            provider->threadManager->dispatch(STAFLocalIPCConnectionThread,
                                              new ConnectionData(connData));
        }
    }
    CATCH_STANDARD_TRACE("STAFTCPRunThread");

    try
    {
        provider->syncSem->post();
    }
    CATCH_STANDARD_TRACE("STAFTCPRunThread");

    return 0;
}


STAFRC_t STAFConnectionProviderConstruct(STAFConnectionProvider_t *provider,
                                         void *constructInfo,
                                         unsigned int constructInfoLevel,
                                         STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidParm;
    if (constructInfoLevel != 1) return kSTAFInvalidAPILevel;

    STAFConnectionProviderConstructInfoLevel1 *cpInfo =
        reinterpret_cast<STAFConnectionProviderConstructInfoLevel1 *>(
            constructInfo);

    try
    {
        static bool initedAtExit =  false;

        if (!initedAtExit)
        {
            STAFMutexSemLock lock(sActiveProvidersSem);

            if (!initedAtExit) atexit(atExit);
        }

        STAFConnectionProviderImpl lipcData;

        lipcData.mode          = cpInfo->mode;
        lipcData.syncSem       = STAFEventSemPtr(new STAFEventSem,
                        STAFEventSemPtr::INIT);
        lipcData.state         = kSTAFConnectionProviderStopped;
        lipcData.ipcName       = STAFString("STAF").adoptImpl();
        lipcData.serverSocket  = 0;
        lipcData.threadManager =
            STAFThreadManagerPtr(new STAFThreadManager,
                                 STAFThreadManagerPtr::INIT);

        // Get instance name

        if (getenv("STAF_INSTANCE_NAME") != NULL)
        {
            lipcData.ipcName = STAFString(
                getenv("STAF_INSTANCE_NAME")).adoptImpl();
        }
        
        for (unsigned int i = 0; i < cpInfo->numOptions; ++i)
        {
            STAFString thisOption =
                       STAFString(cpInfo->optionNames[i]).upperCase();

            if (thisOption == "IPCNAME")
            {
                lipcData.ipcName = 
                    (STAFString(lipcData.ipcName, STAFString::kShallow) +
                     cpInfo->optionValues[i]).adoptImpl();
            }
            else
            {
                if (errorBuffer)
                {
                    *errorBuffer =
                        STAFString(cpInfo->optionNames[i]).adoptImpl();
                }
                return kSTAFInvalidValue;
            }
        }

        // Add each option to a map.

        lipcData.options = STAFObject::createMap();
        lipcData.options->put("IPCName", STAFString(lipcData.ipcName));

        // Setup property values

        lipcData.portProperty = STAFString().adoptImpl();
        lipcData.isSecureProperty = STAFString("0").adoptImpl();

        // Assign logical and physical identifiers
        lipcData.logicalNetworkID = STAFString("local").adoptImpl();
        lipcData.physicalNetworkID = STAFString("local").adoptImpl();

        *provider = new STAFConnectionProviderImpl(lipcData);

        STAFRC_t rc = STAFSocketInit(errorBuffer);

        if (rc != kSTAFOk) return rc;

        *provider = new STAFConnectionProviderImpl(lipcData);

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderConstruct");

    return kSTAFUnknownError;
}


#define HANDLE_START_ERROR(string, function) \
{\
    STAFString theError = STAFString(string) + STAFString(", " function \
                                                          " RC=") + \
                          STAFString(STAFSocketGetLastError()); \
\
    if (errorBuffer) *errorBuffer = theError.adoptImpl();\
    return kSTAFBaseOSError;\
}

STAFRC_t STAFConnectionProviderStart(STAFConnectionProvider_t provider,
                                     void *startInfo,
                                     unsigned int startInfoLevel,
                                     STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (startInfoLevel != 1) return kSTAFInvalidAPILevel;

    STAFConnectionProviderStartInfoLevel1 *cpInfo =
        reinterpret_cast<STAFConnectionProviderStartInfoLevel1 *>(startInfo);

    try
    {
        if (cpInfo->newConnectionFunc == 0) return kSTAFInvalidValue;

        provider->connFunc = cpInfo->newConnectionFunc;
        provider->data = cpInfo->data;
        provider->serverSocket = socket(PF_UNIX, SOCK_STREAM, 0);

        if (!STAFUtilIsValidSocket(provider->serverSocket))
            HANDLE_START_ERROR("No socket available", "socket()");

        // Set the socket to be non-inheritable to ensure that the file
        // descriptor is closed when we fork() and exec().

        unsigned int osRC = 0;
        STAFSocket_t newSocket;

        if (STAFUtilGetNonInheritableSocket(provider->serverSocket,
                                            &newSocket, &osRC))
        {
            HANDLE_START_ERROR("Error getting non-inheritable server socket",
                               "STAFUtilGetNonInheritableSocket()");
        }

        provider->serverSocket = newSocket;

        struct sockaddr_un serverAddress = { 0 };

        serverAddress.sun_family = AF_UNIX;
        strcpy(serverAddress.sun_path, "/tmp/STAFIPC_");
        strcat(serverAddress.sun_path,
            STAFString(provider->ipcName).toCurrentCodePage()->buffer());
        // XXX: sun_len is not available on Linux, what to do?
        // serverAddress.sun_len = strlen(serverAddress.sun_path) + 1;

        // We attempt to remove the socket path in case it was left around
        // from a previous invocation
        unlink(serverAddress.sun_path);

        int bindRC = bind(provider->serverSocket,
                          reinterpret_cast<struct sockaddr *>(&serverAddress),
                          sizeof(serverAddress));

        if (bindRC != 0)
            HANDLE_START_ERROR("Error binding server socket", "bind()");

        int listenRC = listen(provider->serverSocket, SOMAXCONN);

        if (listenRC != 0)
            HANDLE_START_ERROR("Error listening on server socket", "listen()");

        // Ok, the provider is now ready

        provider->syncSem->reset();
        provider->state = kSTAFConnectionProviderActive;
        provider->threadManager->dispatch(STAFTCPRunThread, provider);

        // XXX: Might want to time out on this one.

        provider->syncSem->wait();

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderStart");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderStop(STAFConnectionProvider_t provider,
                                    void *stopInfo,
                                    unsigned int stopInfoLevel,
                                    STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (stopInfoLevel != 0) return kSTAFInvalidAPILevel;

    try
    {
        provider->state = kSTAFConnectionProviderStopped;

        provider->syncSem->reset();

        // Wake up the run thread and release resources acquired when
        // starting provider
	
        STAFSocketClose(provider->serverSocket);
	
        if (provider->syncSem->wait(10000) != 0)
        {
            STAFTrace::trace(
                kSTAFTraceError,
                STAFString("STAFLocalIPCConnectionProviderStop - Timed out "
                           "waiting for run thread to wake up"));
        }

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderStop");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderDestruct(STAFConnectionProvider_t *provider,
                                        void *destructInfo,
                                        unsigned int destructInfoLevel,
                                        STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (*provider == 0) return kSTAFInvalidObject;
    if (destructInfoLevel != 0) return kSTAFInvalidAPILevel;

    try
    {
        STAFMutexSemLock lock(sActiveProvidersSem);

        sActiveProviders.erase(*provider);

        if ((*provider)->state != kSTAFConnectionProviderStopped)
        {
            (*provider)->state = kSTAFConnectionProviderStopped;

            // XXX: Should we check the RC?
            STAFSocketClose((*provider)->serverSocket);
        }

        delete *provider;

        *provider = 0;

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderDestruct");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderConnect(STAFConnectionProvider_t provider,
                                       STAFConnection_t *connection,
                                       void *connectInfo,
                                       unsigned int connectInfoLevel,
                                       STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (connectInfoLevel != 1) return kSTAFInvalidAPILevel;
    if (connection == 0) return kSTAFInvalidParm;

    STAFConnectionProviderConnectInfoLevel1 *cpInfo =
        reinterpret_cast<STAFConnectionProviderConnectInfoLevel1 *>(connectInfo);

    try
    {
        STAFConnectionImpl connImpl;
        STAFString host = cpInfo->endpoint;
    struct sockaddr_un serverAddress = { 0 };

    serverAddress.sun_family = AF_UNIX;
    strcpy(serverAddress.sun_path, "/tmp/STAFIPC_");
    strcat(serverAddress.sun_path,
           STAFString(provider->ipcName).toCurrentCodePage()->buffer());
    // XXX: Linux doesn't support sun_len.  What to do?
    // serverAddress.sun_len = strlen(serverAddress.sun_path) + 1;

        connImpl.clientSocket = socket(PF_UNIX, SOCK_STREAM, 0);

        if (!STAFUtilIsValidSocket(connImpl.clientSocket))
        {
            STAFString error = STAFString("Error creating socket: socket() "
                                          "RC=") + STAFSocketGetLastError();
            if (errorBuffer) *errorBuffer = error.adoptImpl();
            return kSTAFCommunicationError;
        }

        // Set the socket to be non-inheritable

        unsigned int osRC = 0;
        STAFSocket_t newSocket;

        if (STAFUtilGetNonInheritableSocket(connImpl.clientSocket,
                                            &newSocket, &osRC))
        {
            STAFString error = STAFString(
                "Error getting non-inheritable socket, "
                "STAFUtilGetNonInheritableSocket(), OS RC: ") +
                STAFString(osRC);

            if (errorBuffer)
            {
                error += STAFString(*errorBuffer, STAFString::kShallow);
                *errorBuffer = error.adoptImpl();
            }

            STAFSocketClose(connImpl.clientSocket);

            return kSTAFCommunicationError;
        }

        connImpl.clientSocket = newSocket;
        
        int modeRC = STAFSocketSetBlockingMode(connImpl.clientSocket,
                                               kSTAFSocketNonBlocking,
                                               errorBuffer);
        if (modeRC != kSTAFOk)
        {
            STAFString error = STAFString("Error setting socket to non-blocking "
                                          "mode:");
            if (errorBuffer)
            {
                error += STAFString(*errorBuffer, STAFString::kShallow);
                *errorBuffer = error.adoptImpl();
            }

            STAFSocketClose(connImpl.clientSocket);

            return kSTAFCommunicationError;
        }

        int connectRC = connect(
            connImpl.clientSocket,
            reinterpret_cast<struct sockaddr *>(&serverAddress),
            sizeof(serverAddress));

        if ((connectRC < 0) &&
            (STAFSocketGetLastError() == ECONNREFUSED))
        {
            // On Solaris Opteron the connect request intermittently (usually
            // under heavy load) will return ECONNREFUSED.  In this case we
            // will retry the connect request up to 5 times (sleeping 100ms
            // between every request).

            int maxRetry = 5;

            for (int i = 0; i < maxRetry; i++)
            {
                if ((connectRC < 0) &&
                    (STAFSocketGetLastError() == ECONNREFUSED))
                {
                    STAFSocketClose(connImpl.clientSocket);

                    connImpl.clientSocket = socket(PF_UNIX, SOCK_STREAM, 0);

                    modeRC = STAFSocketSetBlockingMode(connImpl.clientSocket,
                                                       kSTAFSocketNonBlocking,
                                                       errorBuffer);

                    STAFThreadManager::sleepCurrentThread(100);

                    connectRC = connect(
                        connImpl.clientSocket,
                        reinterpret_cast<struct sockaddr *>(&serverAddress),
                        sizeof(serverAddress));
                }
                else
                {
                    break;
                }
            }
        }

        if ((connectRC < 0) &&
            (STAFSocketGetLastError() != SOCEINPROGRESS) &&
            (STAFSocketGetLastError() != SOCEWOULDBLOCK) &&
            (STAFSocketGetLastError() != 0))
        {
            STAFString error = STAFString("Error connecting to endpoint: "
                                          "connect() RC=") +
                               STAFSocketGetLastError();
            if (errorBuffer) *errorBuffer = error.adoptImpl();
            STAFSocketClose(connImpl.clientSocket);
            return kSTAFCommunicationError;
        }

        // XXX: this timeout needs to be passed somehow

        unsigned int timeout = 5000;

        if (connectRC < 0)
        {
            fd_set writeSocks;
            FD_ZERO(&writeSocks);
            FD_SET(connImpl.clientSocket, &writeSocks);
            timeval theTimeout = { timeout / 1000, (timeout % 1000) * 1000 };
            int selectRC = select(connImpl.clientSocket + 1, 0,
                                  &writeSocks, 0,
                                  (timeout == 0) ? 0 : &theTimeout);
            if (selectRC < 0)
            {
                STAFString error = STAFString("Error connecting to endpoint: "
                                              "select() RC=") +
                                   STAFSocketGetLastError();
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                STAFSocketClose(connImpl.clientSocket);
                return kSTAFCommunicationError;
            }
            else if (selectRC == 0)
            {
                STAFString error = STAFString("Timed out connecting to "
                                              "endpoint: select() timeout");
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                STAFSocketClose(connImpl.clientSocket);
                return kSTAFCommunicationError;
            }

            // On AIX, select() returns successfully if the target system exists.
            // This is true, even if the connection to the particular port is
            // refused.  So, we put in a zero-byte read to make sure that the
            // socket is actually connected successfully.
            // On Windows, we need to check if errno is SOCEWOULDBLOCK

            int recvRC = recv(connImpl.clientSocket, 0 , 0, STAF_MSG_NOSIGNAL);

            if ((recvRC < 0) && (STAFSocketGetLastError() != SOCEWOULDBLOCK))
            {
                STAFString error = STAFString("Error performing test read on "
                                              "connected endpoint: recv() RC=") +
                                   STAFSocketGetLastError();
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                STAFSocketClose(connImpl.clientSocket);
                return kSTAFCommunicationError;
            }
        }

        modeRC = STAFSocketSetBlockingMode(connImpl.clientSocket,
                                           kSTAFSocketBlocking, errorBuffer);
        if (modeRC != 0)
        {
            STAFString error = STAFString("Error setting socket to blocking "
                                          "mode:");
            if (errorBuffer)
            {
                error += STAFString(*errorBuffer, STAFString::kShallow);
                *errorBuffer = error.adoptImpl();
            }

            STAFSocketClose(connImpl.clientSocket);

            return kSTAFCommunicationError;
        }

        *connection = new STAFConnectionImpl(connImpl);

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderConnect");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderGetMyNetworkIDs(STAFConnectionProvider_t provider,
                                               STAFStringConst_t *logicalID,
                                               STAFStringConst_t *physicalID,
                                               STAFString_t *errorBuffer)
{
    if (provider   == 0) return kSTAFInvalidObject;
    if (logicalID  == 0) return kSTAFInvalidParm;
    if (physicalID == 0) return kSTAFInvalidParm;

    try
    {
        *logicalID = provider->logicalNetworkID;
        *physicalID = provider->physicalNetworkID;

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderGetMyNetworkIDs");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderGetOptions(
    STAFConnectionProvider_t provider,
    STAFObject_t *options,
    STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (options  == 0) return kSTAFInvalidParm;

    try
    {
        STAFObjectConstructReference(options, provider->options->getImpl());

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderGetOptions");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionProviderGetProperty(
    STAFConnectionProvider_t provider,
    STAFConnectionProviderProperty_t property, STAFStringConst_t *value,
    STAFString_t *errorBuffer)
{
    if (provider == 0) return kSTAFInvalidObject;
    if (value    == 0) return kSTAFInvalidParm;
    
    try
    {
        if (property == kSTAFConnectionProviderPortProperty)
            *value = provider->portProperty;
        else if (property == kSTAFConnectionProviderIsSecureProperty)
            *value = provider->isSecureProperty;
        else
            return kSTAFInvalidValue;

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionProviderGetProperty");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionRead(STAFConnection_t connection, void *buffer,
                            unsigned int readLength,
                            STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if ((buffer == 0) && (readLength != 0)) return kSTAFInvalidParm;

    try
    {
        int rc = 0;

        for(unsigned current = 0; current < readLength; current += rc)
        {
            // XXX: There should probably be a select() call here and in write

            int recvSize = (int)STAF_MIN((size_t)(readLength - current), 
                                         sizeof(connection->buffer));
            rc = recv(connection->clientSocket, connection->buffer, recvSize,
                      STAF_MSG_NOSIGNAL);

            if (rc < 0)
            {
                STAFString error = STAFString("Error reading from socket: "
                                              "recv() RC=") +
                                   STAFSocketGetLastError();
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                return kSTAFCommunicationError;
            }
            else if (rc == 0)
            {
                STAFString error = STAFString("Error reading from socket: "
                                              "other side closed socket");
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                return kSTAFCommunicationError;
            }
            else
            {
                memcpy((char *)buffer + current, connection->buffer, rc);
            }
        }

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionRead");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionReadUInt(STAFConnection_t connection,
                                unsigned int *uint, STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if (uint == 0) return kSTAFInvalidParm;

    STAFRC_t rc = STAFConnectionRead(connection, uint, sizeof(unsigned int),
                                     errorBuffer);

    if (rc == kSTAFOk) *uint = STAFUtilConvertLEUIntToNative(*uint);

    return rc;
}


STAFRC_t STAFConnectionReadSTAFString(STAFConnection_t connection,
                                      STAFString_t *stafString,
                                      STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if (stafString == 0) return kSTAFInvalidParm;

    try
    {
        // First, get the size of the string

        unsigned int size = 0;
        STAFRC_t rc = STAFConnectionReadUInt(connection, &size, errorBuffer);

        if (rc != kSTAFOk) return rc;

        // Next read in the actual UTF-8 data

        char *inputData = new char[size];

        rc = STAFConnectionRead(connection, (void *)inputData, size,
                                errorBuffer);

        if (rc != kSTAFOk)
        {
            delete [] inputData;
            return rc;
        }

        // Now, create the actual STAFString

        unsigned int osRC = 0;

        rc = STAFStringConstruct(stafString, inputData, size, &osRC);

        if ((rc == kSTAFBaseOSError) && (errorBuffer != 0))
            *errorBuffer = STAFString(osRC).adoptImpl();

        delete [] inputData;

        return rc;
    }
    CATCH_STANDARD("STAFConnectionReadSTAFString");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionWrite(STAFConnection_t connection, void *buffer,
                             unsigned int writeLength,
                             STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if ((buffer == 0) && (writeLength != 0)) return kSTAFInvalidParm;

    try
    {
        int rc = 0;

        for(unsigned int current = 0; current < writeLength; current += rc)
        {
            int sendSize = (int)STAF_MIN((size_t)(writeLength - current), 
                                         sizeof(connection->buffer));
            memcpy(connection->buffer, (char *)buffer + current, sendSize);

            rc = send(connection->clientSocket, (char *)connection->buffer,
                      (int)sendSize, 0);

            if (rc < 0)
            {
                STAFString error = STAFString("Error writing to socket: "
                                              "send() RC=") +
                                   STAFSocketGetLastError();
                if (errorBuffer) *errorBuffer = error.adoptImpl();
                return kSTAFCommunicationError;
            }
        }

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionWrite");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionWriteUInt(STAFConnection_t connection,
                                 unsigned int uint, STAFString_t *errorBuffer)
{
    unsigned int leUInt = STAFUtilConvertNativeUIntToLE(uint);

    return STAFConnectionWrite(connection, &leUInt, sizeof(unsigned int),
                               errorBuffer);
}


STAFRC_t STAFConnectionWriteSTAFString(STAFConnection_t connection,
                                       STAFStringConst_t stafString,
                                       STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if (stafString == 0) return kSTAFInvalidObject;

    try
    {
        unsigned int osRC = 0;
        unsigned int length = 0;
        const char *buffer = 0;
        STAFRC_t rc = STAFStringGetBuffer(stafString, &buffer, &length, &osRC);

        if ((rc == kSTAFBaseOSError) && (errorBuffer != 0))
        {
            *errorBuffer = STAFString(osRC).adoptImpl();
            return rc;
        }

        rc = STAFConnectionWriteUInt(connection, length, errorBuffer);

        if (rc == kSTAFOk)
        {
            rc = STAFConnectionWrite(connection, const_cast<char *>(buffer),
                                     length, errorBuffer);
        }

        return rc;
    }
    CATCH_STANDARD("STAFConnectionWriteSTAFString");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionGetPeerNetworkIDs(STAFConnection_t connection,
                                         STAFStringConst_t *logicalID,
                                         STAFStringConst_t *physicalID,
                                         STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidObject;
    if (logicalID  == 0) return kSTAFInvalidParm;
    if (physicalID == 0) return kSTAFInvalidParm;

    try
    {
        static STAFString sLocalString("local");

        *logicalID = sLocalString.getImpl();
        *physicalID = sLocalString.getImpl();

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionGetPeerNetworkIDs");

    return kSTAFUnknownError;
}


STAFRC_t STAFConnectionDestruct(STAFConnection_t *connection,
                                STAFString_t *errorBuffer)
{
    if (connection == 0) return kSTAFInvalidParm;
    if (*connection == 0) return kSTAFInvalidObject;

    try
    {
        STAFSocketClose((*connection)->clientSocket);

        delete *connection;

        *connection = 0;

        return kSTAFOk;
    }
    CATCH_STANDARD("STAFConnectionDestruct");

    return kSTAFUnknownError;
}
