/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2004                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <vector>
#include <map>
#include "STAFServiceInterface.h"
#include "STAFPerlService.h"
#include "STAFConnectionProvider.h"
#include "STAFUtil.h"
#include "STAFThread.h"
#include "STAFEventSem.h"
#include "STAFMutexSem.h"
#include "STAFProcess.h"
#include "STAFFileSystem.h"
#include "STAFTrace.h"
#include "STAF_fstream.h"

struct PerlInterpreterData
{
    STAFString fName;
    STAFString fExec;
    STAFString fOptions;
    STAFConnectionProviderPtr fConnProv;
    STAFEventSemPtr fPerlInterpreterExitedSem;
    STAFProcessID_t fPerlInterpreter_PID;
    unsigned int fNumServices;
};

typedef STAFRefPtr<PerlInterpreterData> PerlInterpreterDataPtr;
typedef std::map<STAFString, PerlInterpreterDataPtr> PerlInterpreterDataMap;

struct PerlInterpreterStartThreadData
{
    PerlInterpreterStartThreadData (STAFString_t startString, STAFEventSemPtr
        &exitedSem) :
        fStartString(startString), fPerlInterpreterExitedSem(exitedSem)
    { /* Do nothing */ }

    STAFString_t fStartString;
    STAFEventSemPtr fPerlInterpreterExitedSem;
};

struct STAFProcPerlServiceData
{
    STAFString fName;
    STAFString fExec;
    PerlInterpreterDataPtr fPerlInterpreter;
};

static PerlInterpreterDataMap sPerlInterpreterDataMap;
static STAFMutexSem sPerlInterpreterDataSem;
static STAFString sLocal("local");
static STAFString sIPCName("IPCNAME");

STAFRC_t STAFServiceGetLevelBounds(unsigned int levelID,
                                   unsigned int *minimum,
                                   unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceInit:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceAcceptRequest:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return kSTAFInvalidAPILevel;
        }
    }

    return kSTAFOk;
}


unsigned int STAFDoShutdownPerlInterpreter(STAFConnectionProviderPtr &connProv)
{
    try
    {
        STAFConnectionPtr connPtr = connProv->connect(sLocal);

        connPtr->writeUInt(PERL_SERVICE_PERL_INTERPRETER_EXIT);

        STAFRC_t perlInterpreterRC = connPtr->readUInt();
        STAFString perlInterpreterResultString = connPtr->readString();

        // Note, this last connect is required depending on the thread timing
        // in STAFPerlServiceHelper.cpp.   In that file, if the request
        // thread gets started before the creating thread gets control back,
        // then this call is unnecessary.  Otherwise, we need to send this
        // final request so that the main thread will be able to check the
        // exit flag.

        connPtr = connProv->connect(sLocal);

        connPtr->writeUInt(PERL_SERVICE_PERL_INTERPRETER_FIN);
    }
    catch (STAFConnectionProviderException)
    {
        // These execptions are anticipated in cases where the PerlInterpreter
        // has already shutdown before we send the shutdown commands
    }

    return 0;
}


// Note: You should already own sPerlInterpreterDataSem before
// calling STAFShutdownJMV

unsigned int STAFShutdownPerlInterpreter(STAFString &perlInterpreterName)
{
    try
    {
        PerlInterpreterDataPtr perlInterpreter =
            sPerlInterpreterDataMap[perlInterpreterName];

        sPerlInterpreterDataMap.erase(perlInterpreterName);

        STAFDoShutdownPerlInterpreter(perlInterpreter->fConnProv);
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFShutdownPerlInterpreter()");
    }
    catch (...)
    {   
        STAFTrace::trace(
            kSTAFTraceError, "Caught unknown exception in "
            "PLSTAF.STAFShutdownPerlInterpreter()");
    }

    return 0;
}


STAFRC_t STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                              void *pServiceInfo, unsigned int infoLevel,
                              STAFString_t *pErrorBuffer)
{
    try
    {
        if (infoLevel != 30) return kSTAFInvalidAPILevel;

        STAFProcPerlServiceData data;
        STAFServiceInfoLevel30 *pInfo =
            static_cast<STAFServiceInfoLevel30 *>(pServiceInfo);
        STAFString perlServiceName = pInfo->name;
        STAFString perlServiceExec = pInfo->exec;
        STAFString perlInterpreterExec = "perl ";
        STAFString perlInterpreterOptions = "";
        bool perlInterpreterSpecified = false;
        unsigned int maxLogs = 5; // Defaults to keeping 5 PerlInterpreter logs
        long maxLogSize = 1048576; // Default size of each log is 1M

        data.fName = pInfo->name;
        data.fExec = pInfo->exec;

        //XXX: Can't think of a way to make sure the STAFRoot/lib dir is in the
        // include path, so I'm grabbing it from STAF
        STAFHandlePtr handlePtr;

        STAFRC_t retcode = STAFHandle::create("STAF/Service/STAFPerlService",
            handlePtr);

        if (retcode != kSTAFOk)
        {
            return retcode;
        }

        STAFResultPtr stafRootResultPtr =
               handlePtr->submit("local", "var",
               "resolve string {STAF/Config/STAFRoot}");

        if(stafRootResultPtr->rc != 0)
        {
            return stafRootResultPtr->rc;
        }

        // XXX: Need to unregister this handle here

        STAFString libDir = stafRootResultPtr->result;
        libDir = libDir.replace("\\", "/"); // switch all backslashes to forward
        perlInterpreterOptions +=  "-I" + libDir + "/lib ";


        // Walk through and verify the config options

        for (unsigned int i = 0; i < pInfo->numOptions; ++i)
        {
            STAFString upperOptionName =
                       STAFString(pInfo->pOptionName[i]).upperCase();
            STAFString optionValue(pInfo->pOptionValue[i]);

            if (upperOptionName == "PERLEXEC")
            {
                perlInterpreterExec = optionValue;
                perlInterpreterSpecified = true;
            }
            else if (upperOptionName == "MAXLOGS")
            {
                // Check to make sure it is an integer value > 0

                STAFString maxLogsStr = optionValue;

                if (maxLogsStr.isDigits() && (maxLogsStr.asUInt() > 0))
                {
                    maxLogs = maxLogsStr.asUInt();
                }
                else
                {
                    STAFString errmsg(
                        "Error constructing the Perl Interpreter using "
                        "Perl Service Name: " + perlServiceName +
                        ", PerlServiceExec " + perlServiceExec +
                        ", Perl Interpreter: " + perlInterpreterExec +
                        ", Perl Interpreter Options: " +
                        perlInterpreterOptions +
                        ", RC: " + STAFString(kSTAFInvalidValue) +
                        ", MAXLOGS must be an positive integer");
                    *pErrorBuffer = errmsg.adoptImpl();

                    return kSTAFServiceConfigurationError;
                }
            }
            else if (upperOptionName == "MAXLOGSIZE")
            {
                // Check to make sure it is an integer value > 0

                STAFString maxLogSizeStr = optionValue;

                if (maxLogSizeStr.isDigits() && (maxLogSizeStr.asUInt() > 0))
                {
                    maxLogSize = (long)maxLogSizeStr.asUInt();
                }
                else
                {
                    STAFString errmsg(
                        "Error constructing the Perl Interpreter using "
                        "Perl Service Name: " + perlServiceName +
                        ", PerlServiceExec " + perlServiceExec +
                        ", Perl Interpreter: " + perlInterpreterExec +
                        ", Perl Interpreter Options: " +
                        perlInterpreterOptions +
                        ", RC: " + STAFString(kSTAFInvalidValue) +
                        ", MAXLOGSIZE must be an positive integer");
                    *pErrorBuffer = errmsg.adoptImpl();

                    return kSTAFServiceConfigurationError;
                }
            }
            else if (upperOptionName == "PERLPARMS")
            {
                perlInterpreterOptions = perlInterpreterOptions + " " +
                    optionValue;
            }
            else
            {
                STAFString optionError(pInfo->pOptionName[i]);
                *pErrorBuffer = optionError.adoptImpl();

                return kSTAFServiceConfigurationError;
            }
        }

        // Check to see if the specified PerlInterpreter already exists

        STAFMutexSemLock lock(sPerlInterpreterDataSem);

        if (sPerlInterpreterDataMap.find(perlServiceName) ==
            sPerlInterpreterDataMap.end())
        {
            // The PerlInterpreter doesn't exist so we need to start it

            PerlInterpreterData perlInterpreterData;

            perlInterpreterData.fName = perlServiceName;
            perlInterpreterData.fExec = perlServiceExec;
            perlInterpreterData.fOptions = perlInterpreterOptions;
            perlInterpreterData.fPerlInterpreterExitedSem =
                STAFEventSemPtr(new STAFEventSem, STAFEventSemPtr::INIT);
            perlInterpreterData.fNumServices = 0;

            STAFString perlInterpreterStartString = perlInterpreterOptions +
                " " + perlServiceExec + " " + perlServiceName;

            // Create the connection provider for the PerlInterpreter

            STAFString ipcName = perlServiceName; //CHANGED BY TMG
            STAFStringConst_t optionData[] = { sIPCName.getImpl(),
                                               ipcName.getImpl() };
            STAFConnectionProviderConstructInfoLevel1 constructInfo =
            {
                kSTAFConnectionProviderOutbound,
                1,
                optionData,
                &optionData[1]
            };

            perlInterpreterData.fConnProv =
                STAFConnectionProvider::createRefPtr(ipcName, "STAFLIPC",
                                                     &constructInfo, 1);

            // We need to shutdown any PerlInterpreter that might happen to be
            // leftover from a "bad" exit from a previous STAFProc

            STAFDoShutdownPerlInterpreter(perlInterpreterData.fConnProv);

            // We need to capture stdout/stderr for diagnostic purposes
            // Create directory for PerlInterpreter log file if doesn't
            // already exist

            STAFFSPath logPath;
            logPath.setRoot(pInfo->writeLocation);
            logPath.addDir("lang");
            logPath.addDir("perl");
            logPath.addDir("perlInterpreter");
            logPath.addDir(perlServiceName);

            if (!logPath.exists())
            {
                try
                {
                    // Don't want exceptions here
                    STAFFSEntryPtr dir =
                        logPath.createDirectory(kSTAFFSCreatePath);
                }
                catch (...)
                { /* Do Nothing */ }

                if (!logPath.exists())
                {
                    STAFString errmsg(
                        "Error constructing the Perl Interpreter using "
                        "Perl Service Name: " + perlServiceName +
                        ", PerlServiceExec " + perlServiceExec +
                        ", Perl Interpreter: " + perlInterpreterExec +
                        ", Perl Interpreter Options: " +
                        perlInterpreterOptions +
                        ", Error creating PerlInterpreterLog directory: " +
                        logPath.asString());
                    *pErrorBuffer = errmsg.adoptImpl();

                    return kSTAFServiceConfigurationError;
                }
            }

            // Name the current PerlInterpreter log file - PerlInterpreterLog.1

            logPath.setName("PerlInterpreterLog");
            logPath.setExtension("1");
            STAFString logName = logPath.asString();

            // Instead of replacing the PerlInterpreter log file each time a
            // PerlInterpreter is created, use the following PerlInterpreter
            // OPTIONs to determine when to create a new PerlInterpreter log
            // file and how many PerlInterpreter log files to save:
            // - MaxLogs   : Maximum number of PerlInterpreter log files to keep
            // - MaxLogSize: Maximum size of a PerlInterpreter log file in bytes

            // Open the PerlInterpreter log file

            fstream outfile(logName.toCurrentCodePage()->buffer(),
                            ios::out | ios::app);

            if (!outfile)
            {
                STAFString errmsg(
                    "Error constructing the Perl Interpreter using "
                    "Perl Service Name: " + perlServiceName +
                    ", PerlServiceExec " + perlServiceExec +
                    ", Perl Interpreter: " + perlInterpreterExec +
                    ", Perl Interpreter Options: " + perlInterpreterOptions +
                    ", RC: " + STAFString(kSTAFFileOpenError) +
                    ", Error opening file " + logName);
                *pErrorBuffer = errmsg.adoptImpl();

                return kSTAFServiceConfigurationError;
            }

            // Figure out how big the PerlInterpreter log file is

            outfile.seekp(0, ios::end);
            unsigned int fileLength = (unsigned int)outfile.tellp();

            if (fileLength > maxLogSize)
            {
                // Roll any existing log files (e.g. Rename
                // PerlInterpreterLog.2.out to PerlInterpreterLog.3,
                // PerlInterpreterLog.1 to PerlInterpreterLog.2, etc) and
                // create a new PerlInterpreterLog.1 file.  If the # of
                // existing logs > MAXLOGS, don't save the oldest log.

                outfile.close();

                STAFFSPath fromLogPath(logPath);

                for (int i = maxLogs; i > 0; --i)
                {
                    fromLogPath.setExtension(STAFString(i));

                    if (fromLogPath.exists() && i < maxLogs)
                    {
                        // Rename PerlInterpreterLog.<i> to
                        // PerlInterpreterLog.<i+1>

                        STAFFSPath toLogPath(fromLogPath);
                        toLogPath.setExtension(STAFString(i + 1));

                        fromLogPath.getEntry()->move(toLogPath.asString());
                    }
                }

                // Open a new empty current log file

                outfile.open(logName.toCurrentCodePage()->buffer(),
                             ios::out | ios::trunc);

                if (!outfile)
                {
                    STAFString errmsg(
                        "Error constructing the Perl Interpreter using "
                        "Perl Service Name: " + perlServiceName +
                        ", PerlServiceExec " + perlServiceExec +
                        ", Perl Interpreter: " + perlInterpreterExec +
                        ", Perl Interpreter Options: " +
                        perlInterpreterOptions +
                        ", RC: " + STAFString(kSTAFFileOpenError) +
                        ", Error opening file " + logName);
                    *pErrorBuffer = errmsg.adoptImpl();

                    return kSTAFServiceConfigurationError;
                }
            }

            // Write the PerlInterpreter start information to the
            // PerlInterpreter log file

            STAFString separatorLine("***************************************"
                                     "***************************************");
            STAFString line1("*** " + STAFTimestamp().asString() +
                                  " - Start of Log for PerlServiceName: " +
                                  perlServiceName);
            STAFString line2("*** PerlInterpreter Executable: " +
                                  perlInterpreterExec);
            STAFString line3("*** PerlService Executable: " + perlServiceExec);
            STAFString line4("*** PerlInterpreter Options   :");
            if (perlInterpreterOptions != STAFString(" "))
                line4 += perlInterpreterOptions;
            else
                line4 += " none";

            outfile << separatorLine.toCurrentCodePage()->buffer() << endl
                    << line1.toCurrentCodePage()->buffer() << endl
                    << line2.toCurrentCodePage()->buffer() << endl
                    << line3.toCurrentCodePage()->buffer() << endl
                    << line4.toCurrentCodePage()->buffer() << endl
                    << separatorLine.toCurrentCodePage()->buffer() << endl;

            outfile.close();

            // Start a process for the PerlInterpreter

            STAFProcessStartInfoLevel1 startInfo = { 0 };


            startInfo.command     = perlInterpreterExec.getImpl();

            startInfo.parms       = perlInterpreterStartString.getImpl();
            startInfo.consoleMode = kSTAFProcessSameConsole;

            startInfo.stdoutMode  = kSTAFProcessIOStdout;
            startInfo.stderrMode  = kSTAFProcessIOStdout;

            startInfo.stdoutRedirect = logName.getImpl();


            unsigned int osRC = 0;
            STAFString_t errorBuffer = 0;
            STAFRC_t rc = STAFProcessStart2(
                0, 0, &startInfo, 1, &osRC, &errorBuffer);

            if (rc != kSTAFOk)
            {
                STAFString startError(
                    "Error starting a process for the the Perl Interpreter "
                    "using Perl Service Name: " + perlServiceName +
                    ", PerlServiceExec " + perlServiceExec +
                    ", Perl Interpreter: " + perlInterpreterExec +
                    ", Perl Interpreter Options: " + perlInterpreterOptions +
                    ", RC: " + STAFString(rc) + ", Result: " +
                    STAFString(errorBuffer, STAFString::kShallow));

                // Add more details to the error msg when the Perl executable
                // cannot be found.

                if (rc == 10)
                {
                    if (!perlInterpreterSpecified)
                    {
                        startError = startError + "  Make sure the perl " +
                            "executable is in the PATH.";
                    }
                    else
                    {
                        startError = startError + "  Make sure " +
                            perlInterpreterExec + " exists.";
                    }
                }

                *pErrorBuffer = startError.adoptImpl();

                return kSTAFServiceConfigurationError;
            }

            // Now we need to wait for it to start

            bool perlInterpreterReady = false;

            for (int i = 0; (i < 30) & !perlInterpreterReady; ++i)
            {
                // XXX: Need to check to see if the PerlInterpreter is actually
                // still running

                try
                {
                    // First connect to the PerlInterpreter
                    STAFConnectionPtr connPtr =
                        perlInterpreterData.fConnProv->connect(sLocal);

                    // Now see if it's alive

                    connPtr->writeUInt(PERL_SERVICE_PERL_INTERPRETER_PING);
                    STAFRC_t perlInterpreterRC = connPtr->readUInt();
                    STAFString perlInterpreterResultString =
                        connPtr->readString();

                    if (perlInterpreterRC != kSTAFOk)
                    {
                        perlInterpreterResultString =
                            "Error starting PerlInterpreter: " +
                            perlInterpreterResultString;
                        *pErrorBuffer = perlInterpreterResultString.adoptImpl();

                        return kSTAFServiceConfigurationError;
                    }

                    perlInterpreterReady = true;
                }
                catch (STAFException e)
                {
                    //e.trace("PLSTAF.STAFServiceConstruct - error connecting");
                }

                if (!perlInterpreterReady)
                    STAFThreadSleepCurrentThread(1000, 0);
            }

            // If we didn't time out waiting for the PerlInterpreter then we
            // add the PerlInterpreter to the map of PerlInterpreters

            if (!perlInterpreterReady)
            {
                *pErrorBuffer = STAFString(
                    "Unable to connect to Perl Interpreter").adoptImpl();
                return kSTAFServiceConfigurationError;
            }

            sPerlInterpreterDataMap[data.fName] = PerlInterpreterDataPtr(new
                PerlInterpreterData(perlInterpreterData),
                PerlInterpreterDataPtr::INIT);
        }
        else
        {
            *pErrorBuffer = STAFString("There is already a service named " +
                perlServiceName + " registered with STAF").adoptImpl();
            return kSTAFServiceConfigurationError;
        }

        // Ok. We have a PerlInterpreter.  Now let's try to load the service.
        data.fPerlInterpreter = sPerlInterpreterDataMap[data.fName];

        STAFConnectionPtr connPtr =
            data.fPerlInterpreter->fConnProv->connect(sLocal);

        connPtr->writeUInt(PERL_SERVICE_LOAD);
        connPtr->writeString(data.fName);
        connPtr->writeString(data.fExec);
        connPtr->writeString(pInfo->writeLocation);
        connPtr->writeUInt(pInfo->serviceType);

        unsigned int constructRC = connPtr->readUInt();
        STAFString constructResult = connPtr->readString();

        if (constructRC != kSTAFOk)
        {
            if (data.fPerlInterpreter->fNumServices == 0)
                STAFShutdownPerlInterpreter(perlServiceName);

            *pErrorBuffer = constructResult.adoptImpl();
            return constructRC;
        }

        // The service is now loaded

        STAFProcPerlServiceData *pService = new STAFProcPerlServiceData(data);

        *pServiceHandle = pService;
        ++pService->fPerlInterpreter->fNumServices;

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFServiceConstruct()"); // Should do something here
    }
    catch (...)
    {
        STAFTrace::trace(
            kSTAFTraceError,
            "Caught unknown exception in PLSTAF.STAFServiceConstruct()");
    }

    return kSTAFUnknownError;
}

STAFRC_t STAFServiceDestruct(STAFServiceHandle_t *serviceHandle,
                             void *pDestructInfo,
                             unsigned int destructLevel,
                             STAFString_t *pErrorBuffer)
{
    try
    {
        if (destructLevel != 0) return kSTAFInvalidAPILevel;

        STAFProcPerlServiceData *pData =
            static_cast<STAFProcPerlServiceData *>(*serviceHandle);

        STAFMutexSemLock lock(sPerlInterpreterDataSem);

        if (--pData->fPerlInterpreter->fNumServices == 0)
            STAFShutdownPerlInterpreter(pData->fPerlInterpreter->fName);

        delete pData;
        *serviceHandle = 0;

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFServiceDestruct()");
    }
    catch (...)
    {
        STAFTrace::trace(
            kSTAFTraceError,
            "Caught unknown exception in PLSTAF.STAFServiceDestruct()");
    }

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceInit(STAFServiceHandle_t serviceHandle,
                         void *pInitInfo, unsigned int initLevel,
                         STAFString_t *pErrorBuffer)
{
    try
    {
        if (initLevel != 30) return kSTAFInvalidAPILevel;

        STAFProcPerlServiceData *pData =
            static_cast<STAFProcPerlServiceData *>(serviceHandle);
        STAFServiceInitLevel30 *pInfo =
            static_cast<STAFServiceInitLevel30 *>(pInitInfo);
        STAFConnectionPtr connPtr =
            pData->fPerlInterpreter->fConnProv->connect(sLocal);

        connPtr->writeUInt(PERL_SERVICE_INIT);
        connPtr->writeString(pData->fName);
        connPtr->writeString(pInfo->parms);
        connPtr->writeString(pInfo->writeLocation);

        unsigned int initRC = connPtr->readUInt();
        STAFString initResult = connPtr->readString();

        if (initRC != kSTAFOk)
        {
            *pErrorBuffer = initResult.adoptImpl();
            return initRC;
        }

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFServiceInit()");
    }
    catch (...)
    {
        STAFTrace::trace(
            kSTAFTraceError,
            "Caught unknown exception in PLSTAF.STAFServiceInit()");
    }

    return kSTAFUnknownError;
}

STAFRC_t STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                         void *pTermInfo, unsigned int termLevel,
                         STAFString_t *pErrorBuffer)
{
    try
    {
        if (termLevel != 0) return kSTAFInvalidAPILevel;

        STAFProcPerlServiceData *pData =
            static_cast<STAFProcPerlServiceData *>(serviceHandle);
        STAFConnectionPtr connPtr =
            pData->fPerlInterpreter->fConnProv->connect(sLocal);

        connPtr->writeUInt(PERL_SERVICE_TERM);
        connPtr->writeString(pData->fName);

        unsigned int termRC = connPtr->readUInt();
        STAFString termResult = connPtr->readString();

        if (termRC != kSTAFOk)
        {
            *pErrorBuffer = termResult.adoptImpl();
            return termRC;
        }

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFServiceTerm()");
    }
    catch (...)
    {
        STAFTrace::trace(
            kSTAFTraceError,
            "Caught unknown exception in PLSTAF.STAFServiceTerm()");
    }

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                  void *pRequestInfo, unsigned int reqLevel,
                                  STAFString_t *pResultBuffer)
{
    try
    {
        if (reqLevel != 30) return kSTAFInvalidAPILevel;

        STAFProcPerlServiceData *pData =
            static_cast<STAFProcPerlServiceData *>(serviceHandle);
        STAFServiceRequestLevel30 *pInfo =
            static_cast<STAFServiceRequestLevel30 *>(pRequestInfo);
        STAFConnectionPtr connPtr =
            pData->fPerlInterpreter->fConnProv->connect(sLocal);

        unsigned int  machineLength          = 0;
        const char   *machineBuffer          = 0;
        unsigned int  machineNicknameLength  = 0;
        const char   *machineNicknameBuffer  = 0;
        unsigned int  handleNameLength       = 0;
        const char   *handleNameBuffer       = 0;
        unsigned int  requestLength          = 0;
        const char   *requestBuffer          = 0;
        unsigned int  userLength             = 0;
        const char   *userBuffer             = 0;
        unsigned int  endpointLength         = 0;
        const char   *endpointBuffer         = 0;
        unsigned int  stafInstanceUUIDLength = 0;
        const char   *stafInstanceUUIDBuffer = 0;
        unsigned int  physicalInterfaceIDLength = 0;
        const char   *physicalInterfaceIDBuffer = 0;

        STAFStringGetBuffer(pInfo->machine, &machineBuffer,
                            &machineLength, 0);
        STAFStringGetBuffer(pInfo->machineNickname, &machineNicknameBuffer,
                            &machineNicknameLength, 0);
        STAFStringGetBuffer(pInfo->handleName, &handleNameBuffer,
                            &handleNameLength, 0);
        STAFStringGetBuffer(pInfo->request, &requestBuffer,
                            &requestLength, 0);
        STAFStringGetBuffer(pInfo->user, &userBuffer, &userLength, 0);
        STAFStringGetBuffer(pInfo->endpoint, &endpointBuffer,
                            &endpointLength, 0);
        STAFStringGetBuffer(pInfo->stafInstanceUUID, &stafInstanceUUIDBuffer,
                            &stafInstanceUUIDLength, 0);
        STAFStringGetBuffer(pInfo->physicalInterfaceID,
                            &physicalInterfaceIDBuffer,
                            &physicalInterfaceIDLength, 0);

        // IMPORTANT:  Increase the numFields value if add a field to the
        // ServiceRequest class for a new STAFServiceInterfaceLevel.

        unsigned int numFields = 16;

        unsigned int bufferLength = (numFields * sizeof(unsigned int)) +
            pData->fName.length() + machineLength + machineNicknameLength +
            handleNameLength + requestLength + userLength + endpointLength +
            stafInstanceUUIDLength + physicalInterfaceIDLength;

        STAFBuffer<char> buffer(new char[bufferLength], STAFBuffer<char>::INIT,
                                STAFBuffer<char>::ARRAY);
        unsigned int *uintBuffer =
                     reinterpret_cast<unsigned int *>((char *)buffer);

        uintBuffer[0] =
            STAFUtilConvertNativeUIntToLE(PERL_SERVICE_ACCEPT_REQUEST);
        uintBuffer[1] = STAFUtilConvertNativeUIntToLE(
            bufferLength - (2 * sizeof(unsigned int)));
        uintBuffer[2] = pData->fName.length();
        uintBuffer[3] = pInfo->handle;
        uintBuffer[4] = pInfo->trustLevel;
        uintBuffer[5] = machineLength;
        uintBuffer[6] = machineNicknameLength;
        uintBuffer[7] = handleNameLength;
        uintBuffer[8] = requestLength;
        uintBuffer[9] = pInfo->diagEnabled;
        uintBuffer[10] = pInfo->requestNumber;
        uintBuffer[11] = userLength;
        uintBuffer[12] = endpointLength;
        uintBuffer[13] = stafInstanceUUIDLength;
        uintBuffer[14] = pInfo->isLocalRequest;
        uintBuffer[15] = physicalInterfaceIDLength;

        char *currBuffer = buffer + (numFields * sizeof(unsigned int));

        memcpy(currBuffer, pData->fName.buffer(), pData->fName.length());
        currBuffer += pData->fName.length();
        memcpy(currBuffer, machineBuffer, machineLength);
        currBuffer += machineLength;
        memcpy(currBuffer, machineNicknameBuffer, machineNicknameLength);
        currBuffer += machineNicknameLength;
        memcpy(currBuffer, handleNameBuffer, handleNameLength);
        currBuffer += handleNameLength;
        memcpy(currBuffer, requestBuffer, requestLength);
        currBuffer += requestLength;
        memcpy(currBuffer, userBuffer, userLength);
        currBuffer += userLength;
        memcpy(currBuffer, endpointBuffer, endpointLength);
        currBuffer += endpointLength;
        memcpy(currBuffer, stafInstanceUUIDBuffer, stafInstanceUUIDLength);
        currBuffer += stafInstanceUUIDLength;
        memcpy(currBuffer, physicalInterfaceIDBuffer,
               physicalInterfaceIDLength);
        currBuffer += physicalInterfaceIDLength;

        /* Debug:
        cout << "Buffer length: " << bufferLength << endl;
        cout << endl << "Buffer: " << endl;

        for (unsigned int i = 0; i < (bufferLength - 8); ++i)
        {
            if ((i != 0) && (i % 24 == 0)) cout << endl;
            unsigned int currChar = buffer[i + 8];
            if (currChar < 16) cout << "0";
            cout << hex << currChar << dec << " ";
        }

        cout << endl;
        */

        connPtr->write(buffer, bufferLength);

        unsigned int reqRC = connPtr->readUInt();
        STAFString reqResult = connPtr->readString();

        *pResultBuffer = reqResult.adoptImpl();
        return reqRC;
    }
    catch (STAFException &e)
    {
        e.trace("PLSTAF.STAFServiceAcceptRequest");
    }
    catch (...)
    {
        STAFTrace::trace(
            kSTAFTraceError,
            "Caught unknown exception in PLSTAF.STAFServiceAcceptRequest()");
    }

    *pResultBuffer = STAFString().adoptImpl();

    return kSTAFUnknownError;
}
