#############################################################################
# Software Testing Automation Framework (STAF)                              #
# (C) Copyright IBM Corp. 2001                                              #
#                                                                           #
# This software is licensed under the Common Public License (CPL) V1.0.     #
#############################################################################

# This file contains tests for the STAF Tcl support
package require STAF
package require STAFMon
package require STAFLog

if {[STAF::Register "Tcl Test"] != 0} {
    puts "Error registering with STAF, RC: $STAF::RC"
    exit $STAF::RC
}

puts "Using handle $STAF::Handle"

#######################################
# First test some basic functionality #
#######################################

puts "Testing basic functionality"

STAF::Submit local ping ping

if {$STAF::Result != "PONG"} {
    puts "Wrong output for ping request"
    exit 1
}

if {[STAF::Submit local var "resolve string {STAF/Config/MachineNickname}"] != 0} {
    puts "Error resolving machine nickname, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
}

set self $STAF::Result

###############################################
# Next, lets test the monitor service wrapper #
###############################################

puts "Testing Monitor service functions"

# Log the message

set monitorMessage "Hello World"

if {[STAF::Monitor::Log $monitorMessage] != $STAF::kOk} {
    puts "Error on STAF::Monitor::Log, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Try to retrieve it

set request "query machine $self handle $STAF::Handle"

if {[STAF::Submit local monitor $request] != $STAF::kOk} {
    puts "Error querying monitor info, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Make sure we got back the correct message

if {[string first $monitorMessage $STAF::Result] == -1} {
    puts "Wrong output for monitor query request"
    puts "Expected to find '<Timestamp> $monitorMessage'"
    puts "Found '$STAF::Result'"
    exit 1
}

##########################################
# Now, lets test the log service wrapper #
##########################################

puts "Testing Log service functions"

# Setup logging

STAF::Log::Init TclTest handle "Fatal Error Warning Info"

# Log the message

set logMessage "A log message"

if {[STAF::Log::Log info $logMessage] != $STAF::kOk} {
    puts "Error on STAF::Log::Log, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Try to retrieve it

set request "query machine $self handle $STAF::Handle logname TclTest"

if {[STAF::Submit local log $request] != $STAF::kOk} {
    puts "Error on STAF::Log::Log, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Make sure we got back the correct message

if {[string first $logMessage $STAF::Result] == -1} {
    puts "Wrong output for log query request"
    puts "Expected to find '<Timestamp> $logMessage'"
    puts "Found '$STAF::Result'"
    exit 1
}

# Try to retrieve it from monitor

set request "query machine $self handle $STAF::Handle"

if {[STAF::Submit local Monitor $request] != $STAF::kOk} {
    puts "Error on querying monitor info, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Make sure we got back the correct message from monitor

if {[string first INFO:$logMessage $STAF::Result] == -1} {
    puts "Wrong output for monitor query request"
    puts "Expected to find '<Timestamp> INFO:$logMessage'"
    puts "Found '$STAF::Result'"
    exit 1
}

# Delete the log file

set request "delete machine $self handle $STAF::Handle logname TclTest confirm" 

if {[STAF::Submit local log $request] != 0} {
    puts "Error deleting log file, RC: $STAF::RC, Result: $STAF::Result"
    exit 1
}

# Log the message so that Monitor shouldn't get it

if {[STAF::Log::Log status $logMessage] != $STAF::kOk} {
    puts "Error on STAF::Log::Log, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Try to retrieve it

set request "query machine $self handle $STAF::Handle logname TclTest"

if {[STAF::Submit local log $request] != $STAF::kOk} {
    puts "Error on STAF::Log::Log, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Make sure we got back the correct message

if {[string first $logMessage $STAF::Result] == -1} {
    puts "Wrong output for log query request"
    puts "Expected to find '<Timestamp> $logMessage'"
    puts "Found '$STAF::Result'"
    exit 1
}

# Try to retrieve it from monitor

set request "query machine $self handle $STAF::Handle"

if {[STAF::Submit local Monitor $request] != $STAF::kOk} {
    puts "Error on querying monitor info, RC: $STAF::RC, Result: $STAF::Result"
    exit $STAF::RC
} 

# Make sure we got back the correct (old) message from monitor

if {[string first INFO:$logMessage $STAF::Result] == -1} {
    puts "Wrong output for monitor query request"
    puts "Expected to find '<Timestamp> INFO:$logMessage'"
    puts "Found '$STAF::Result'"
    exit 1
}

##########################
# Test privacy functions #
##########################

puts "Testing privacy functions"

set data "secret"
set dataWithPD [STAF::AddPrivacyDelimiters $data]

set expectedResult "!!@secret@!!"
if {[STAF::AddPrivacyDelimiters $data] != $expectedResult} {
    puts "Error: \[STAF::AddPrivacyDelimiters $data\] : [STAF::AddPrivacyDelimiters $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "^!!@secret^@!!"
if {[STAF::EscapePrivacyDelimiters $dataWithPD] != $expectedResult} {
    puts "Error: \[STAF::EscapePrivacyDelimiters $dataWithPD\] = [STAF::EscapePrivacyDelimiters $dataWithPD]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "************"
if {[STAF::MaskPrivateData $dataWithPD] != $expectedResult} {
    puts "Error: \[STAF::MaskPrivateData $dataWithPD\] = [STAF::MaskPrivateData $dataWithPD]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "secret"
if {[STAF::RemovePrivacyDelimiters $dataWithPD] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $dataWithPD\] = [STAF::RemovePrivacyDelimiters $dataWithPD]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set data "!!@Msg: ^!!@Pw is ^^!!@secret^^@!!.^@!!@!!"
set expectedResult "Msg: Pw is secret."
if {[STAF::RemovePrivacyDelimiters $data] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $data\] = [STAF::RemovePrivacyDelimiters $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "Msg: Pw is secret."
if {[STAF::RemovePrivacyDelimiters $data 0] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $data 0\] = [STAF::RemovePrivacyDelimiters $data 0]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "Msg: !!@Pw is ^!!@secret^@!!.@!!"
if {[STAF::RemovePrivacyDelimiters $data 1] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $data 1\] = [STAF::RemovePrivacyDelimiters $data 1]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

set expectedResult "Msg: Pw is !!@secret@!!."
if {[STAF::RemovePrivacyDelimiters $data 2] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $data 2\] = [STAF::RemovePrivacyDelimiters $data 2]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

# Test private methods passing in an empty string

set data ""
set expectedResult ""
if {[STAF::AddPrivacyDelimiters $data] != $expectedResult} {
    puts "Error: \[STAF::AddPrivacyDelimiters $data\] : [STAF::AddPrivacyDelimiters $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

if {[STAF::EscapePrivacyDelimiters $data] != $expectedResult} {
    puts "Error: \[STAF::EscapePrivacyDelimiters $data\] : [STAF::EscapePrivacyDelimiters $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

if {[STAF::MaskPrivateData $data] != $expectedResult} {
    puts "Error: \[STAF::MaskPrivateData $data\] : [STAF::MaskPrivateData $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

if {[STAF::RemovePrivacyDelimiters $data] != $expectedResult} {
    puts "Error: \[STAF::RemovePrivacyDelimiters $data\] : [STAF::RemovePrivacyDelimiters $data]"
    puts "       Should return the following instead: $expectedResult"
    exit 1
}

##################################
# Now, let's test STAF::datatype #
##################################

puts "Testing data type functions"

# For each type (None is handled a little differently),
#   create (without value)
#   validate type
#   get value
#   set value
#   get value
#   create (with value)
#   get value
#   set value
#   get value
#
# For list and map, also do the following
#   append to STAF::datatype value
#   get value

# Testing data type None

puts "Testing data type: None"

set dtNone [STAF::datatype createNone]

if {[STAF::datatype getType $dtNone] != $STAF::NoneType} {
    puts "Wrong type for a None data type"
    puts "Expected: $STAF::NoneType"
    puts "Found: [STAF::datatype getType $dtNone]"
    exit 1
}

if {[STAF::datatype getValue $dtNone] != "None"} {
    puts "Wrong value for a None data type"
    puts "Expected: None"
    puts "Found: [STAF::datatype getValue $dtNone]"
    exit 1
}

if {[catch {STAF::datatype setValue dtNone "Ok"}] == 0} {
    puts "Error: Successfully set the type of a None data type"
    exit 1
}

set dtNone2 [STAF::datatype createNone "NoneValue"]

if {[STAF::datatype getValue $dtNone2] != "None"} {
    puts "Wrong value for a None data type"
    puts "Expected: None"
    puts "Found: [STAF::datatype getValue $dtNone]"
    exit 1
}

# Testing data type Context

puts "Testing data type: Context"

set dtContext [STAF::datatype createContext]

if {[STAF::datatype getType $dtContext] != $STAF::ContextType} {
    puts "Wrong type for a Context data type"
    puts "Expected: $STAF::ContextType"
    puts "Found: [STAF::datatype getType $dtContext]"
    exit 1
}

# Testing data type Scalar

puts "Testing data type: Scalar"

set dtScalar [STAF::datatype createScalar]

if {[STAF::datatype getType $dtScalar] != "$STAF::ScalarType"} {
    puts "Wrong type for a Scalar data type"
    puts "Expected: $STAF::ScalarType"
    puts "Found: [STAF::datatype getType $dtScalar]"
    exit 1
}

if {[STAF::datatype getValue $dtScalar] != ""} {
    puts "Wrong value for data type"
    puts "Expected: <Empty string>"
    puts "Found: [STAF::datatype getValue $dtScalar]"
    exit 1
}

set scalarTestValue  "Scalar Test Value"

STAF::datatype setValue dtScalar $scalarTestValue

if {[STAF::datatype getValue $dtScalar] != $scalarTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $scalarTestValue"
    puts "Found: [STAF::datatype getValue $dtScalar]"
    exit 1
}

set scalarInitValue "Scalar Init Value"
set dtScalar [STAF::datatype createScalar $scalarInitValue]

if {[STAF::datatype getValue $dtScalar] != $scalarInitValue} {
    puts "Wrong value for data type"
    puts "Expected: $scalarInitValue"
    puts "Found: [STAF::datatype getValue $dtScalar]"
    exit 1
}

STAF::datatype setValue dtScalar $scalarTestValue

if {[STAF::datatype getValue $dtScalar] != $scalarTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $scalarTestValue"
    puts "Found: [STAF::datatype getValue $dtScalar]"
    exit 1
}

# Testing data type List

puts "Testing data type: List"

set dtList [STAF::datatype createList]

if {[STAF::datatype getType $dtList] != "$STAF::ListType"} {
    puts "Wrong type for a List data type"
    puts "Expected: $STAF::ListType"
    puts "Found: [STAF::datatype getType $dtList]"
    exit 1
}

if {[STAF::datatype getValue $dtList] != ""} {
    puts "Wrong value for data type"
    puts "Expected: <Empty string>"
    puts "Found: [STAF::datatype getValue $dtList]"
    exit 1
}

set listTestValue [list "List" "Test" "Value"]

STAF::datatype setValue dtList $listTestValue

if {[STAF::datatype getValue $dtList] != $listTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $listTestValue"
    puts "Found: [STAF::datatype getValue $dtList]"
    exit 1
}

set listInitValue [list "List" "Init" "Value"]
set dtList [STAF::datatype createList $listInitValue]

if {[STAF::datatype getValue $dtList] != $listInitValue} {
    puts "Wrong value for data type"
    puts "Expected: $listInitValue"
    puts "Found: [STAF::datatype getValue $dtList]"
    exit 1
}

STAF::datatype setValue dtList $listTestValue

if {[STAF::datatype getValue $dtList] != $listTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $listTestValue"
    puts "Found: [STAF::datatype getValue $dtList]"
    exit 1
}

lappend listTestValue "And"
lappend listTestValue "More"
lappend dtList "And"
lappend dtList "More"

if {[STAF::datatype getValue $dtList] != $listTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $listTestValue"
    puts "Found: [STAF::datatype getValue $dtList]"
    exit 1
}

# Testing data type Map

puts "Testing data type: Map"

set dtMap [STAF::datatype createMap]

if {[STAF::datatype getType $dtMap] != "$STAF::MapType"} {
    puts "Wrong type for a Map data type"
    puts "Expected: $STAF::MapType"
    puts "Found: [STAF::datatype getType $dtMap]"
    exit 1
}

if {[STAF::datatype getValue $dtMap] != ""} {
    puts "Wrong value for data type"
    puts "Expected: <Empty string>"
    puts "Found: [STAF::datatype getValue $dtMap]"
    exit 1
}

set mapTest(key1) value1
set mapTest(key2) value2
set mapTestValue [array get mapTest]

STAF::datatype setValue dtMap $mapTestValue

if {[STAF::datatype getValue $dtMap] != $mapTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $mapTestValue"
    puts "Found: [STAF::datatype getValue $dtMap]"
    exit 1
}

set mapInit(key3) value3
set mapInit(key4) value4
set mapInitValue [array get mapInit]
set dtMap [STAF::datatype createMap $mapInitValue]

if {[STAF::datatype getValue $dtMap] != $mapInitValue} {
    puts "Wrong value for data type"
    puts "Expected: $mapInitValue"
    puts "Found: [STAF::datatype getValue $dtMap]"
    exit 1
}

STAF::datatype setValue dtMap $mapTestValue

if {[STAF::datatype getValue $dtMap] != $mapTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $mapTestValue"
    puts "Found: [STAF::datatype getValue $dtMap]"
    exit 1
}

set mapAdditionalValue(key5) value5
append mapTestValue " " [array get mapAdditionalValue]
append dtMap        " " [array get mapAdditionalValue]

if {[STAF::datatype getValue $dtMap] != $mapTestValue} {
    puts "Wrong value for data type"
    puts "Expected: $mapTestValue"
    puts "Found: [STAF::datatype getValue $dtMap]"
    exit 1
}

############################################
# Now, let's test STAF::mapclassdefinition #
############################################

puts "Testing map class definition functions"

# create

set mcdName "foo"

set mcd [STAF::mapclassdef create $mcdName]

if {[STAF::mapclassdef getName $mcd] != $mcdName} {
    puts "Wrong name for map class def"
    puts "Expected: $mcdName"
    puts "Found: [STAF::mapclassdef getName $mcd]"
    exit 1
}

# addKey
# setKeyProperty
# getKeys

set mcdKey1Name        "key1"
set mcdKey1DisplayName "Name 1"
set mcdKey2Name        "key2"
set mcdKey2DisplayName "Name 2"
set mcdKey2PName       "display-short-name"
set mcdKey2PValue      "N2"

STAF::mapclassdef addKey mcd $mcdKey1Name $mcdKey1DisplayName
STAF::mapclassdef addKey mcd $mcdKey2Name $mcdKey2DisplayName
STAF::mapclassdef setKeyProperty mcd $mcdKey2Name $mcdKey2PName $mcdKey2PValue

set mcdKeysObj [STAF::mapclassdef getKeys $mcd]

if {[STAF::datatype getType $mcdKeysObj] != "$STAF::ListType"} {
    puts "Wrong type for map class def keys"
    puts "Expected: $STAF::ListType"
    puts "Found: [STAF::datatype getType $mcdKeysObj]"
    exit 1
}

foreach mcdKeyObj [STAF::datatype getValue $mcdKeysObj] {
    if {[STAF::datatype getType $mcdKeyObj] != "$STAF::MapType"} {
        puts "Wrong type for a map class def key"
        puts "Expected: $STAF::MapType"
        puts "Found: [STAF::datatype getType $mcdKeysObj]"
        exit 1
    }

    array set mcdKey [STAF::datatype getValue $mcdKeyObj]

    if {$mcdKey(key) == $mcdKey1Name} {
        if {$mcdKey(display-name) != $mcdKey1DisplayName} {
            puts "Error: Invalid key display-name in map class definition"
            puts "Expected: $mcdKey1DisplayName"
            puts "Found: $mcdKey(display-name)"
            exit 1
        }
    } elseif {$mcdKey(key) == $mcdKey2Name} {
        if {$mcdKey(display-name) != $mcdKey2DisplayName} {
            puts "Error: Invalid key display-name in map class definition"
            puts "Expected: $mcdKey2DisplayName"
            puts "Found: $mcdKey(display-name)"
            exit 1
        }

        if {$mcdKey($mcdKey2PName) != $mcdKey2PValue} {
            puts "Error: Invalid key property value in map class definition"
            puts "Expected: $mcdKey2PValue"
            puts "Found: $mcdKey($mcdKey2PName)"
            exit 1
        }
    } else {
        puts "Error: Invalid key in map class definition"
        puts "Expected: <No key>"
        puts "Found: $mcdKey(key)"
        exit 1
    }
}

# createInstance

set mcdInstanceObj [STAF::mapclassdef createInstance $mcd]

if {[STAF::datatype getType $mcdInstanceObj] != "$STAF::MapType"} {
    puts "Wrong type for map class def keys"
    puts "Expected: $STAF::MapType"
    puts "Found: [STAF::datatype getType $mcdInstanceObj]"
    exit 1
}

array set mcdInstance [STAF::datatype getValue $mcdInstanceObj]

if {$mcdInstance(staf-map-class-name) != $mcdName} {
    puts "Wrong map class definition type for map class definition instance"
    puts "Expected: $mcdName"
    puts "Found: $mcdInstance(staf-map-class-name)"
    exit 1
}

# getMapClassDefinitionObject

if {[STAF::mapclassdef getMapClassDefinitionObject $mcd] != $mcd} {
    puts "Wrong map class definition object"
    puts "Expected: $mcd"
    puts "Found: [STAF::mapclassdef getMapClassDefinitionObject $mcd]"
    exit 1
}

# create(with map class definition)

set mcd2Name "bar"

set mcd2 [STAF::mapclassdef create $mcd2Name [STAF::mapclassdef getMapClassDefinitionObject $mcd]]

if {[STAF::mapclassdef getName $mcd2] != $mcdName} {
    puts "Wrong name for map class def"
    puts "Expected: $mcdName"
    puts "Found: [STAF::mapclassdef getName $mcd2]"
    exit 1
}

if {[STAF::mapclassdef getMapClassDefinitionObject $mcd2] != $mcd} {
    puts "Wrong map class definition object"
    puts "Expected: $mcd"
    puts "Found: [STAF::mapclassdef getMapClassDefinitionObject $mcd2]"
    exit 1
}

##################################
# Now, let's test STAF::mcontext #
##################################

puts "Testing marshalling context functions"

# create

set context [STAF::mcontext create]

if {[STAF::datatype getType $context] != "$STAF::ContextType"} {
    puts "Wrong type for marshalling context"
    puts "Expected: $STAF::ContextType"
    puts "Found: [STAF::datatype getType $context]"
    exit 1
}

if {[STAF::mcontext getRootObject $context] != [STAF::datatype createNone]} {
    puts "Error: New marshalling context should have a root object of None"
    puts "Found: [STAF::mcontext getRootObject $context]"
    exit 1
}

# XXX: Need an actual test of some type here

# setMapClassDefinition
# hasMapClassDefinition
# getMapClassDefinition

STAF::mcontext setMapClassDefinition context $mcd

if {! [STAF::mcontext hasMapClassDefinition $context $mcdName]} {
    puts "Error: marshalling context does not contain map class definition"
    puts "Expected to find map class definition $mcdName"
    puts "Did not find it"
    exit 1
}

set testMCD [STAF::mcontext getMapClassDefinition $context $mcdName]

if {$testMCD != $mcd} {
    puts "Error: map class definitions do not match"
    puts "Map class 1: $testMCD"
    puts "Map class 2: $mcd"
    exit 1
}

# getMapClassMap

set mapClassMap [STAF::mcontext getMapClassMap $context]

# getMapClassDefinitionNames

set mcd3Name "baz"
set mcd3 [STAF::mapclassdef create $mcd3Name]

STAF::mapclassdef addKey mcd3 $mcdKey1Name $mcdKey1DisplayName
STAF::mapclassdef addKey mcd3 $mcdKey2Name $mcdKey2DisplayName

STAF::mcontext setMapClassDefinition context $mcd3

set mcdNames [STAF::mcontext getMapClassDefinitionNames $context]

if {[llength $mcdNames] != 2} {
    puts "Error: Incorrect number of map class definition in marshalling context"
    puts "Expected: 2"
    puts "Found: [llength $mcdNames]"
    exit 1
}

if {( [lsearch -exact $mcdNames $mcdName] == -1 ) ||
    ( [lsearch -exact $mcdNames $mcd3Name] == -1 )} {
    puts "Error: Invalid map class definition names in marshalling context"
    puts "Expected: $ncdName and $mcd3Name"
    puts "Found: $mcdNames"
    exit 1
}

# setRootObject
# getRootObject

set testArray(a) "Hello"
set testArray(b) "World"
set testArrayString [array get testArray]
set arrayObject [STAF::datatype createMap $testArrayString]

STAF::mcontext setRootObject context $arrayObject
set testArrayObject [STAF::mcontext getRootObject $context]

if {$testArrayObject != $arrayObject} {
    puts "Error: Incorrect root object"
    puts "Expected: $arrayObject"
    puts "Found: $testArrayObject"
    exit 1
}

# getPrimaryObject

if {[STAF::mcontext getPrimaryObject $context] != $context} {
    puts "Error: Wrong primary object for marshalling context"
    puts "Expected: $context"
    puts "Found: [STAF::mcontext getPrimaryObject $context]"
    exit 1
}

set context2 [STAF::mcontext create]
STAF::mcontext setRootObject context2 $arrayObject

if {[STAF::mcontext getPrimaryObject $context2] != $arrayObject} {
    puts "Error: Wrong primary object for marshalling context"
    puts "Expected: $arrayObject"
    puts "Found: [STAF::mcontext getPrimaryObject $context2]"
    exit 1
}

#############
# Finish up #
#############

if {[STAF::UnRegister] != 0} {
    puts "Error unregistering with STAF, RC: $STAF::RC"
    exit $STAF::RC
}

puts "All tests successful"

exit 0
