/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include <vector>
#include <list>
#include <map>
#include "STAF_fstream.h"
#include "STAFString.h"
#include "STAFError.h"
#include "STAFException.h"
#include "STAFRefPtr.h"
#include "STAFMutexSem.h"
#include "STAFEventSem.h"
#include "STAFRWSem.h"
#include "STAFCommandParser.h"
#include "STAFServiceInterface.h"
#include "STAFTimestamp.h"
#include "STAFUtil.h"
#include "STAFInternalUtil.h"
#include "STAFResPoolService.h"
#include "STAFFileSystem.h"

int debug = 0; // XXX

typedef STAFRefPtr<STAFCommandParser> STAFCommandParserPtr;

// Resource Data - contains data for a entry in a resource pool

struct ResourceData
{
    ResourceData() : owned(0)
    { /* Do Nothing */ }

    ResourceData(const STAFString &aEntry) : entry(aEntry), owned(0)
    { /* Do Nothing */ }

    STAFString   entry;            // Entry value
    unsigned int owned;            // 0 means Available; 1 means Owned
    STAFString   orgUUID;          // Originating request's STAF UUID
    STAFString   orgMachine;       // Originating request's machine name
    STAFString   orgName;          // Originating request's handle name
    STAFHandle_t orgHandle;        // Originating request's handle
    STAFString   orgUser;          // Originating request's handle user
    STAFString   orgEndpoint;      // Originating request's endpoint
    STAFString   requestedTime;    // Time made request
    STAFString   acquiredTime;     // Time acquired the resource
};

// ResourceList --  Ordered list of entries in a resource pool

typedef std::vector<ResourceData> ResourceList;

// Request Data - contains data for a pending request for a resource

struct RequestData
{
    RequestData() : wakeup(new STAFEventSem(), STAFEventSemPtr::INIT),
                    retCode(kSTAFTimeout)
    {
        requestedTime = STAFTimestamp::now().asString();
        wakeup->reset();
        *garbageCollectedPtr = false;
    }

    RequestData(const STAFString &aUUID, const STAFString &aMachine,
                const STAFString &aHandleName, const STAFHandle_t aHandle,
                const STAFString aUser, const STAFString aEndpoint)
        : orgUUID(aUUID), orgMachine(aMachine), orgName(aHandleName),
          orgHandle(aHandle), orgUser(aUser), orgEndpoint(aEndpoint),
          wakeup(new STAFEventSem(), STAFEventSemPtr::INIT),
          retCode(kSTAFTimeout),
          garbageCollectedPtr(new bool, STAFRefPtr<bool>::INIT)
    {
        requestedTime = STAFTimestamp::now().asString();
        wakeup->reset();
        *garbageCollectedPtr = false;
    }

    STAFString      orgUUID;       // Originating request's STAF UUID
    STAFString      orgMachine;    // Originating request's machine name
    STAFString      orgName;       // Originating request's handle name
    STAFHandle_t    orgHandle;     // Originating request's STAF handle
    STAFString      orgUser;       // Originating request's user
    STAFString      orgEndpoint;   // Originating request's endpoint
    STAFString      requestedTime; // Originating request's date/time
    STAFEventSemPtr wakeup;        // Semaphore to wake up a pending request
    STAFRC_t        retCode;       // Return code indicating if request was
                                   //   successfully satisfied
    STAFString      resultBuffer;  // Entry obtained if retCode = 0
    STAFRefPtr<bool> garbageCollectedPtr;
};

typedef STAFRefPtr<RequestData> RequestDataPtr;

// RequestList  --  Ordered list of pending requests for a pool

typedef std::list<RequestDataPtr> RequestList;

static const unsigned int sCurrFileFormat = 1;

// Resource Pool Data - contains all data for a resource pool

struct PoolData
{
    PoolData() : usedResources(0),
                 accessSem(new STAFMutexSem(), STAFMutexSemPtr::INIT)
    { /* Do nothing */ }

    PoolData(const STAFString &aPoolName, const STAFString &aPoolDescription)
        : poolName(aPoolName), poolDescription(aPoolDescription), 
          numResources(0), usedResources(0), 
          accessSem(new STAFMutexSem(), STAFMutexSemPtr::INIT)
    { fileFormat = sCurrFileFormat; }

    unsigned int    fileFormat;      // Format of the pool file
                                     //  "0" - REXX version in STAF 2.2 and <
                                     //   1  - C++ version in STAF 2.3 and later
    STAFString      poolName;        // Pool Name
    STAFString      poolDescription; // Pool Description
    unsigned int    numResources;    // Total # of entries in ResourceList
    unsigned int    usedResources;   // # of entries used in ResourceList
    ResourceList    resourceList;    // List of entries in a resource pool
    RequestList     requestList;     // List of pending requests
    STAFMutexSemPtr accessSem;       // Semaphore to control access to PoolData
};

typedef STAFRefPtr<PoolData> PoolDataPtr;


// PoolMap -- KEY:   Pool name in upper case, 
//            VALUE: Pointer to PoolData information

typedef std::map<STAFString, PoolDataPtr> PoolMap;

// Read/Write File Return Codes

enum ReadFileRC
{
    kReadorWriteOk = 0,
    kReadEndOfFile = 1,
    kReadInvalidFormat = 2,
    kFileOpenError = 3
};

// RESPOOL Service Data

struct ResPoolServiceData
{
    unsigned int  fDebugMode;            // Debug Mode flag
    STAFString    fShortName;            // Short service name
    STAFString    fName;                 // Registered service name
    STAFString    fLocalMachineName;     // Logical identifier for the local
                                         //   machine
    STAFString    fPoolDir;              // Pool Directory
    STAFHandlePtr fHandlePtr;            // Respool service's STAF handle
    STAFCommandParserPtr fCreateParser;  // RESPOOL CREATE command parser
    STAFCommandParserPtr fDeleteParser;  // RESPOOL DELETE command parser
    STAFCommandParserPtr fQueryParser;   // RESPOOL QUERY command parser
    STAFCommandParserPtr fRequestParser; // RESPOOL REQUEST command parser
    STAFCommandParserPtr fAddParser;     // RESPOOL ADD command parser
    STAFCommandParserPtr fRemoveParser;  // RESPOOL REMOVE command parser
    STAFCommandParserPtr fReleaseParser; // RESPOOL RELEASE command parser
    STAFCommandParserPtr fListParser;    // RESPOOL LIST command parser
    STAFCommandParserPtr 
        fSTAFCallbackParser; // RESPOOL STAF_CALLBACK command parser
    STAFCommandParserPtr fHelpParser;    // RESPOOL HELP command parser
    STAFCommandParserPtr fVersionParser; // RESPOOL VERSION command parser
    STAFCommandParserPtr fParmsParser;   // RESPOOL PARMS command parser

    STAFRWSemPtr         fPoolMapRWSem;  // Read/Write semaphore to control 
                                         //   access to the PoolMap
    PoolMap              fPoolMap;       // Map of all resource pools

    // Map Class Definitions for marshalled results
    STAFMapClassDefinitionPtr fPoolClass; 
    STAFMapClassDefinitionPtr fPoolInfoClass;
    STAFMapClassDefinitionPtr fSettingsClass;
    STAFMapClassDefinitionPtr fRequestClass;
    STAFMapClassDefinitionPtr fResourceClass;
    STAFMapClassDefinitionPtr fResourceOwnerClass;
};

typedef STAFRefPtr<ResPoolServiceData> ResPoolServiceDataPtr;

// Static Variables

static STAFString sLineSep;
static STAFString sPoolExt("rpl");
static const STAFString sVersionInfo("3.1.1");
static const STAFString sPool("POOL");
static const STAFString sDescription("DESCRIPTION");
static const STAFString sFirst("FIRST");
static const STAFString sTimeout("TIMEOUT");
static const STAFString sEntry("ENTRY");
static const STAFString sForce("FORCE");
static const STAFString sDirectory("DIRECTORY");
static const STAFString sLeftCurlyBrace(kUTF8_LCURLY);
static const STAFString sSemiColon(kUTF8_SCOLON);
static const STAFString sColon(kUTF8_COLON);
static const STAFString sSlash(kUTF8_SLASH);
static const STAFString sSpecSeparator(sColon + sSlash + sSlash);
static const STAFString sLocal("local");
static const STAFString sHelp("help");
static const STAFString sVar("var");
static const STAFString sResStrResolve("RESOLVE REQUEST ");
static const STAFString sString(" STRING ");

// Prototypes

static STAFResultPtr handleCreate(STAFServiceRequestLevel30 *, 
                                  ResPoolServiceData *);
static STAFResultPtr handleDelete(STAFServiceRequestLevel30 *,
                                  ResPoolServiceData *);
static STAFResultPtr handleQuery(STAFServiceRequestLevel30 *,
                                 ResPoolServiceData *);
static STAFResultPtr handleRequest(STAFServiceRequestLevel30 *,
                                   ResPoolServiceData *);
static STAFResultPtr handleRelease(STAFServiceRequestLevel30 *,
                                   ResPoolServiceData *);
static STAFResultPtr handleAdd(STAFServiceRequestLevel30 *,
                               ResPoolServiceData *);
static STAFResultPtr handleRemove(STAFServiceRequestLevel30 *,
                                  ResPoolServiceData *);
static STAFResultPtr handleSTAFCallback(STAFServiceRequestLevel30 *,
                                        ResPoolServiceData *);
static STAFResultPtr handleList(STAFServiceRequestLevel30 *,
                                ResPoolServiceData *);
static STAFResultPtr handleHelp(STAFServiceRequestLevel30 *,
                                ResPoolServiceData *);
static STAFResultPtr handleVersion(STAFServiceRequestLevel30 *,
                                   ResPoolServiceData *);

static STAFResultPtr resolveStr(STAFServiceRequestLevel30 *pInfo, 
                                ResPoolServiceData *pData,
                                const STAFString &theString);

static STAFResultPtr resolveOp(STAFServiceRequestLevel30 *pInfo, 
                               ResPoolServiceData *pData,
                               STAFCommandParseResultPtr &parsedResult,
                               const STAFString &fOption,
                               unsigned int optionIndex = 1);

STAFResultPtr resolveOpLocal(ResPoolServiceData *pData,
                             STAFCommandParseResultPtr &parsedResult,
                             const STAFString &fOption,
                             unsigned int optionIndex = 1);

static void registerHelpData(ResPoolServiceData *pData,
                             unsigned int errorNumber,
                             const STAFString &shortInfo,
                             const STAFString &longInfo);

static void unregisterHelpData(ResPoolServiceData *pData,
                               unsigned int errorNumber);

void readUIntFromFile(istream &input, unsigned int &data,
                      unsigned int length = 4);

void writeUIntToFile(ostream &output, unsigned int data,
                     unsigned int length = 4);

void readStringFromFile(istream &input, STAFString &inString);

void writeStringToFile(ostream &output, STAFString &outString);

unsigned int readPoolFile(const STAFString &fileName, PoolData &poolData);

unsigned int writePoolFile(const STAFString &fileName, PoolData &poolData);


// Begin implementation

STAFRC_t STAFServiceGetLevelBounds(unsigned int levelID,
                                   unsigned int *minimum,
                                   unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceInit:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceAcceptRequest:
        {
            *minimum = 30;
            *maximum = 30;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return kSTAFInvalidAPILevel;
        }
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                              void *pServiceInfo, unsigned int infoLevel,
                              STAFString_t *pErrorBuffer)
{
    STAFRC_t rc = kSTAFUnknownError;

    try
    {
        if (infoLevel != 30) return kSTAFInvalidAPILevel;

        STAFServiceInfoLevel30 *pInfo =
            reinterpret_cast<STAFServiceInfoLevel30 *>(pServiceInfo);

        ResPoolServiceData data;
        data.fDebugMode = 0;
        data.fShortName = pInfo->name;
        data.fName = "STAF/Service/";
        data.fName += pInfo->name;
        
        for (unsigned int i = 0; i < pInfo->numOptions; ++i)
        {            
            if (STAFString(pInfo->pOptionName[i]).upperCase() == "DEBUG")
            {
                data.fDebugMode = 1;
            }           
            else
            {
                STAFString optionError(pInfo->pOptionName[i]);
                *pErrorBuffer = optionError.adoptImpl();
                return kSTAFServiceConfigurationError;
            }
        }        

        // Set service handle

        *pServiceHandle = new ResPoolServiceData(data);

        return kSTAFOk;
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                "STAFResPoolService.cpp: STAFServiceConstruct").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFResPoolService.cpp: STAFServiceConstruct: Caught "
                    "unknown exception in STAFServiceConstruct()");
        *pErrorBuffer = error.adoptImpl();
    }    

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceInit(STAFServiceHandle_t serviceHandle,
                         void *pInitInfo, unsigned int initLevel,
                         STAFString_t *pErrorBuffer)
{
    STAFRC_t retCode = kSTAFUnknownError;

    try
    {    
        if (initLevel != 30) return kSTAFInvalidAPILevel;

        ResPoolServiceData *pData =
            reinterpret_cast<ResPoolServiceData *>(serviceHandle);
        
        STAFServiceInitLevel30 *pInfo =
            reinterpret_cast<STAFServiceInitLevel30 *>(pInitInfo);        

        retCode = STAFHandle::create(pData->fName, pData->fHandlePtr);
        
        if (retCode != kSTAFOk)
            return retCode;
        
        //CREATE options
        pData->fCreateParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fCreateParser->addOption("CREATE", 1,
                                       STAFCommandParser::kValueNotAllowed);
        pData->fCreateParser->addOption("POOL", 1,
                                       STAFCommandParser::kValueRequired);
        pData->fCreateParser->addOption("DESCRIPTION", 1,
                                       STAFCommandParser::kValueRequired);
        pData->fCreateParser->addOptionNeed("CREATE", "POOL");
        pData->fCreateParser->addOptionNeed("CREATE", "DESCRIPTION");
        
        //DELETE options
        pData->fDeleteParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fDeleteParser->addOption("DELETE", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOption("POOL", 1,
                                        STAFCommandParser::kValueRequired);
        pData->fDeleteParser->addOption("CONFIRM",  1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOption("FORCE",  1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOptionNeed("DELETE", "POOL");
        pData->fDeleteParser->addOptionNeed("DELETE", "CONFIRM");
        
        //QUERY options
        pData->fQueryParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fQueryParser->addOption("QUERY", 1,
                                       STAFCommandParser::kValueNotAllowed);
        pData->fQueryParser->addOption("POOL", 1,
                                       STAFCommandParser::kValueRequired);
        pData->fQueryParser->addOptionNeed("QUERY", "POOL");

        //REQUEST options
        pData->fRequestParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fRequestParser->addOption("REQUEST", 1,
                                         STAFCommandParser::kValueNotAllowed);
        pData->fRequestParser->addOption("POOL", 1,
                                         STAFCommandParser::kValueRequired);
        pData->fRequestParser->addOption("TIMEOUT", 1,
                                         STAFCommandParser::kValueRequired);
        pData->fRequestParser->addOption("FIRST", 1,
                                         STAFCommandParser::kValueNotAllowed);
        pData->fRequestParser->addOption("RANDOM", 1,
                                         STAFCommandParser::kValueNotAllowed);
        pData->fRequestParser->addOptionGroup("FIRST RANDOM", 0, 1);
        pData->fRequestParser->addOptionNeed("REQUEST", "POOL");

        //ADD options
        pData->fAddParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fAddParser->addOption("ADD", 1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fAddParser->addOption("POOL", 1,
                                     STAFCommandParser::kValueRequired);
        pData->fAddParser->addOption("ENTRY", 0,
                                     STAFCommandParser::kValueRequired);
        pData->fAddParser->addOptionNeed("ADD", "POOL");
        pData->fAddParser->addOptionNeed("ADD", "ENTRY");

        //REMOVE options
        pData->fRemoveParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fRemoveParser->addOption("REMOVE", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fRemoveParser->addOption("POOL", 1,
                                        STAFCommandParser::kValueRequired);
        pData->fRemoveParser->addOption("ENTRY", 0,
                                        STAFCommandParser::kValueRequired);
        pData->fRemoveParser->addOption("CONFIRM", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fRemoveParser->addOption("FORCE", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fRemoveParser->addOptionNeed("REMOVE", "POOL");
        pData->fRemoveParser->addOptionNeed("REMOVE", "ENTRY");
        pData->fRemoveParser->addOptionNeed("REMOVE", "CONFIRM");

        //RELEASE options
        pData->fReleaseParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fReleaseParser->addOption("RELEASE", 1,
                                         STAFCommandParser::kValueNotAllowed);
        pData->fReleaseParser->addOption("POOL", 1,
                                         STAFCommandParser::kValueRequired);
        pData->fReleaseParser->addOption("ENTRY", 1,
                                         STAFCommandParser::kValueRequired);
        pData->fReleaseParser->addOption("FORCE", 1,
                                         STAFCommandParser::kValueNotAllowed);
        pData->fReleaseParser->addOptionNeed("RELEASE", "POOL");
        pData->fReleaseParser->addOptionNeed("RELEASE", "ENTRY");

        //LIST options
        pData->fListParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fListParser->addOption("LIST", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOption("POOLS", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOption("SETTINGS", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOptionNeed("POOLS SETTINGS", "LIST");
        pData->fListParser->addOptionGroup("POOLS SETTINGS", 0, 1);
                                      
        //STAF_CALLBACK options
        pData->fSTAFCallbackParser = STAFCommandParserPtr(new 
            STAFCommandParser, STAFCommandParserPtr::INIT);
        pData->fSTAFCallbackParser->addOption("STAF_CALLBACK", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fSTAFCallbackParser->addOption("HANDLEDELETED", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fSTAFCallbackParser->addOption("HANDLE", 1,
                                      STAFCommandParser::kValueRequired);
        pData->fSTAFCallbackParser->addOption("UUID", 1,
                                      STAFCommandParser::kValueRequired);
        pData->fSTAFCallbackParser->addOption("MACHINE", 1,
                                      STAFCommandParser::kValueRequired);
        pData->fSTAFCallbackParser->addOption("KEY", 1,
                                      STAFCommandParser::kValueRequired);

        //HELP options
        pData->fHelpParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fHelpParser->addOption("HELP", 1,
                                      STAFCommandParser::kValueNotAllowed);

        //VERSION options
        pData->fVersionParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fVersionParser->addOption("VERSION", 1,
                                         STAFCommandParser::kValueNotAllowed);

        // PARMS        
        pData->fParmsParser = STAFCommandParserPtr(new STAFCommandParser,
                                                   STAFCommandParserPtr::INIT);
        pData->fParmsParser->addOption("DIRECTORY", 1,
                                       STAFCommandParser::kValueRequired);
        STAFCommandParseResultPtr parsedResult = 
            pData->fParmsParser->parse(pInfo->parms);

        if (parsedResult->rc != kSTAFOk) 
        {
            *pErrorBuffer = parsedResult->errorBuffer.adoptImpl();
            return parsedResult->rc;
        }

        if (parsedResult->optionTimes(sDirectory))
        {
            STAFResultPtr dirResult = resolveOpLocal(pData, parsedResult,
                                                     sDirectory);

            if (dirResult->rc != kSTAFOk)
            {
                *pErrorBuffer = dirResult->result.adoptImpl();
                return dirResult->rc;
            }

            pData->fPoolDir = dirResult->result;
        }
        
        // Construct the map class for general pool information output

        pData->fPoolClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/Pool");

        pData->fPoolClass->addKey("poolName",    "Pool Name");
        pData->fPoolClass->addKey("description", "Description");
        
        // Construct the map class for listing service settings

        pData->fSettingsClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/Settings");

        pData->fSettingsClass->addKey("directory",  "Directory");

        // Construct the map class for detailed pool information output

        pData->fPoolInfoClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/PoolInfo");

        pData->fPoolInfoClass->addKey("description",  "Description");
        pData->fPoolInfoClass->addKey("requestList",  "Pending Requests");
        pData->fPoolInfoClass->addKey("resourceList", "Resources");

        // Construct map class for a pending request

        pData->fRequestClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/Request");

        pData->fRequestClass->addKey("machine",    "Machine");
        pData->fRequestClass->addKey("handleName", "Handle Name");
        pData->fRequestClass->addKey("handle",     "Handle");
        pData->fRequestClass->addKey("user",       "User");
        pData->fRequestClass->addKey("endpoint",   "Endpoint");
        pData->fRequestClass->addKey("requestedTimestamp",
                                     "Date-Time Requested");

        // Construct map class for a resource owner

        pData->fResourceOwnerClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/ResourceOwner");

        pData->fResourceOwnerClass->addKey("machine",    "Machine");
        pData->fResourceOwnerClass->addKey("handleName", "Handle Name");
        pData->fResourceOwnerClass->addKey("handle",     "Handle");
        pData->fResourceOwnerClass->addKey("user",       "User");
        pData->fResourceOwnerClass->addKey("endpoint",   "Endpoint");
        pData->fResourceOwnerClass->addKey("requestedTimestamp",
                                           "Date-Time Requested");
        pData->fResourceOwnerClass->addKey("acquiredTimestamp",
                                           "Date-Time Acquired");
        
        // Construct map class for a resource

        pData->fResourceClass = STAFMapClassDefinition::create(
            "STAF/Service/ResPool/Resource");

        pData->fResourceClass->addKey("entry", "Entry");
        pData->fResourceClass->addKey("owner", "Owner");

        // Get line separator

        STAFResultPtr result = pData->fHandlePtr->submit(
            "local", "VAR", "RESOLVE STRING {STAF/Config/Sep/Line}");

        if (result->rc != 0)
        {
            *pErrorBuffer = result->result.adoptImpl();
            return result->rc;
        }
        else sLineSep = result->result;

        // Get local machine name (logical identifier)

        result = pData->fHandlePtr->submit(
            "local", "VAR", "RESOLVE STRING {STAF/Config/Machine}");

        if (result->rc != 0)
        {
            *pErrorBuffer = result->result.adoptImpl();
            return result->rc;
        }
        else pData->fLocalMachineName = result->result;

        // Register Help Data
        registerHelpData(pData, kSTAFResPoolNotEntryOwner,
            STAFString("Not resource pool entry owner"),
            STAFString("You are not the owner of the entry you are trying to "
                       "RELEASE.  Use the FORCE option if you are sure that "
                       "the correct entry is specified."));

        registerHelpData(pData, kSTAFResPoolHasPendingRequests,
            STAFString("Resource pool has pending requests"),
            STAFString("The resource pool you are trying to DELETE has pending"
                       " requests.  If necessary, use the FORCE option."));

        registerHelpData(pData, kSTAFResPoolNoEntriesAvailable,
            STAFString("Resource pool has no entries available"),
            STAFString("The resource pool has no entries."));

        registerHelpData(pData, kSTAFResPoolCreatePoolPathError,
            STAFString("Error creating pool path"),
            STAFString("The directory specified by the DIRECTORY parameter "
                       "when registering the service or the default "
                       "directory could not be created."));

        registerHelpData(pData, kSTAFResPoolInvalidFileFormat,
            STAFString("Invalid pool file format"),
            STAFString("An error occurred reading the resource pool file "
                       "due to an error in the file format.  If you are "
                       "using the latest version of the Resource Pool "
                       "service, contact the STAF authors."));

        registerHelpData(pData, kSTAFResPoolEntryIsOwned,
            STAFString("Resource pool entry is owned"),
            STAFString("A resource pool entry you specified to REMOVE is "
                       "owned.  Use the FORCE option if you are sure that "
                       "the correct entry is specified."));

        // Determine the pool directory

        STAFFSPath poolFilePath;

        if (pData->fPoolDir != STAFString())
        {
            // Set to pool directory using DIRECTORY parameter from RESPOOL
            // Service Configuration

            poolFilePath.setRoot(pData->fPoolDir);
        }
        else
        {   // Assign Default Pool Directory value if no DIRECTORY parameter
            // -  Use <pInfo->writeLocation>/<lower-case service name>
            
            poolFilePath.setRoot(pInfo->writeLocation);
            poolFilePath.addDir("service");
            poolFilePath.addDir(pData->fShortName.toLowerCase());
            pData->fPoolDir = poolFilePath.asString();
        }

        // Find all the pools (*.rpl files) in the pool directory
        
        STAFFSEntryType_t entryType = kSTAFFSFile;
        
        if (poolFilePath.exists())
        {
            STAFFSEnumPtr dirEnum = 
                poolFilePath.getEntry()->enumerate(kUTF8_STAR, sPoolExt, 
                                                   entryType);
            // Initialize each pool

            for (; dirEnum->isValid(); dirEnum->next())
            {
                STAFFSEntryPtr entry = dirEnum->entry();

                STAFString fileName = entry->path().asString();

                // Read the pool file and store its data in PoolData

                PoolData poolData;
                
                unsigned int status = readPoolFile(fileName, poolData);

                if (status == kReadEndOfFile)
                {
                    STAFString error( 
                        "STAFResPoolService.cpp: STAFServiceInit: "
                        "Invalid file contents in resource pool " + fileName);
                    cout << error << endl;
                    *pErrorBuffer = error.adoptImpl();
                    return kSTAFResPoolInvalidFileFormat;
                }
                else if (status == kReadInvalidFormat)
                {
                    STAFString error =
                       "STAFResPoolService.cpp: STAFServiceInit: " 
                       "Invalid file format (" + poolData.fileFormat;
                    error += ") in resource pool file " + fileName;
                    cout << error << endl;
                    *pErrorBuffer = error.adoptImpl();
                    return kSTAFResPoolInvalidFileFormat;
                }
                else if (status == kFileOpenError)
                {
                    STAFString error(
                       "STAFResPoolService.cpp: STAFServiceInit: "
                       "Error opening resource pool file " + fileName);
                    cout << error << endl;
                    *pErrorBuffer = error.adoptImpl();
                    return kSTAFFileOpenError;
                }

                // Add the pool data to the Pool Map

                pData->fPoolMap.insert(PoolMap::value_type(
                    poolData.poolName.toUpperCase(),
                    PoolDataPtr(new PoolData(poolData), PoolDataPtr::INIT)));
            }
        }
        else
        {
            // Create the pool directory
            try
            {
                STAFFSEntryPtr pooldir = 
                    poolFilePath.createDirectory(kSTAFFSCreatePath);
            }
            catch (...)
            { 
                STAFString error("STAFResPoolService.cpp: STAFServiceInit: "
                                 "Invalid ResPool Directory: " +
                                 pData->fPoolDir);
                cout << error << endl;
                *pErrorBuffer = error.adoptImpl();
                return kSTAFResPoolCreatePoolPathError;
            }
        }

        pData->fPoolMapRWSem = STAFRWSemPtr(new STAFRWSem, STAFRWSemPtr::INIT);
    }

    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                        "STAFResPoolService.cpp: STAFServiceInit").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFResPoolService.cpp: STAFServiceInit: Caught unknown "
                    "exception in STAFServiceServiceInit()");
        *pErrorBuffer = error.adoptImpl();
    }    

    return retCode;
}


STAFRC_t STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                  void *pRequestInfo, unsigned int reqLevel,
                                  STAFString_t *pResultBuffer)
{
    if (reqLevel != 30) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        STAFResultPtr result(new STAFResult(kSTAFOk, STAFString()),
                             STAFResultPtr::INIT);        

        STAFServiceRequestLevel30 *pInfo =
            reinterpret_cast<STAFServiceRequestLevel30 *>(pRequestInfo);

        ResPoolServiceData *pData =
            reinterpret_cast<ResPoolServiceData *>(serviceHandle);

        STAFString request(pInfo->request);
        STAFString action = request.subWord(0, 1).toLowerCase();
        
        // Call functions for the request

        if (action == "create")
            result = handleCreate(pInfo, pData);
        else if (action == "delete")
            result = handleDelete(pInfo, pData);
        else if (action == "query")
            result = handleQuery(pInfo, pData);
        else if (action == "request")
            result = handleRequest(pInfo, pData);
        else if (action == "add")
            result = handleAdd(pInfo, pData);
        else if (action == "remove")
            result = handleRemove(pInfo, pData);
        else if (action == "release")
            result = handleRelease(pInfo, pData);
        else if (action == "list")
            result = handleList(pInfo, pData);
        else if (action == "staf_callback") 
            result = handleSTAFCallback(pInfo, pData);
        else if (action == "help")
            result = handleHelp(pInfo, pData);
        else if (action == "version")
            result = handleVersion(pInfo, pData);
        else
        {
            result = STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                                   request.subWord(0, 1)), 
                                   STAFResultPtr::INIT);
        }

        *pResultBuffer = result->result.adoptImpl();
        retCode = result->rc;
    }
    catch (STAFException &e)
    { 
        *pResultBuffer = getExceptionString(e,
                         "STAFResPoolService.cpp: "
                         "STAFServiceAcceptRequest").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFResPoolService.cpp: STAFServiceAcceptRequest: Caught "
                    "unknown exception in STAFServiceAcceptRequest()");
        *pResultBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                         void *pTermInfo, unsigned int termLevel,
                         STAFString_t *pErrorBuffer)
{
    if (termLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        retCode = kSTAFOk;
        
        ResPoolServiceData *pData =
            reinterpret_cast<ResPoolServiceData *>(serviceHandle);

        // Un-register Help Data

        unregisterHelpData(pData, kSTAFResPoolNotEntryOwner);
        unregisterHelpData(pData, kSTAFResPoolHasPendingRequests);
        unregisterHelpData(pData, kSTAFResPoolNoEntriesAvailable);
        unregisterHelpData(pData, kSTAFResPoolCreatePoolPathError);
        unregisterHelpData(pData, kSTAFResPoolInvalidFileFormat);
        unregisterHelpData(pData, kSTAFResPoolEntryIsOwned);
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                        "STAFResPoolService.cpp: STAFServiceTerm").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFResPoolService.cpp: STAFServiceTerm: Caught unknown "
                    "exception in STAFServiceTerm()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceDestruct(STAFServiceHandle_t *serviceHandle,
                             void *pDestructInfo, unsigned int destructLevel,
                             STAFString_t *pErrorBuffer)
{
    if (destructLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        ResPoolServiceData *pData =
            reinterpret_cast<ResPoolServiceData *>(*serviceHandle);

        delete pData;
        *serviceHandle = 0;

        retCode = kSTAFOk;
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                    "STAFResPoolService.cpp: STAFServiceDestruct").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFResPoolService.cpp: STAFServiceDestruct: Caught "
                    "unknown exception in STAFServiceDestruct()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


// Handles resource pool creation requests

STAFResultPtr handleCreate(STAFServiceRequestLevel30 *pInfo, 
                           ResPoolServiceData *pData)
{
    // Verify the requester has at least trust level 4

    VALIDATE_TRUST(4, pData->fShortName, "CREATE", pData->fLocalMachineName);

    // Parse the result

    STAFCommandParseResultPtr parsedResult =
        pData->fCreateParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }

    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Set the poolDescription variable (resolve the pool description)

    resultPtr = resolveOp(pInfo, pData, parsedResult, sDescription);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolDescription = resultPtr->result;
 
    // Get a write lock on the Pool Map for the duration of this block

    STAFRWSemWLock wLock(*pData->fPoolMapRWSem);

    // Verify that the resource pool does not already exist in the Pool Map
    
    if ((pData->fPoolMap.find(poolName.toUpperCase())) != pData->fPoolMap.end())
    {
        return STAFResultPtr(new STAFResult(kSTAFAlreadyExists, poolName),
                             STAFResultPtr::INIT);
    }

    // Set the path for the resource pool file

    STAFFSPath poolFilePath;
    poolFilePath.setRoot(pData->fPoolDir);
    poolFilePath.setName(poolName);
    poolFilePath.setExtension(sPoolExt);

    PoolData poolData(poolName, poolDescription);

    // Write the pool data
    
    STAFString fileName = poolFilePath.asString();
    
    if (writePoolFile(fileName, poolData) != kReadorWriteOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFFileWriteError, fileName),
                             STAFResultPtr::INIT);
    }

    // Update in memory data structures

    pData->fPoolMap.insert(PoolMap::value_type(poolData.poolName.toUpperCase(),
                    PoolDataPtr(new PoolData(poolData), PoolDataPtr::INIT)));
    
    // Return an OK result

    return STAFResultPtr(new STAFResult(kSTAFOk, STAFString()),
                         STAFResultPtr::INIT);
}


// Handles resource pool deletion requests

STAFResultPtr handleDelete(STAFServiceRequestLevel30 *pInfo, 
                           ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 4

    VALIDATE_TRUST(4, pData->fShortName, "DELETE", pData->fLocalMachineName);
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = pData->fDeleteParser->parse(
        pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;
    
    // Get a write lock on the Pool Map for the duration of this block.
    // Don't need a lock on Pool Data because I have a write lock on Pool Map.

    STAFRWSemWLock wLock(*pData->fPoolMapRWSem);

    // Make sure the resource pool is in the Pool Map and get a pointer to it

    PoolMap::iterator poolIterator;
    PoolDataPtr poolPtr;
    
    if ((poolIterator = pData->fPoolMap.find(poolName.toUpperCase())) == 
        pData->fPoolMap.end())
    {
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                             STAFResultPtr::INIT);
    }
    
    // Set a pointer to the resource pool to be deleted

    poolPtr = (*poolIterator).second;

    // Check if there are any pending requests and if FORCE is not specified.
    
    if ((poolPtr->requestList.size() != 0) && 
        (parsedResult->optionTimes(sForce) == 0))
    {
        return STAFResultPtr(new STAFResult(kSTAFResPoolHasPendingRequests,
                             poolName), STAFResultPtr::INIT);
    }
    
    // Now delete the resource pool file
    
    STAFFSPath poolFilePath;
    poolFilePath.setRoot(pData->fPoolDir);
    poolFilePath.setName(poolName);
    poolFilePath.setExtension(sPoolExt);
    
    try
    {
        poolFilePath.getEntry()->remove();
    }
    catch (STAFException &e)
    { 
        result = getExceptionString(e, 
                 "STAFResPoolSerice.cpp: Delete ").adoptImpl();
        return STAFResultPtr(new STAFResult(kSTAFFileDeleteError,
                             result), STAFResultPtr::INIT);                
    }

    // Remove the resource pool the Pool Map

    pData->fPoolMap.erase(poolIterator);

    // If there are any pending requests for this pool, wakeup each pending
    // requester with no resource specified.
  
    if (poolPtr->requestList.size() > 0)
    {
        for (RequestList::iterator iter = poolPtr->requestList.begin();
             iter != poolPtr->requestList.end(); ++iter)
        {
            (*iter)->retCode = kSTAFDoesNotExist;
            (*iter)->resultBuffer = poolName;
            (*iter)->wakeup->post();
        }
    }

    // Return an Ok result

    return STAFResultPtr(new STAFResult(kSTAFOk, STAFString()),
                         STAFResultPtr::INIT);
}


// Handles resource pool add entry requests

STAFResultPtr handleAdd(STAFServiceRequestLevel30 *pInfo, 
                        ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 4

    VALIDATE_TRUST(4, pData->fShortName, "ADD", pData->fLocalMachineName);
    
    // Parse the result

    STAFCommandParseResultPtr parsedResult = 
        pData->fAddParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Get a write lock on the Pool Map for the duration of this block
    
    STAFRWSemWLock wLock(*pData->fPoolMapRWSem);

    // Make sure the resource pool is in the Pool Map and get a pointer to it

    PoolMap::iterator poolIterator;
    
    poolIterator = pData->fPoolMap.find(poolName.toUpperCase());

    if (poolIterator == pData->fPoolMap.end())
    {
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                             STAFResultPtr::INIT);
    }
        
    // Create a copy of the pool to update

    PoolDataPtr poolPtr = (*poolIterator).second;
    PoolData newPool = *(poolPtr);

    // Get each entry to be added and check if entry is already in the pool.       
    // If there are any duplicate entries to be added, return an error and
    // do not update the resource pool.

    unsigned int numEntriesToAdd = parsedResult->optionTimes(sEntry);
    unsigned int i;

    for (i = 1; i <= numEntriesToAdd; i++)
    {
        STAFString thisEntry = parsedResult->optionValue(sEntry, i);

        for (unsigned int j = 0; j < newPool.resourceList.size(); j++)
        {
            if (newPool.resourceList[j].entry == thisEntry)
            {
                 return STAFResultPtr(new STAFResult(kSTAFAlreadyExists,
                                                     thisEntry),
                                      STAFResultPtr::INIT);
            }
        }

        // Add to the end of the new pool's resource list if not a duplicate

        ResourceData resourceData(thisEntry);
        newPool.resourceList.push_back(resourceData);
        newPool.numResources++;
    }

    // Delete the old pool file and write the new pool file

    STAFFSPath poolFilePath;
    poolFilePath.setRoot(pData->fPoolDir);
    poolFilePath.setName(poolName);
    poolFilePath.setExtension(sPoolExt);

    try
    {
        poolFilePath.getEntry()->remove();
    }
    catch (STAFException &e)
    {
        result = getExceptionString(e,
                 "STAFResPoolService.cpp: Add ").adoptImpl();
        return STAFResultPtr(new STAFResult(kSTAFFileDeleteError, result),
                             STAFResultPtr::INIT);                
    }

    STAFString fileName = poolFilePath.asString();

    if (writePoolFile(fileName, newPool) != kReadorWriteOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFFileWriteError, fileName),
                             STAFResultPtr::INIT);
    }
    
    // Update the pool in memory

    poolPtr->resourceList = newPool.resourceList;
    poolPtr->numResources = newPool.numResources;
        
    // For each resource entry added, if there are any pending requests,
    // tell the next requester the resource he can have, wake the requester
    // up, and remove the satisfied pending request.

    for (i = poolPtr->numResources - numEntriesToAdd;
         i < poolPtr->numResources; i++)
    {
        if (poolPtr->requestList.size() > 0)
        {
            // Get the first request from the pending request list

            RequestDataPtr requestDataPtr = poolPtr->requestList.front();
                
            // Assign a resource to the request

            requestDataPtr->retCode = kSTAFOk;
            requestDataPtr->resultBuffer = poolPtr->resourceList[i].entry;

            // Update the resource entry's ownership information

            poolPtr->resourceList[i].owned = 1;
            poolPtr->usedResources++;
            poolPtr->resourceList[i].orgUUID = requestDataPtr->orgUUID;
            poolPtr->resourceList[i].orgMachine = requestDataPtr->orgMachine;
            poolPtr->resourceList[i].orgName = requestDataPtr->orgName;
            poolPtr->resourceList[i].orgHandle = requestDataPtr->orgHandle;
            poolPtr->resourceList[i].orgUser = requestDataPtr->orgUser;
            poolPtr->resourceList[i].orgEndpoint = requestDataPtr->orgEndpoint;
            poolPtr->resourceList[i].requestedTime = 
                requestDataPtr->requestedTime;
            poolPtr->resourceList[i].acquiredTime = 
                STAFTimestamp::now().asString();

            // Wakeup the requester

            requestDataPtr->wakeup->post();
            
            // Remove the request from the pending request list

            poolPtr->requestList.pop_front();
        }
        else
        {   // No more pending requests
            break;
        }
    }

    // Return an Ok result
    
    return STAFResultPtr(new STAFResult(kSTAFOk, result), STAFResultPtr::INIT);
}


// Handles resource pool remove entry requests

STAFResultPtr handleRemove(STAFServiceRequestLevel30 *pInfo, 
                           ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 4

    VALIDATE_TRUST(4, pData->fShortName, "REMOVE", pData->fLocalMachineName);
    
    // Parse the result

    STAFCommandParseResultPtr parsedResult = 
        pData->fRemoveParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Get a write lock on the Pool Map for the duration of this block

    STAFRWSemWLock wLock(*pData->fPoolMapRWSem);

    // Make sure the specified resource pool exists in Pool Map

    PoolMap::iterator poolIterator;
    
    poolIterator = pData->fPoolMap.find(poolName.toUpperCase());

    if (poolIterator == pData->fPoolMap.end())
    {  
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                             STAFResultPtr::INIT);
    }

    PoolDataPtr poolPtr = (*poolIterator).second;
    
    if (poolPtr->resourceList.size() == 0) 
    {
        return STAFResultPtr(new STAFResult(kSTAFResPoolNoEntriesAvailable,
                             poolName), STAFResultPtr::INIT);
    }

    unsigned int numEntriesToRemove = parsedResult->optionTimes(sEntry);

    // Create a copy of the pool to update

    PoolData newPool = *(poolPtr);

    // Get each of the entries to be removed; If an invalid entry is found,
    //  return with an appropriate error message and do not remove any other
    //  valid entries specified on the request.
   
    for (unsigned int i = 1; i <= numEntriesToRemove; i++)
    {
        STAFString thisEntry = parsedResult->optionValue(sEntry, i);
        int resid = -1;

        // Check if this entry is in the pool - set resid to its index

        for (unsigned int j = 0; 
             j < newPool.resourceList.size() && resid == -1; j++)
        {
            if (thisEntry == newPool.resourceList[j].entry) resid = j;
        }

        // Remove the entry if found and if owned with FORCE specified

        if ((resid != -1) &&
            (!newPool.resourceList[resid].owned || 
             (parsedResult->optionTimes(sForce) != 0)))
        {
            newPool.resourceList.erase(newPool.resourceList.begin() + resid);

            newPool.numResources--;

            if (newPool.resourceList[resid].owned)
            {
                newPool.usedResources--;
            }
        }
        else if (resid == -1) 
        {
            return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, thisEntry),
                                 STAFResultPtr::INIT);
        }
        else
        {   //  Cannot remove pool entry if owned and FORCE not specified

            return STAFResultPtr(new STAFResult(
                kSTAFResPoolEntryIsOwned, thisEntry), STAFResultPtr::INIT);
        }
    } // end for each entry specified

    // Delete the old pool file and write the new pool file

    STAFFSPath poolFilePath;
    poolFilePath.setRoot(pData->fPoolDir);
    poolFilePath.setName(poolName);
    poolFilePath.setExtension(sPoolExt);

    try
    {
        poolFilePath.getEntry()->remove();
    }
    catch (STAFException &e)
    { 
        result = getExceptionString(e,
                             "STAFResPoolService.cpp: Remove ").adoptImpl();
        return STAFResultPtr(new STAFResult(kSTAFFileDeleteError, result),
                             STAFResultPtr::INIT);                
    }

    STAFString fileName = poolFilePath.asString();

    if (writePoolFile(fileName, newPool) != kReadorWriteOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFFileWriteError, fileName),
                             STAFResultPtr::INIT);
    }

    // Update the pool in memory

    poolPtr->resourceList = newPool.resourceList;
    poolPtr->numResources = newPool.numResources;
    poolPtr->usedResources = newPool.usedResources;

    // Result an Ok result

    return STAFResultPtr(new STAFResult(kSTAFOk, result), STAFResultPtr::INIT);
}


// Handles resource pool request requests

STAFResultPtr handleRequest(STAFServiceRequestLevel30 *pInfo, 
                            ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 3

    VALIDATE_TRUST(3, pData->fShortName, "REQUEST", pData->fLocalMachineName);
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = pData->fRequestParser->parse(
        pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Set the FIRST variable (if first = 0, a random request is performed)

    unsigned int first = parsedResult->optionTimes(sFirst);
    
    // Set the timeout variable (resolve the timeout value and check if numeric)

    unsigned int timeout = STAF_EVENT_SEM_INDEFINITE_WAIT;

    if (parsedResult->optionTimes(sTimeout) > 0)
    {
        resultPtr = resolveOp(pInfo, pData, parsedResult, sTimeout);

        if (resultPtr->rc != 0) return resultPtr;

        if (!resultPtr->result.isDigits())
        {
            return STAFResultPtr(new STAFResult(kSTAFInvalidValue, 
                                 resultPtr->result), STAFResultPtr::INIT);
        }

        timeout = (resultPtr->result).asUInt();
    }

    RequestDataPtr requestDataPtr;
    PoolMap::iterator poolIterator;
    PoolDataPtr poolPtr;
    
    // Get a read lock on the Pool Map for the duration of this block
    {
        STAFRWSemRLock rLock(*pData->fPoolMapRWSem);

        // Make sure the specified resource pool is in the PoolMap

        poolIterator = pData->fPoolMap.find(poolName.toUpperCase());

        if (poolIterator == pData->fPoolMap.end())
        {  
            return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                                 STAFResultPtr::INIT);
        }
    
        poolPtr = (*poolIterator).second;
    
        // Lock the poolData semaphore for the duration of this block
        {
            STAFMutexSemLock lock(*poolPtr->accessSem);

            if (poolPtr->resourceList.size() == 0)
            {
                return STAFResultPtr(new STAFResult(
                    kSTAFResPoolNoEntriesAvailable, poolName),
                    STAFResultPtr::INIT);
            }
            
            if (debug) // XXX
                cout << "Got a respool request for " << poolName << "\n";
            
            // Register for notification when the handle ends
            
            STAFString request = "STAF_NOTIFY REGISTER ONENDOFHANDLE ";
            request += pInfo->handle;
            request += " MACHINE ";
            request += pInfo->machine;
            request += " UUID ";
            request += pInfo->stafInstanceUUID;
            request += " SERVICE ";
            request += pData->fShortName;
            request += " KEY ";
            request += "ResPoolEntry";
                        
            STAFResultPtr notifyResult = pData->fHandlePtr->submit("LOCAL", "HANDLE",
                 request);
            
            // Check if there are any available resources

            int availResources = poolPtr->numResources - poolPtr->usedResources;

            if (availResources > 0)
            {
                // Get the index of a random resource

                int rIndex = rand() % availResources;

                // Find the first or Nth available resource in the resource list
       
                unsigned int i;  // Index of the available index

                for (i = 0; i < poolPtr->resourceList.size(); i++)
                {
                    if ((!poolPtr->resourceList[i].owned) &&
                        (first || !rIndex--)) 
                        break;
                }

                // Mark the resource as OWNED and return the entry info

                poolPtr->usedResources++;
                poolPtr->resourceList[i].owned = 1;
                poolPtr->resourceList[i].orgUUID = pInfo->stafInstanceUUID;
                poolPtr->resourceList[i].orgMachine = pInfo->machine;
                poolPtr->resourceList[i].orgName = pInfo->handleName;
                poolPtr->resourceList[i].orgHandle = pInfo->handle;
                poolPtr->resourceList[i].orgUser = pInfo->user;
                poolPtr->resourceList[i].orgEndpoint = pInfo->endpoint;
                STAFString currentTime = STAFTimestamp::now().asString();
                poolPtr->resourceList[i].requestedTime = currentTime;
                poolPtr->resourceList[i].acquiredTime = currentTime;

                // Return the entry assigned to the request

                return STAFResultPtr(new STAFResult(kSTAFOk, 
                                     poolPtr->resourceList[i].entry),
                                     STAFResultPtr::INIT);
            }

            // Else no resources currently available; put me on the pending list
    
            RequestData requestData(pInfo->stafInstanceUUID, pInfo->machine,
                                    pInfo->handleName, pInfo->handle,
                                    pInfo->user, pInfo->endpoint);

            requestDataPtr = RequestDataPtr(new RequestData(requestData),
                                            RequestDataPtr::INIT);

            poolPtr->requestList.push_back(requestDataPtr);

        } // End block for locking the PoolData access semaphore

    } // End block for putting a read lock on the PoolMap

    // Wait for the specified time for a resource to become available

    requestDataPtr->wakeup->wait(timeout);
            
    // Check if the handle that submitted this request was garbage
    // collected while we were waiting for a resource to become
    // available, and if so, return an error.
    // Save the cost of getting a semaphore lock by first checking if
    // the request was garbage collected

    if (*requestDataPtr->garbageCollectedPtr)
    {
        // Lock the poolData semaphore for the duration of this block

        STAFMutexSemLock lock(*poolPtr->accessSem);

        if (*requestDataPtr->garbageCollectedPtr)
        {
            return STAFResultPtr(
                new STAFResult(
                    kSTAFRequestCancelled,
                    "The handle that submitted this request no longer exists"),
                STAFResultPtr::INIT);
        }
    }

    // If request timed out, remove the request from the list.  Save the
    // cost of getting a semaphore lock by first checking if the request
    // timed out.

    if (requestDataPtr->retCode == kSTAFTimeout)
    { 
        // Lock the poolData semaphore for the duration of this block

        STAFMutexSemLock lock(*poolPtr->accessSem);

        if (requestDataPtr->retCode == kSTAFTimeout)
        {
            poolPtr->requestList.remove(requestDataPtr);
        }
    }

    // Return the return code and result assigned to the request

    return STAFResultPtr(new STAFResult(requestDataPtr->retCode,
                         requestDataPtr->resultBuffer), STAFResultPtr::INIT);
}                    


// Handles resource pool release entry requests

STAFResultPtr handleRelease(STAFServiceRequestLevel30 *pInfo, 
                            ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 3

    VALIDATE_TRUST(3, pData->fShortName, "RELEASE", pData->fLocalMachineName);
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = 
        pData->fReleaseParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Set the releaseEntry variable (resolve the entry value)

    resultPtr = resolveOp(pInfo, pData, parsedResult, sEntry);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString releaseEntry = resultPtr->result;
   
    // Get a read lock on the Pool Map for the duration of this block

    STAFRWSemRLock rLock(*pData->fPoolMapRWSem);

    // Make sure that the resource pool is in pData->poolMap

    PoolMap::iterator poolIterator;
    PoolDataPtr poolPtr;

    poolIterator = pData->fPoolMap.find(poolName.toUpperCase());

    if (poolIterator == pData->fPoolMap.end())
    {  
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                             STAFResultPtr::INIT);
    }
    
    poolPtr = (*poolIterator).second;
    
    // Lock the poolData semaphore for the duration of this block
    
    STAFMutexSemLock lock(*poolPtr->accessSem);
    
    // Find the entry in the resource pool

    int resid = -1;

    for (unsigned int i = 0; i < poolPtr->resourceList.size(); i++)
    {
        if (releaseEntry == poolPtr->resourceList[i].entry)
        {
            resid = i;
            break;
        }
    }

    // If the entry is not in the resource pool, return an error

    if (resid == -1)
    { 
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, releaseEntry),
                             STAFResultPtr::INIT);
    }

    // Check if the entry requested to be released is owned

    if (poolPtr->resourceList[resid].owned)
    {
        // Check if you are the owner

        unsigned int owner = 0;   // 0 means not the owner

        if ((poolPtr->resourceList[resid].orgUUID == pInfo->stafInstanceUUID) &&
            (poolPtr->resourceList[resid].orgHandle == pInfo->handle))
        {
            owner = 1;            // 1 means you are the owner
        }

        // If you are not the owner and FORCE is specified, need trust level 4

        unsigned int force = parsedResult->optionTimes(sForce);
        
        if (!owner && force)
        {
            // Verify the requester has at least trust level 4

            VALIDATE_TRUST2(4, pData->fShortName, "RELEASE FORCE",
                            pData->fLocalMachineName);
        }

        // To release, you must be the owner or FORCE must be specified

        if (owner || force)
        {
            // Mark the resource as available

            poolPtr->usedResources--;
            poolPtr->resourceList[resid].owned = 0;

            // If there are any pending requests for this pool, wake up the 
            // first requester and tell him the resource he can have.

            if (poolPtr->requestList.size() > 0)
            {
                RequestDataPtr requestDataPtr = poolPtr->requestList.front();

                // Assign the resource to the request

                requestDataPtr->retCode = kSTAFOk;
                requestDataPtr->resultBuffer = 
                    poolPtr->resourceList[resid].entry;

                // Update the resource entry's ownership information

                poolPtr->resourceList[resid].owned = 1;
                poolPtr->usedResources++;
                poolPtr->resourceList[resid].orgUUID =  requestDataPtr->orgUUID;
                poolPtr->resourceList[resid].orgMachine = 
                    requestDataPtr->orgMachine;
                poolPtr->resourceList[resid].orgName = requestDataPtr->orgName;
                poolPtr->resourceList[resid].orgHandle = 
                    requestDataPtr->orgHandle;
                poolPtr->resourceList[resid].orgUser = requestDataPtr->orgUser;
                poolPtr->resourceList[resid].orgEndpoint =
                    requestDataPtr->orgEndpoint;
                poolPtr->resourceList[resid].requestedTime = 
                    requestDataPtr->requestedTime;
                poolPtr->resourceList[resid].acquiredTime = 
                    STAFTimestamp::now().asString();
                
                // Wakeup the requester

                requestDataPtr->wakeup->post();

                // Remove the satisfied request from the pending request list

                poolPtr->requestList.pop_front();
            }
        }
        else
        {
            return STAFResultPtr(new STAFResult(kSTAFResPoolNotEntryOwner, 
                                 releaseEntry), STAFResultPtr::INIT);
        }
    } // end if entry was owned

    // Return an Ok result

    return STAFResultPtr(new STAFResult(kSTAFOk, result), 
                         STAFResultPtr::INIT);
}


STAFResultPtr handleList(STAFServiceRequestLevel30 *pInfo, 
                         ResPoolServiceData *pData)
{    
    // Verify the requester has at least trust level 2

    VALIDATE_TRUST(2, pData->fShortName, "LIST", pData->fLocalMachineName);

    // Parse the request

    STAFCommandParseResultPtr parsedResult = pData->fListParser->parse(
        pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    STAFObjectPtr mc = STAFObject::createMarshallingContext();

    if (parsedResult->optionTimes("SETTINGS"))
    {
        // LIST SETTINGS

        // Create a marshalled map containing settings for the service
        
        mc->setMapClassDefinition(pData->fSettingsClass->reference());

        // Write the service settings to a map

        STAFObjectPtr outputMap = pData->fSettingsClass->createInstance();
        outputMap->put("directory", pData->fPoolDir);

        mc->setRootObject(outputMap);
    }
    else
    {
        // LIST POOLS (This is the default if only LIST is specified)

        // Create a marshalled list of maps containing pool information

        mc->setMapClassDefinition(pData->fPoolClass->reference());
        STAFObjectPtr outputList = STAFObject::createList();

        // Get a read lock on the Pool Map for the duration of this block

        STAFRWSemRLock rLock(*pData->fPoolMapRWSem);

        // Write all the pool names to the result variable

        PoolMap::iterator poolIterator;

        for(poolIterator = pData->fPoolMap.begin(); 
            poolIterator != pData->fPoolMap.end(); ++poolIterator)
        {
            STAFObjectPtr poolInfoMap = pData->fPoolClass->createInstance();
            poolInfoMap->put("poolName", poolIterator->second->poolName);
            poolInfoMap->put("description", poolIterator->second->poolDescription);
            outputList->append(poolInfoMap);
        }           

        mc->setRootObject(outputList);
    }

    // Return an Ok result and the marshalled output for a list of the pools

    return STAFResultPtr(new STAFResult(kSTAFOk, mc->marshall()),
                         STAFResultPtr::INIT);
}


// Handles resource pool query requests

STAFResultPtr handleQuery(STAFServiceRequestLevel30 *pInfo, 
                          ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Verify the requester has at least trust level 2

    VALIDATE_TRUST(2, pData->fShortName, "QUERY", pData->fLocalMachineName);
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = pData->fQueryParser->parse(
        pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    // Set the poolName variable (resolve the pool name)

    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, sPool);

    if (resultPtr->rc != 0) return resultPtr;

    STAFString poolName = resultPtr->result;

    // Create a marshalled map of general pool information

    STAFObjectPtr mc = STAFObject::createMarshallingContext();
    mc->setMapClassDefinition(pData->fPoolInfoClass->reference());
    mc->setMapClassDefinition(pData->fRequestClass->reference());
    mc->setMapClassDefinition(pData->fResourceClass->reference());
    mc->setMapClassDefinition(pData->fResourceOwnerClass->reference());

    STAFObjectPtr poolInfoMap = pData->fPoolInfoClass->createInstance();

    // Get a read lock on the Pool Map for the duration of this block

    STAFRWSemRLock rLock(*pData->fPoolMapRWSem);

    // Make sure the resource pool is in the Pool Map and get a pointer to it

    PoolMap::iterator poolIterator;
    PoolDataPtr poolPtr;

    poolIterator = pData->fPoolMap.find(poolName.toUpperCase());

    if (poolIterator == pData->fPoolMap.end())
    {  
        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, poolName),
                             STAFResultPtr::INIT);
    }

    // Set a pointer to the resource pool being queried

    poolPtr = (*poolIterator).second;
    
    poolInfoMap->put("description", poolPtr->poolDescription);

    // Create an empty list object for marshalled list of pending requests

    STAFObjectPtr requestList = STAFObject::createList();
    
    // Append an entry to requestList for each pending request

    if (poolPtr->requestList.size() > 0)
    {
        RequestList::iterator iList;
        unsigned int requestNum = 1;

        for (iList = poolPtr->requestList.begin(); 
             iList != poolPtr->requestList.end(); ++iList)
        {
            // Add an entry in a marshalling context as a map to requestList

            STAFObjectPtr requestMap = pData->fRequestClass->createInstance();

            requestMap->put("machine", (*iList)->orgMachine);
            requestMap->put("handleName", (*iList)->orgName);
            requestMap->put("handle", STAFString((*iList)->orgHandle));
            requestMap->put("user", (*iList)->orgUser);
            requestMap->put("endpoint", (*iList)->orgEndpoint);
            requestMap->put("requestedTimestamp", (*iList)->requestedTime);

            requestList->append(requestMap);
        }
    }

    poolInfoMap->put("requestList", requestList);
      
    // Create an empty list object for the marshalled list of the pool's
    // resource entries

    STAFObjectPtr resourceList = STAFObject::createList();

    // Append an entry to resourceList for each resource

    for (unsigned int i = 0; i < poolPtr->resourceList.size(); i++)
    {
        // Create a map of information about each resource

        STAFObjectPtr resourceMap = pData->fResourceClass->createInstance();

        resourceMap->put("entry", poolPtr->resourceList[i].entry);

        if (!poolPtr->resourceList[i].owned)
        {
            resourceMap->put("owner", STAFObject::createNone());
        }
        else
        {
            // Create a map of information about the resource owner

            STAFObjectPtr resourceOwnerMap =
                pData->fResourceOwnerClass->createInstance();

            resourceOwnerMap->put("machine",
                                  poolPtr->resourceList[i].orgMachine);
            resourceOwnerMap->put("handleName",
                                  poolPtr->resourceList[i].orgName);
            resourceOwnerMap->put("handle", STAFString(
                                  poolPtr->resourceList[i].orgHandle));
            resourceOwnerMap->put("user", poolPtr->resourceList[i].orgUser);
            resourceOwnerMap->put("endpoint",
                                  poolPtr->resourceList[i].orgEndpoint);
            resourceOwnerMap->put("requestedTimestamp",
                                  poolPtr->resourceList[i].requestedTime);
            resourceOwnerMap->put("acquiredTimestamp",
                                  poolPtr->resourceList[i].acquiredTime);

            resourceMap->put("owner", resourceOwnerMap);
        }

        // Add the entry in a marshalling context as a map to resourceList

        resourceList->append(resourceMap);
    }
    
    poolInfoMap->put("resourceList", resourceList);

    // Set the marshalling context's root object

    mc->setRootObject(poolInfoMap);

    // Return the marshalled result of the query

    return STAFResultPtr(new STAFResult(kSTAFOk, mc->marshall()), 
                         STAFResultPtr::INIT);
}


STAFResultPtr handleSTAFCallback(STAFServiceRequestLevel30 *pInfo, 
                                 ResPoolServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;
    
    // Don't check the TRUST level, but make sure the requesting handle
    // is the STAFProc handle

    if (pInfo->handle != 1)
    {
        // XXX Return a non-zero RC?
        return STAFResultPtr(new STAFResult(kSTAFOk, result), 
                              STAFResultPtr::INIT);
    }
    
    STAFCommandParseResultPtr parsedResult = 
        pData->fSTAFCallbackParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
  
    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, "HANDLE");

    if (resultPtr->rc != 0) return resultPtr;

    STAFString handle = resultPtr->result;
    
    resultPtr = resolveOp(pInfo, pData, parsedResult, "MACHINE");

    if (resultPtr->rc != 0) return resultPtr;

    STAFString machine = resultPtr->result;
    
    resultPtr = resolveOp(pInfo, pData, parsedResult, "UUID");

    if (resultPtr->rc != 0) return resultPtr;

    STAFString uuid = resultPtr->result;
    
    resultPtr = resolveOp(pInfo, pData, parsedResult, "KEY");

    if (resultPtr->rc != 0) return resultPtr;

    STAFString key = resultPtr->result;
    
    if (debug) // XXX
        cout << "Respool handleSTAFCallback " << "handle " << handle 
            << " machine " << machine << " key " << key << "\n";

    // Get a read lock on the Pool Map for the duration of this block

    STAFRWSemRLock rLock(*pData->fPoolMapRWSem);

    PoolMap::iterator poolIterator;

    // For each pool, remove any pending requests that were requested by
    // this garbage collected handle and remove any resources that are owned
    // by this garbage collected handle

    for(poolIterator = pData->fPoolMap.begin();
        poolIterator != pData->fPoolMap.end(); ++poolIterator)
    {
        if (debug) // XXX
            cout << poolIterator->second->poolName << " - " 
                << poolIterator->second->poolDescription << sLineSep;
            
        // First, remove any pending requests for this pool that were
        // requested by this garbage collected handle that no longer exists 

        if (poolIterator->second->requestList.size() > 0)
        {
            bool removedRequest = true;

            while (removedRequest)
            {
                removedRequest = false;

                for (RequestList::iterator iter =
                     poolIterator->second->requestList.begin();
                     iter != poolIterator->second->requestList.end(); ++iter)
                {
                    // Check if this garbage collected handle was the requester

                    if (((*iter)->orgMachine == machine) &&
                        (STAFString((*iter)->orgHandle) == handle) &&
                        (STAFString((*iter)->orgUUID) == uuid))
                    {
                        // The requester for this entry in the request list
                        // no longer exists, so remove this pending request

                        poolIterator->second->requestList.remove(*iter);
                        
                        // Set the flag indicating that the pending request
                        // has been garbage collected and wake up the
                        // requester

                        *(*iter)->garbageCollectedPtr = true;
                        (*iter)->wakeup->post();
                        
                        // Break since can't continue iterating a list
                        // that has been changed

                        removedRequest = true;
                        break;
                    }
                }
            }
        }

        // Second, release any resources in this pool that are owned by this
        // garbage collected handle that no longer exist and then wake up any
        // pending requests

        for(int resid = 0; resid < poolIterator->second->numResources; 
            ++resid)
        {
            if ((poolIterator->second->resourceList.size() != 0) &&
                (poolIterator->second->resourceList[resid].owned))
            {
                // Check if this garbage collected handle is the owner of the
                // resource

                // XXX: This needs to be dealing with the UUID I think, but
                //      it seems we need to update the handle notification
                //      mechanism

                if ((poolIterator->second->resourceList[resid].orgMachine == 
                    machine) && (STAFString(
                    poolIterator->second->resourceList[resid].orgHandle) == 
                    handle) && (STAFString(
                    poolIterator->second->resourceList[resid].orgUUID) == 
                    uuid))
                {
                    // The owner of this resource no longer exists
                    
                    if (debug) // XXX
                        cout << "Garbage collected handle owns entry "
                             << poolIterator->second->resourceList[resid].entry
                             << " in pool " << poolIterator->second->poolName
                             << endl;
                        
                    // Mark the resource as available
    
                    poolIterator->second->usedResources--;
                    poolIterator->second->resourceList[resid].owned = 0;
    
                    // If there are any pending requests for this pool, 
                    // wake up the first requester and tell him the resource 
                    // he can have.

                    if (poolIterator->second->requestList.size() > 0)
                    {
                        RequestDataPtr requestDataPtr = 
                            poolIterator->second->requestList.front();
        
                        // Assign the resource to the request
        
                        requestDataPtr->retCode = kSTAFOk;
                        requestDataPtr->resultBuffer = 
                            poolIterator->second->resourceList[resid].entry;
        
                        // Update the resource entry's ownership information
        
                        poolIterator->second->resourceList[resid].owned = 1;
                        poolIterator->second->usedResources++;
                        poolIterator->second->resourceList[resid].orgUUID = 
                            requestDataPtr->orgUUID;
                        poolIterator->second->resourceList[resid].orgMachine = 
                            requestDataPtr->orgMachine;
                        poolIterator->second->resourceList[resid].orgName = 
                            requestDataPtr->orgName;
                        poolIterator->second->resourceList[resid].orgHandle = 
                            requestDataPtr->orgHandle;
                        poolIterator->second->resourceList[resid].requestedTime = 
                            requestDataPtr->requestedTime;
                        poolIterator->second->resourceList[resid].acquiredTime = 
                            STAFTimestamp::now().asString();
                        
                        // Wakeup the requester
        
                        requestDataPtr->wakeup->post();
        
                        // Remove the satisfied request from the pending 
                        // request list
        
                        poolIterator->second->requestList.pop_front();
                    }
                }
            }
        }
    }

    return STAFResultPtr(new STAFResult(kSTAFOk, result), STAFResultPtr::INIT);
}


STAFResultPtr handleHelp(STAFServiceRequestLevel30 *pInfo, 
                         ResPoolServiceData *pData)
{
    // Verify the requester has at least trust level 1

    VALIDATE_TRUST(1, pData->fShortName, "HELP", pData->fLocalMachineName);
    
    // Return help text

    STAFString help("RESPOOL Service Help" + sLineSep + sLineSep);        

    help += "CREATE  POOL <PoolName> DESCRIPTION <Pooltext>";
    help += sLineSep;
    help += "DELETE  POOL <PoolName> CONFIRM [FORCE]";
    help += sLineSep;
    help += "QUERY   POOL <PoolName>";
    help += sLineSep;
    help += "REQUEST POOL <PoolName> [TIMEOUT <Timeout>] [FIRST | RANDOM]";
    help += sLineSep;
    help += "ADD     POOL <PoolName> ENTRY <Value> [ENTRY <Value>]...";
    help += sLineSep;
    help += "REMOVE  POOL <PoolName> ENTRY <Value> [ENTRY <Value>]... CONFIRM [FORCE]";
    help += sLineSep;
    help += "RELEASE POOL <PoolName> ENTRY <Value> [FORCE]";
    help += sLineSep;
    help += "LIST    [POOLS | SETTINGS]" + sLineSep;
    help += "VERSION" + sLineSep;
    help += "HELP" + sLineSep;

    return STAFResultPtr(new STAFResult(kSTAFOk, help), STAFResultPtr::INIT);
}


STAFResultPtr handleVersion(STAFServiceRequestLevel30 *pInfo, 
                            ResPoolServiceData *pData)
{    
    // Verify the requester has at least trust level 1

    VALIDATE_TRUST(1, pData->fShortName, "VERSION", pData->fLocalMachineName);
    
    return STAFResultPtr(new STAFResult(kSTAFOk, sVersionInfo), 
                         STAFResultPtr::INIT);
}


STAFResultPtr resolveStr(STAFServiceRequestLevel30 *pInfo,
                         ResPoolServiceData *pData, 
                         const STAFString &theString)
{
    return pData->fHandlePtr->submit(sLocal, sVar, sResStrResolve +
                                     STAFString(pInfo->requestNumber) +
                                     sString +
                                     pData->fHandlePtr->wrapData(theString));
}


STAFResultPtr resolveOp(STAFServiceRequestLevel30 *pInfo, 
                        ResPoolServiceData *pData,
                        STAFCommandParseResultPtr &parsedResult,
                        const STAFString &fOption, unsigned int optionIndex)
{
    STAFString optionValue = parsedResult->optionValue(fOption, optionIndex);

    if (optionValue.find(sLeftCurlyBrace) == STAFString::kNPos)
    {
        return STAFResultPtr(new STAFResult(kSTAFOk, optionValue),
                             STAFResultPtr::INIT);
    }

    return resolveStr(pInfo, pData, optionValue);
}


STAFResultPtr resolveOpLocal(ResPoolServiceData *pData,
                             STAFCommandParseResultPtr &parsedResult,
                             const STAFString &fOption, 
                             unsigned int optionIndex)
{
    STAFString optionValue = parsedResult->optionValue(fOption, optionIndex);

    if (optionValue.find(sLeftCurlyBrace) == STAFString::kNPos)
    {
        return STAFResultPtr(new STAFResult(kSTAFOk, optionValue),
                             STAFResultPtr::INIT);
    }
        
    return pData->fHandlePtr->submit(sLocal, sVar, sResStrResolve +
                                     sString +
                                     pData->fHandlePtr->wrapData(optionValue));
}


void registerHelpData(ResPoolServiceData *pData, unsigned int errorNumber,
                      const STAFString &shortInfo, const STAFString &longInfo)
{
    static STAFString regString("REGISTER SERVICE %C ERROR %d INFO %C "
                                "DESCRIPTION %C");

    pData->fHandlePtr->submit(sLocal, sHelp, STAFHandle::formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber,
        shortInfo.getImpl(), longInfo.getImpl()));
}


void unregisterHelpData(ResPoolServiceData *pData, unsigned int errorNumber)
{
    static STAFString regString("UNREGISTER SERVICE %C ERROR %d");

    pData->fHandlePtr->submit(sLocal, sHelp, STAFHandle::formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber));
}


void writeUIntToFile(ostream &output, unsigned int data, unsigned int length)
{
    union
    {
        char bytes[4];
        unsigned int uint;
    };

    uint = STAFUtilSwapUInt(STAFUtilConvertNativeUIntToLE(data));

    output.write(&bytes[4 - length], length);
}


void readUIntFromFile(istream &input, unsigned int &data, unsigned int length)
{
    union
    {
        char bytes[4];
        unsigned int uint;
    };

    uint = 0;

    input.read(&bytes[4 - length], length);

    data = STAFUtilConvertLEUIntToNative(STAFUtilSwapUInt(uint));
}


void readStringFromFile(istream &input, STAFString &inString)
{
    // First read in the UTF-8 data

    unsigned int stringLength = 0;

    readUIntFromFile(input, stringLength);

    char *inputData = new char[stringLength];

    input.read(inputData, stringLength);

    try
    {
        inString = STAFString(inputData, stringLength, STAFString::kUTF8);
    }
    catch(...)
    {
        delete [] inputData;
        throw;
    }

    delete [] inputData;
}


void writeStringToFile(ostream &output, STAFString &outString)
{
    unsigned int stringLength = outString.length();

    writeUIntToFile(output, stringLength);
    output.write(outString.buffer(), stringLength);
}


// Read a resource pool file and store its contents in PoolData.
// Currently, there are two formats of resource pool files supported:
// - 1 is the current binary format used by the C++ Respool Service in STAF 2.3
//   and later
// - "0" is the old text format used by the REXX Respool Service in STAF 2.2
//   and earlier

unsigned int readPoolFile(const STAFString &fileName, PoolData &poolData)
{
    unsigned int rc = kReadorWriteOk;

    // Open the pool file (in binary mode)

    fstream poolfile(fileName.toCurrentCodePage()->buffer(),
                     ios::in | STAF_ios_binary);

    if (!poolfile) return kFileOpenError;
    
    // Check the format of the file to determine how to read the file

    readUIntFromFile(poolfile, poolData.fileFormat);

    if (poolfile.eof()) return kReadEndOfFile;

    if (poolData.fileFormat == sCurrFileFormat)
    {
        // This is the current binary format used by the C++ Respool Service

        readStringFromFile(poolfile, poolData.poolName);
        readStringFromFile(poolfile, poolData.poolDescription);
        readUIntFromFile(poolfile, poolData.numResources);
        
        STAFString entry;
        
        for (unsigned int i = 0; i < poolData.numResources; i++) 
        {
            readStringFromFile(poolfile, entry);
            
            ResourceData resourceData(entry);
            poolData.resourceList.push_back(resourceData);
        }
    }
    else
    {
        // Check if has the old text format used by the Rexx Respool Service

        // Close and reopen the pool file (in text mode)

        poolfile.close();
        poolfile.open(fileName.toCurrentCodePage()->buffer(), ios::in);
        if (!poolfile) return kFileOpenError;

        STAFString line;
        char poolLine[1024];

        // Read pool file format line

        poolLine[0] = 0;
        poolfile.getline(poolLine, 1024);
        
        if (poolfile.eof()) return kReadEndOfFile;
        
        STAFString line1(poolLine);
        STAFString fileFormatString = line1;

        // Validate file format and process based on file format

        if (fileFormatString != STAFString("0"))
        {   
            // Unknown/invalid file format
            return kReadInvalidFormat;
        }
       
        // A String "0" format was used by the old REXX Respool service

        poolData.fileFormat = 0;

        // Read pool name line
        poolLine[0] = 0;
        poolfile.getline(poolLine, 1024);

        if (poolfile.eof()) return kReadEndOfFile;

        STAFString line2(poolLine);
        poolData.poolName = line2;
        
        // Read pool description line
        poolLine[0] = 0;
        poolfile.getline(poolLine, 1024);
            
        if (poolfile.eof()) return kReadEndOfFile;
           
        STAFString line3(poolLine);
        poolData.poolDescription = line3;

        // Read resource entry lines
        poolLine[0] = 0;
        poolfile.getline(poolLine, 1024);
 
        unsigned int i = 0;

        for (i = 0; !poolfile.eof() && !rc; i++)
        {
            STAFString entry(poolLine);
            ResourceData resourceData(entry);
            poolData.resourceList.push_back(resourceData);
            
            // Get next line
            poolLine[0] = 0;
            poolfile.getline(poolLine, 1024);       
        }

        poolData.numResources = i;
    }

    return rc;
}


// Write the resource pool file.  Note that pool files are always written
// in the current format.

unsigned int writePoolFile(const STAFString &fileName, PoolData &poolData)
{
    // Open the pool file

    fstream poolfile(fileName.toCurrentCodePage()->buffer(),
                    ios::out | STAF_ios_binary);
    if (!poolfile) return kFileOpenError;
    
    // Write to the pool file

    writeUIntToFile(poolfile, sCurrFileFormat);
    writeStringToFile(poolfile, poolData.poolName);
    writeStringToFile(poolfile, poolData.poolDescription);
    unsigned int numResources = poolData.resourceList.size();
    writeUIntToFile(poolfile, numResources);
    
    for (unsigned int i = 0; i < numResources; i++) 
    {
        writeStringToFile(poolfile, poolData.resourceList[i].entry);
    }

    return kReadorWriteOk;
}

