/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import com.ibm.staf.*;
import com.ibm.staf.wrapper.STAFLog;
import java.util.Map;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.io.File;
import org.python.core.Py;
import org.python.core.PyObject;
import org.python.core.PyCode;
import org.python.core.__builtin__;

/**
 * The representation of a STAX job.  A STAX job is created by the STAXParser
 * when the STAX service receives an EXECUTE request specifying a file or string
 * containing XML that defines a STAX job.
 */
public class STAXJob implements STAXThreadCompleteListener,
                                STAXSTAFQueueListener
{
    // For debugging - Counts and prints the number of cache gets/adds
    static final boolean COUNT_PYCODE_CACHES = false;

    static final String STAX_JOB_EVENT = new String("Job");

    // Logfile types:

    static final int SERVICE_LOG = 1;

    /**
     * Indicates to log to the STAX Job Log file
     */
    public static final int JOB_LOG = 2;

    /**
     * Indicates to log to the STAX Job User Log file
     */
    public static final int USER_JOB_LOG = 3;

    public static final int NO_NOTIFY_ONEND = 0;
    public static final int NOTIFY_ONEND_BY_HANDLE = 1;
    public static final int NOTIFY_ONEND_BY_NAME = 2;

    /** 
      * Creates a new STAXJob instance passing in a STAX object which
      * represents the STAX service that is executing this job.
      */
    public STAXJob(STAX staxService)
    {
        fSTAX = staxService;
        STAXThread thread = new STAXThread(this);
        thread.addCompletionNotifiee(this);
        fThreadMap.put(new Integer(thread.getThreadNumber()), thread);
        fClearlogs = fSTAX.getClearlogs();
        fLogTCElapsedTime = fSTAX.getLogTCElapsedTime();
        fLogTCNumStarts   = fSTAX.getLogTCNumStarts();
        fLogTCStartStop   = fSTAX.getLogTCStartStop();
    }

    /**
     * Gets the STAX object which represents the job's STAX service
     * @return an instance of the job's STAX service
     */
    public STAX getSTAX() { return fSTAX; }

    /**
     * Gets the next number for a thread in a job
     * @return a number for the next thread in a job
     */
    public int getNextThreadNumber()
    {
        synchronized (fNextThreadNumberSynch)
        {
            return fNextThreadNumber++;
        }
    }

    /**
     * Gets the name of the function that should be called to start the
     * execution of a job
     * @return the name of the starting function for a job
     */
    public String getStartFunction() { return fStartFunction; }

    /**
     * Sets the name of the function that should be called to start the
     * execution of a job
     * @param  startFunction   the name of the starting function for a job
     */
    public void setStartFunction(String startFunction)
    { fStartFunction = startFunction; }

    public String getStartFuncArgs() { return fStartFuncArgs; }

    public void setStartFuncArgs(String startFuncArgs)
    { fStartFuncArgs = startFuncArgs; }

    public void setExecuteAndHold()
    { fExecuteAndHold = true; }

    public String getJobName() { return fJobName; }

    public void setJobName(String jobName) { fJobName = jobName; }

    public int getJobNumber() { return fJobNumber; }

    public void setJobNumber(int jobNumber) { fJobNumber = jobNumber; }
    
    public int getNextProcNumber()
    {
        synchronized (fNextProcNumberSynch)
        { 
            return fProcNumber++;
        }
    }
    
    public int getNextCmdNumber()
    {
        synchronized (fNextCmdNumberSynch)
        {
            return fCmdNumber++;
        }
    }
    
    public int getNextProcessKey()
    {
        synchronized (fNextProcessKeySynch)
        {
            return fProcessKey++;
        }
    }
    public String getJobDataDir() { return fJobDataDir; }
    
    public String getXmlMachine() { return fXmlMachine; }

    public void setXmlMachine(String machName) 
    { fXmlMachine = machName; }

    public String getXmlFile() { return fXmlFile; }

    public void setXmlFile(String fileName) { fXmlFile = fileName; }

    public List getScripts() { return fScripts; }

    public void setScript(String script) { fScripts.add(script); }

    public List getScriptFiles() { return fScriptFiles; }

    public void setScriptFile(String fileName) { fScriptFiles.add(fileName); }
    
    public String getScriptFileMachine() { return fScriptFileMachine; }

    public void setScriptFileMachine(String machName) 
    { fScriptFileMachine = machName; }

    public String getSourceMachine() { return fSourceMachine; }

    public void setSourceMachine(String machName) 
    { fSourceMachine = machName; }

    public String getSourceHandleName() { return fSourceHandleName; }

    public void setSourceHandleName(String handleName) 
    { fSourceHandleName = handleName; }

    public int getSourceHandle() { return fSourceHandle; }

    public void setSourceHandle(int handle) 
    { fSourceHandle = handle; }

    public boolean getClearlogs() { return fClearlogs; }

    public String getClearLogsAsString()
    {
        if (fClearlogs)
            return "Enabled";
        else
            return "Disabled";
    }

    public void setClearlogs(boolean clearlogs) 
    { fClearlogs = clearlogs; }

    public int getNotifyOnEnd() { return fNotifyOnEnd; }

    public String getNotifyOnEndAsString()
    {
        if (fNotifyOnEnd == STAXJob.NOTIFY_ONEND_BY_NAME)
            return "By Name";
        else if (fNotifyOnEnd == STAXJob.NOTIFY_ONEND_BY_HANDLE)
            return "By Handle";
        else 
            return "No";
    }

    public void setNotifyOnEnd(int notifyFlag) { fNotifyOnEnd = notifyFlag; }
    
    public PyObject getResult() { return fResult; }

    public void setResult(PyObject result) 
    { fResult = result; }

    public STAXTimestamp getStartTimestamp() { return fStartTimestamp; }
    
    public STAXTimestamp getEndTimestamp() { return fEndTimestamp; }
    
    public Integer getJobNumberAsInteger() { return new Integer(fJobNumber); }

    public STAFHandle getSTAFHandle() { return fHandle; }

    public boolean getLogTCElapsedTime() { return fLogTCElapsedTime; }

    public String getLogTCElapsedTimeAsString()
    {
        if (fLogTCElapsedTime)
            return "Enabled";
        else
            return "Disabled";
    }

    public void setLogTCElapsedTime(boolean logTCElapsedTime)
    { fLogTCElapsedTime = logTCElapsedTime; }

    public boolean getLogTCNumStarts() { return fLogTCNumStarts; }

    public String getLogTCNumStartsAsString()
    {
        if (fLogTCNumStarts)
            return "Enabled";
        else
            return "Disabled";
    }

    public void setLogTCNumStarts(boolean logTCNumStarts)
    { fLogTCNumStarts = logTCNumStarts; }

    public boolean getLogTCStartStop() { return fLogTCStartStop; }
    
    public String getLogTCStartStopAsString()
    {
        if (fLogTCStartStop)
            return "Enabled";
        else
            return "Disabled";
    }

    public void setLogTCStartStop(boolean logTCStartStop)
    { fLogTCStartStop = logTCStartStop; }

    public STAXThread getMainThread()
    {
        synchronized (fThreadMap)
        {
            return (STAXThread)fThreadMap.get(new Integer(1));
        }
    }

    public void addFunction(STAXFunctionAction function)
    {
        synchronized(this)
        {    
            if (fFunctionMap.containsKey(function.getName()))
            {
                /* XXX: Do something */
            }

            fFunctionMap.put(function.getName(), function);
        }
    }

    public STAXAction getFunction(String name)
    {
        synchronized(this)
        {
            return (STAXAction)fFunctionMap.get(name);
        }
    }
    
    public HashMap getFunctionMap()
    {
        synchronized(this)
        {
            return fFunctionMap;
        }
    }

    public boolean functionExists(String name)
    {
        synchronized(this)
        {
            return fFunctionMap.containsKey(name);
        }
    }

    public void addDefaultAction(STAXAction action)
    {
        fDefaultActions.addFirst(action);
    }

    public void addCompletionNotifiee(STAXJobCompleteListener listener)
    {
        synchronized (fCompletionNotifiees)
        {
            fCompletionNotifiees.addLast(listener);
        }
    }

    public STAFResult addCompletionNotifiee2(STAXJobCompleteNotifiee notifiee)
    {
        synchronized (fCompletionNotifiees)
        {
            // Check if the notifiee is already in the list

            Iterator iter = fCompletionNotifiees.iterator();

            while (iter.hasNext())
            {
                Object notifieeObj = iter.next();

                if (notifieeObj instanceof
                    com.ibm.staf.service.stax.STAXJobCompleteNotifiee)
                {
                    STAXJobCompleteNotifiee aNotifiee =
                        (STAXJobCompleteNotifiee)notifieeObj;

                    if (aNotifiee.getMachine().equals(notifiee.getMachine()) &&
                        aNotifiee.getHandle() == notifiee.getHandle() &&
                        aNotifiee.getHandleName().equals(notifiee.getHandleName()))
                    {
                        return new STAFResult(
                            STAFResult.AlreadyExists,
                            "A notifiee is already registered for machine=" +
                            notifiee.getMachine() +
                            ", handle=" + notifiee.getHandle() +
                            ", handleName=" + notifiee.getHandleName());
                    }
                }

            }

            // Add to the end of the list

            fCompletionNotifiees.addLast(notifiee);
        }

        return new STAFResult(STAFResult.Ok, "");
    }

    public STAFResult removeCompletionNotifiee(STAXJobCompleteNotifiee notifiee)
    {
        boolean found = false;

        synchronized (fCompletionNotifiees)
        {
            // Check if notifiee exists.  If so, remove the notifiee

            Iterator iter = fCompletionNotifiees.iterator();

            while (iter.hasNext())
            {
                Object notifieeObj = iter.next();

                if (notifieeObj instanceof
                    com.ibm.staf.service.stax.STAXJobCompleteNotifiee)
                {
                    STAXJobCompleteNotifiee aNotifiee =
                        (STAXJobCompleteNotifiee)notifieeObj;
                
                    if (aNotifiee.getMachine().equals(notifiee.getMachine()) &&
                        aNotifiee.getHandle() == notifiee.getHandle() &&
                        aNotifiee.getHandleName().equals(notifiee.getHandleName()))
                    {
                        // Remove notifiee from list

                        fCompletionNotifiees.remove(aNotifiee);
                        found = true;
                        return new STAFResult(STAFResult.Ok, "");
                    }
                }
            }
        }
                
        return new STAFResult(
            STAFResult.DoesNotExist,
            "No notifiee registered for machine=" + notifiee.getMachine() +
            ", handle=" + notifiee.getHandle() +
            ", handleName=" + notifiee.getHandleName());
    }

    public LinkedList getCompletionNotifiees()
    {
        synchronized(fCompletionNotifiees)
        {
            return new LinkedList(fCompletionNotifiees);
        }
    }

    // The combination of __builtin__.compile and the exec version, which
    // takes a code object, is used so that if Jython code is executed
    // more than once, the compiled code is cached, eliminating the need
    // to recompile the string each time it is executed.
    //
    // Input Arguments:
    //  - codeString:  must be valid Python code; however, like ordinary
    //                 Python code, the existence of variables will not be
    //                 checked until the code is executed.
    //  - kind:        is either 'exec' if the string is made up of
    //                 statements, 'eval' if it is an expression.

    public PyCode getCompiledPyCode(String codeString, String kind)
    {
        PyCode codeObject = (PyCode)fCompiledPyCodeCache.get(codeString);

        if (codeObject == null)
        {
            if (COUNT_PYCODE_CACHES) fCompiledPyCodeCacheAdds++;

            // XXX: Would prefer to do a "eval" instead of "exec" so that
            // could remove assignment to STAXPyEvalResult, but I don't
            // know how to evaluate a PyCode object in STAXThread.pyCompile()
            // via a Python Interpreter as it's eval method only accepts a
            // string, not a PyCode object.

            if (kind.equals("eval"))
            {
                // Set to avoid error of setting STAXPyEvalResult to nothing
                if (codeString.equals("")) codeString = "None";
                
                codeObject = __builtin__.compile(
                    "STAXPyEvalResult = " + codeString, "<pyEval string>",
                    "exec");
            }
            else
            {
                codeObject = __builtin__.compile(
                    codeString, "<pyExec string>", "exec");
            }

            fCompiledPyCodeCache.put(codeString, codeObject);
        }
        else
        {
            if (COUNT_PYCODE_CACHES) fCompiledPyCodeCacheGets++;
        }
        
        return codeObject;
    }
    
    // Queue listener methods

    public void registerSTAFQueueListener(String msgType, 
                                          STAXSTAFQueueListener listener)
    {
        synchronized (fQueueListenerMap)
        {
            TreeSet listenerSet = (TreeSet)fQueueListenerMap.get(msgType);

            if (listenerSet == null)
            {
                listenerSet = new TreeSet(new STAXObjectComparator());
                fQueueListenerMap.put(msgType, listenerSet);
            }

            listenerSet.add(listener);
        }
    }

    public void unregisterSTAFQueueListener(String msgType, 
                                            STAXSTAFQueueListener listener)
    {
        synchronized (fQueueListenerMap)
        {
            TreeSet listenerSet = (TreeSet)fQueueListenerMap.get(msgType);

            if (listenerSet != null)
                listenerSet.remove(listener);
        }
    }

    //
    // Data management functions
    //

    public boolean setData(String dataName, Object data)
    {
        // Return true if added successfully, else return false.

        synchronized (fDataMap)
        {
            if (!fDataMap.containsKey(dataName))
            {
                fDataMap.put(dataName, data);
                return true;
            }
        }

        return false;
    }
 
    public Object getData(String dataName)
    {
        synchronized (fDataMap)
        { return fDataMap.get(dataName); }
    }

    public boolean removeData(String dataName)
    {
        // Return true if removed successfully, else return false.

        synchronized (fDataMap)
        {
            if (fDataMap.containsKey(dataName))
            {
                fDataMap.remove(dataName);
                return true;
            }
        }

        return false;
    }

    //
    // Execution methods
    //

    public void startExecution() 
        throws STAFException, STAXException
    {
        fHandle = new STAFHandle("STAX/Job/" + fJobNumber);

        if (fClearlogs) clearLogs();

        // Assign the name of the directory that the job can store data in
        fJobDataDir = fSTAX.getDataDir() + STAX.fileSep + "job" +
            STAX.fileSep + "Job" + fJobNumber;
        
        clearJobDataDir();

        fSTAFQueueMonitor = new STAFQueueMonitor(this);
        fSTAFQueueMonitor.start();

        // Initialize data for the Job

        fSTAX.visitJobManagementHandlers(new STAXVisitorHelper(this)
        {
            public void visit(Object o, Iterator iter)
            {
                STAXJobManagementHandler handler = (STAXJobManagementHandler)o;

                handler.initJob((STAXJob)fData);
            }
        });

        // Register to listen for messages that STAF requests have completed

        registerSTAFQueueListener("STAF/RequestComplete", this);

        // Generate job start event
        
        HashMap jobStartMap = new HashMap();
        jobStartMap.put("type", "job");
        jobStartMap.put("block", "main");
        jobStartMap.put("status", "begin");
        jobStartMap.put("jobID", String.valueOf(fJobNumber));
        jobStartMap.put("startFunction", fStartFunction);
        jobStartMap.put("jobName", fJobName);

        generateEvent(STAXJob.STAX_JOB_EVENT, jobStartMap, true);

        // Get the current date and time and set as the starting date/time
        fStartTimestamp = new STAXTimestamp();

        STAXThread thread = (STAXThread)fThreadMap.get(new Integer(1));

        STAXAction action = (STAXAction)fFunctionMap.get(fStartFunction);

        if (action == null)
        {
            throw new STAXInvalidStartFunctionException(fStartFunction);
        }
        else
        {   // Call the main function passing any arguments
            action = new STAXCallAction("'" + fStartFunction + "'", fStartFuncArgs);
        }

        ArrayList actionList = new ArrayList();
        
        // If HOLD was specified on EXECUTE request, add hold action as
        // the first action in the actionList.
        if (fExecuteAndHold)
           actionList.add(new STAXHoldAction("'main'"));

        // Add additional Python variables about the Job
        thread.pySetVar("STAXJobID", new Integer(fJobNumber));
        thread.pySetVar("STAXJobName", fJobName);
        thread.pySetVar("STAXJobXMLFile", fXmlFile);
        thread.pySetVar("STAXJobXMLMachine", fXmlMachine);
        thread.pySetVar("STAXJobStartDate", fStartTimestamp.getDateString());
        thread.pySetVar("STAXJobStartTime", fStartTimestamp.getTimeString());
        thread.pySetVar("STAXJobSourceMachine", fSourceMachine);
        thread.pySetVar("STAXJobSourceHandleName", fSourceHandleName);
        thread.pySetVar("STAXJobSourceHandle", new Integer(fSourceHandle));
        thread.pySetVar("STAXJobStartFunctionName", fStartFunction);
        thread.pySetVar("STAXJobStartFunctionArgs", fStartFuncArgs);
        thread.pySetVar("STAXCurrentFunction", Py.None);

        if (!fScriptFileMachine.equals(""))
            thread.pySetVar("STAXJobScriptFileMachine", fScriptFileMachine);
        else
            thread.pySetVar("STAXJobScriptFileMachine", fSourceMachine);

        thread.pySetVar("STAXJobScriptFiles", fScriptFiles.toArray());
        thread.pySetVar("STAXJobHandle", fHandle);

        thread.pySetVar("STAXServiceName", fSTAX.getServiceName());
        thread.pySetVar("STAXServiceMachine", fSTAX.getLocalMachineName());
        thread.pySetVar("STAXServiceMachineNickname",
                        fSTAX.getLocalMachineNickname());
        thread.pySetVar("STAXEventServiceName", fSTAX.getEventServiceName());
        thread.pySetVar("STAXEventServiceMachine",
                        fSTAX.getEventServiceMachine());

        thread.pySetVar("STAXJobUserLog", new STAFLog(
            STAFLog.MACHINE,
            fSTAX.getServiceName().toUpperCase() + "_Job_" +
            fJobNumber + "_User",
            fHandle, 0));
        thread.pySetVar("STAXJobLogName",
                        fSTAX.getServiceName().toUpperCase() +
                        "_Job_" + fJobNumber);
        thread.pySetVar("STAXJobUserLogName",
                        fSTAX.getServiceName().toUpperCase() +
                        "_Job_" + fJobNumber + "_User");
        
        thread.pySetVar("STAXJobWriteLocation", fJobDataDir);
        thread.pySetVar("STAXMessageLog", new Integer(0));
        thread.pySetVar("STAXLogMessage", new Integer(0));

        if (fLogTCElapsedTime)
            thread.pySetVar("STAXLogTCElapsedTime", new Integer(1));
        else
            thread.pySetVar("STAXLogTCElapsedTime", new Integer(0));

        if (fLogTCNumStarts)
            thread.pySetVar("STAXLogTCNumStarts", new Integer(1));
        else
            thread.pySetVar("STAXLogTCNumStarts", new Integer(0));

        if (fLogTCStartStop)
            thread.pySetVar("STAXLogTCStartStop", new Integer(1));
        else
            thread.pySetVar("STAXLogTCStartStop", new Integer(0));

        // Add default signal handlers to the actionList for the main block.
        addDefaultSignalHandlers(actionList);
     
        // Put the "main" function/block at the bottom of the stack,
        // with its action being a sequence group that contains the
        // default actions (scripts and signalhandlers) in the <stax> 
        // element and then the call of the main function. 
      
        while (!fDefaultActions.isEmpty())
        {
            actionList.add((STAXAction)fDefaultActions.removeLast());
        }
        actionList.add(action);  // Add call of main function
        action = new STAXSequenceAction(actionList);
        thread.pushAction(new STAXBlockAction("main", action));

        // Log a "start" job message in the Service and Job logs
        String msg = "JobID: " + fJobNumber + 
                     ", File: " + fXmlFile +
                     ", Machine: " + fXmlMachine + 
                     ", Function: " + fStartFunction +
                     ", Args: " + fStartFuncArgs +
                     ", JobName: ";

        if (fJobName.equals(""))
            msg += "<N/A>";
        else
            msg += fJobName;

        Iterator it = fScriptFiles.iterator();

        while (it.hasNext())
            msg += ", ScriptFile: " + (String)it.next();

        if (! fScriptFileMachine.equals(""))
            msg += ", ScriptFileMachine: " + fScriptFileMachine;

        it = fScripts.iterator();
        while (it.hasNext())
            msg += ", Script: " + (String)it.next();

        log(STAXJob.SERVICE_LOG, "start", msg);
        log(STAXJob.JOB_LOG, "start", msg);
        
        thread.pySetVar("STAXJob", this);
 
        // Schedule the thread to run 
        thread.schedule();
    }

    public void addDefaultSignalHandlers(ArrayList actionList)
    {
        // Array of default SignalHandlers.  Each signal is described by
        // a signal name, an action, and a signal message variable name
        
        String[][] defaultSHArray = 
        {
            {
                "STAXPythonEvaluationError", "terminate", "STAXPythonEvalMsg"
            },
            {
                "STAXProcessStartError", "continue", "STAXProcessStartErrorMsg"
            },
            {
                "STAXProcessStartTimeout", "continue",
                "STAXProcessStartTimeoutMsg"
            },
            {
                "STAXCommandStartError", "terminate",
                "STAXCommandStartErrorMsg"
            },
            {
                "STAXFunctionDoesNotExist", "terminate",
                "STAXFunctionDoesNotExistMsg"
            },
            {
                "STAXInvalidBlockName", "terminate", "STAXInvalidBlockNameMsg"
            },
            {
                "STAXBlockDoesNotExist", "continue", "STAXBlockDoesNotExistMsg"
            },
            {
                "STAXLogError", "continue", "STAXLogMsg"
            },
            {
                "STAXTestcaseMissingError", "continue",
                "STAXTestcaseMissingMsg"
            },
            {
                "STAXInvalidTcStatusResult", "continue",
                "STAXInvalidTcStatusResultMsg"
            },
            {
                "STAXInvalidTimerValue", "terminate",
                "STAXInvalidTimerValueMsg"
            },
            {
                "STAXNoSuchSignalHandler", "continue",
                "STAXNoSuchSignalHandlerMsg"
            },
            {
                "STAXEmptyList", "continue", "STAXEmptyListMsg"
            },
            {
                "STAXFunctionArgValidate", "terminate",
                "STAXFunctionArgValidateMsg"
            },
            {
                "STAXImportError", "terminate", "STAXImportErrorMsg"
            },
            {
                "STAXInvalidTestcaseMode", "continue",
                "STAXInvalidTestcaseModeMsg"
            }
        };

        // Add default SignalHandlers to actionList

        for (int i = 0; i < defaultSHArray.length; i++)
        {
            ArrayList signalHandlerActionList = new ArrayList();

            String signalName = defaultSHArray[i][0];
            String signalAction = defaultSHArray[i][1];
            String signalMsgVarName = defaultSHArray[i][2];
            String signalMsgText = "";

            if (signalAction.equals("terminate"))
            {
                signalMsgText = "'" + signalName + " signal raised. " +
                    "Terminating job. '" + " + " + signalMsgVarName;
            }
            else if (signalAction.equals("continue"))
            {
                signalMsgText = "'" + signalName + " signal raised. " +
                    "Continuing job. '" + " + " + signalMsgVarName;
            }
            
            // Log the error message in the STAX job log and send to the STAX
            // job monitor.

            signalHandlerActionList.add(new STAXLogAction(
                signalMsgText, "'error'", "1", "1", STAXJob.JOB_LOG));

            if (signalAction.equals("terminate"))
            {
                // Add a Terminate Job Action to the action list
                signalHandlerActionList.add(
                    new STAXTerminateAction("'main'"));
            }

            // Add the signalhandler action to the action list

            actionList.add(
                new STAXSignalHandlerAction(
                    "'" + signalName + "'",
                    new STAXSequenceAction(signalHandlerActionList)));
        }
    }

    //
    // Event methods
    //

    public void generateEvent(String eventSubType, Map details)
    {
        generateEvent(eventSubType, details, false);
    }

    public void generateEvent(String eventSubType, Map details, 
                              boolean notifyAll)
    {
        String detailsString = "";
        String key = "";
        String value = "";

        Iterator keyIter = details.keySet().iterator();
        
        while (keyIter.hasNext())
        {
            key = (String)keyIter.next();
            
            value = (String)details.get(key);
            
            if (value == null)
            {
                // do nothing
            }
            else
            {
                detailsString += "property " + STAFUtil.wrapData(
                    key + "=" + value) + " ";
            }
        }
        
        // The type machine should always be the local machine
        // The details parm must already be in the :length: format        
        STAFResult result = submitSync(
                        fSTAX.getEventServiceMachine(),
                        fSTAX.getEventServiceName(), "generate type " +
                        fSTAX.getServiceName().toUpperCase() + "/" + 
                        fSTAX.getLocalMachineName() + "/" + fJobNumber +
                        " subtype " + eventSubType +  " " + detailsString);
                        
        // Debug
        if (false)
        {
            System.out.println("Event:\n" +  "generate type " +
                        fSTAX.getServiceName().toUpperCase() + "/" + 
                        fSTAX.getLocalMachineName() + "/" + fJobNumber +
                        " subtype " + eventSubType + " " + details);
        }
        
        if (notifyAll)
        {
            submitSync(fSTAX.getEventServiceMachine(),
                       fSTAX.getEventServiceName(), "generate type " +
                       fSTAX.getServiceName().toUpperCase() + "/" + 
                       fSTAX.getLocalMachineName() + " subtype " + 
                       eventSubType + " " + detailsString);
        
            // Debug
            if (false)
            {
                System.out.println("Event:\n" + "generate type " +
                            fSTAX.getServiceName().toUpperCase() + "/" + 
                            fSTAX.getLocalMachineName() + " subtype " + 
                            eventSubType +  " " + details);
            }
        }
    }

    public STAFResult log(int logfile, String level, String message)
    {
        String serviceName = fSTAX.getServiceName().toUpperCase();
        String logName;

        if (logfile == STAXJob.SERVICE_LOG)
            logName = serviceName + "_Service";
        else if (logfile == STAXJob.JOB_LOG)
            logName = serviceName + "_Job_" + fJobNumber;
        else if (logfile == STAXJob.USER_JOB_LOG)
            logName = serviceName + "_Job_" + fJobNumber + "_User";
        else
        {   // Log to STAX_Service log if invalid logfile specified 
            System.out.println("STAX Service Error: Invalid Logfile" + 
                               logfile);
            logName = serviceName + "_Service";
        }

        String logRequest = "LOG MACHINE LOGNAME " + 
                            STAFUtil.wrapData(logName) +
                            " LEVEL " + level + 
                            " MESSAGE " + STAFUtil.wrapData(message);
                                
        STAFResult result = submitSync("LOCAL", "LOG", logRequest);

        // Check if the result was unsuccesful (except ignore RC 2)
        if (result.rc != 0 && result.rc != 2)
        {
            if (logfile != STAXJob.USER_JOB_LOG)
                {
                System.out.println("STAX Log failed with RC " +
                     result.rc + " and Result " + result.result +
                     "  level: " + level +
                     "  logRequest: " + logRequest);
                } 
        }
        return result;
    }
    
    public void clearLogs()
    {
        String serviceName = fSTAX.getServiceName().toUpperCase();
        String jobLogName;
        String userJobLogName;
        
        jobLogName = serviceName + "_Job_" + fJobNumber;
        userJobLogName = serviceName + "_Job_" + fJobNumber + "_User";
        
        String logRequest = "DELETE MACHINE " +
            fSTAX.getLocalMachineNickname() + " LOGNAME " + 
            STAFUtil.wrapData(jobLogName) + " CONFIRM";
                                
        STAFResult result = submitSync("LOCAL", "LOG", logRequest);
        
        logRequest = "DELETE MACHINE " + fSTAX.getLocalMachineNickname() +
            " LOGNAME " +  STAFUtil.wrapData(userJobLogName) + " CONFIRM";
                                
        result = submitSync("LOCAL", "LOG", logRequest);
    }
    
    public void clearJobDataDir()
    {
        // Delete the job data directory and recreate it
        
        File dir = new File(fJobDataDir);

        if (dir.exists())
        {
            String deleteDirRequest = "DELETE ENTRY " +
                STAFUtil.wrapData(fJobDataDir) + " RECURSE CONFIRM";
                                
            submitSync("local", "FS", deleteDirRequest);
        }

        if (!dir.exists())
        {
            dir.mkdirs();
        }
    }

    // Submit methods
    
    public STAFResult submitAsync(String location, String service, 
        String request, STAXSTAFRequestCompleteListener listener)
    {
        STAFResult result;
        
        synchronized(fRequestMap)
        {
            result = fHandle.submit2(
                STAFHandle.ReqQueue, location, service, request);
            
            if (result.rc == STAFResult.Ok)
            {
                try
                {
                    Integer requestNumber = new Integer(result.result);                                        
                        
                    fRequestMap.put(requestNumber, listener);
                            
                    result.result = 
                        (new Integer(requestNumber.intValue())).toString();
                }
                catch (NumberFormatException e)
                {
                    System.out.println(e.toString());
                    result.result = "0";
                }
            }
        }
        
        return result;
    }
    
    public STAFResult submitSync(String location, String service, 
                                 String request)
    {
        STAFResult result = fHandle.submit2(STAFHandle.ReqSync, 
            location, service, request);
        
        return result;
    }

    public STAFResult submitAsyncForget(String location, String service, 
                                        String request)
    {
        STAFResult result = fHandle.submit2(STAFHandle.ReqFireAndForget, 
            location, service, request);
        
        return result;
    }
    
    // STAXSTAFQueueListener method

    public void handleQueueMessage(STAXSTAFMessage message, STAXJob job)
    {
        int requestNumber = message.getRequestNumber();

        STAXSTAFRequestCompleteListener listener = null;

        synchronized (fRequestMap)
        {
            Integer key = new Integer(requestNumber);

            listener = (STAXSTAFRequestCompleteListener)fRequestMap.get(key);

            if (listener != null)
            {
                fRequestMap.remove(key);
            }
        }

        if (listener != null)
        {
            listener.requestComplete(requestNumber, new STAFResult(
                message.getRequestRC(), message.getRequestResult()));
        }
        else
        {   // Log a message in the job log
            String msg = "STAXJob.handleQueueMessage: " +
                         " No listener found for message:\n" +
                         STAFMarshallingContext.unmarshall(
                             message.getResult()).toString();
            job.log(STAXJob.JOB_LOG, "warning", msg);
        }
    }

    // STAXThreadCompleteListener method

    public void threadComplete(STAXThread thread, int endCode)
    {
        // Debug:
        if (false)
        {
            System.out.println("Thread #" + thread.getThreadNumber() +
                               " complete");
        }

        boolean jobComplete = false;

        synchronized (fThreadMap)
        {
            fThreadMap.remove(new Integer(thread.getThreadNumber()));
            if (fThreadMap.isEmpty()) jobComplete = true;
        }

        if (jobComplete == true)
        {
            // Perform terminateJob on interested parties

            fSTAX.visitJobManagementHandlers(new STAXVisitorHelper(this)
            {
                public void visit(Object o, Iterator iter)
                {
                    STAXJobManagementHandler handler =
                                             (STAXJobManagementHandler)o;

                    handler.terminateJob((STAXJob)fData);
                }
            });

            // Get the current date and time and set as job ending date/time

            fEndTimestamp = new STAXTimestamp();
            
            // Set the result so that it is available upon job completion.

            try
            {
                fResult = thread.pyObjectEval("STAXResult");
            }
            catch (STAXPythonEvaluationException e)
            {
                fResult = Py.None;
            }

            // Generate job completion event
            
            HashMap jobEndMap = new HashMap();
            jobEndMap.put("type", "job");
            jobEndMap.put("block", "main");
            jobEndMap.put("status", "end");
            jobEndMap.put("jobID", String.valueOf(fJobNumber));
            jobEndMap.put("result", fResult.toString()); 
                
            generateEvent(STAXJob.STAX_JOB_EVENT, jobEndMap, true);
            
            // Log the result from the job in a Status message in the Job log

            STAFMarshallingContext mc = STAFMarshallingContext.
                unmarshall(fResult.toString());
            log(STAXJob.JOB_LOG, "status", "Job Result: " + mc);

            // Log a Stop message for the job in the STAX Service and Job logs

            String msg = "JobID: " + fJobNumber; 
            log(STAXJob.SERVICE_LOG, "stop", msg);
            log(STAXJob.JOB_LOG, "stop", msg);

            submitSync("local", "QUEUE", "QUEUE TYPE STAF/Service/STAX/End " +
                       "MESSAGE " + STAFUtil.wrapData(""));

            // Debug
            if (COUNT_PYCODE_CACHES && STAX.CACHE_PYTHON_CODE)
            {   
                System.out.println(
                    "Job " + fJobNumber + ": " +
                    " cacheGets=" + fCompiledPyCodeCacheGets +
                    " cacheAdds=" + fCompiledPyCodeCacheAdds);
            }
            
            while (!fCompletionNotifiees.isEmpty())
            {
                STAXJobCompleteListener listener =
                    (STAXJobCompleteListener)fCompletionNotifiees.removeFirst();

                if (listener != null) listener.jobComplete(this);
            }

            try
            {
                fHandle.unRegister();
            }
            catch (STAFException e)
            {
                /* Do Nothing */
            }

            // Clear out job's private variables
            
            fScripts = null;
            fScriptFiles = null;
            fFunctionMap = null;
            fThreadMap = null;
            fHandle = null;
            fDefaultActions = null;
            fSTAFQueueMonitor = null;
            fCompletionNotifiees = null;
            fQueueListenerMap = null;
            fDataMap = null;
            fRequestMap = null;
            fCompiledPyCodeCache = null;
        }
    }

    private STAX fSTAX;
    private Object fNextThreadNumberSynch = new Object();
    private int fNextThreadNumber = 1;
    private int fJobNumber = 0;
    private Object fNextProcNumberSynch = new Object();
    private int fProcNumber = 1;
    private Object fNextCmdNumberSynch = new Object();
    private int fCmdNumber = 1;
    private Object fNextProcessKeySynch = new Object();
    private int fProcessKey = 1;
    private String fJobDataDir = new String();
    private String fJobName = new String();
    private String fXmlMachine = new String();
    private String fXmlFile = new String();
    private String fStartFunction = new String();
    private String fStartFuncArgs = "None";
    private boolean fClearlogs;
    private List fScripts = new ArrayList();
    private List fScriptFiles = new ArrayList();
    private String fScriptFileMachine = new String();
    private HashMap fFunctionMap = new HashMap();
    // I don't think we need a whole map, just a reference to thread 1
    private HashMap fThreadMap = new HashMap();
    private STAFHandle fHandle = null;
    private LinkedList fDefaultActions = new LinkedList();
    private STAFQueueMonitor fSTAFQueueMonitor = null;
    private LinkedList fCompletionNotifiees = new LinkedList();    
    private boolean fExecuteAndHold = false;
    private String fSourceMachine = new String();  // Source == Originating
    private String fSourceHandleName = new String();
    private int fSourceHandle;
    private int fNotifyOnEnd = STAXJob.NO_NOTIFY_ONEND;
    private PyObject fResult = Py.None;
    private STAXTimestamp fStartTimestamp;
    private STAXTimestamp fEndTimestamp;
    private HashMap fQueueListenerMap = new HashMap();
    private HashMap fDataMap = new HashMap();
    private HashMap fRequestMap = new HashMap(); // Map of active STAF requests
    private boolean fLogTCElapsedTime;
    private boolean fLogTCNumStarts;
    private boolean fLogTCStartStop;
    private HashMap fCompiledPyCodeCache = new HashMap();
    private long fCompiledPyCodeCacheAdds = 0;
    private long fCompiledPyCodeCacheGets = 0;
    
    class STAFQueueMonitor extends Thread
    {
        STAFQueueMonitor(STAXJob job)
        {
            fJob = job;
        }

        public void run()
        {
            for (;;)
            {
                STAFResult result = submitSync("local", "QUEUE", "GET WAIT");

                if (result.rc != STAFResult.Ok)
                {
                    // XXX: Maybe we should do something to prevent an infinite
                    //      loop
                    continue;
                }

                // Debug:
                if (false) System.out.println("Result: " + result.result);

                // Break up result
                STAXSTAFMessage msg = new STAXSTAFMessage(result.result);

                String type = msg.getType();

                if (type != null &&
                    type.equalsIgnoreCase("STAF/Service/STAX/End"))
                {
                    return;
                }

                notifyListeners(msg);
            }
        }

        public void notifyListeners(STAXSTAFMessage msg)
        {

            // Perform a lookup of registered message handlers for
            // this message prefix and pass the STAXSTAFMessage to
            // the registered message handlers.

            synchronized (fQueueListenerMap) 
            {
                Iterator mapIter = fQueueListenerMap.keySet().iterator();

                int listenersFound = 0;

                while (mapIter.hasNext())
                {
                    String msgType = (String)mapIter.next();
                    String theType = msg.getType();

                    if ((msgType == null) ||
                        // Messages from 3.x clients that STAX is interested
                        // in will have a type
                        (theType != null &&
                         theType.equalsIgnoreCase(msgType)) ||
                        // The following is for messages from 2.x clients
                        // which will have a null type and the message will
                        // begin with the type (which for processes will be
                        // STAF/PROCESS/END instead of STAF/Process/End
                        (theType == null && msg.getMessage().toUpperCase().
                         startsWith(msgType.toUpperCase())))
                    {
                        TreeSet listenerSet =
                                (TreeSet)fQueueListenerMap.get(msgType);

                        if (listenerSet == null) continue;

                        Iterator iter = listenerSet.iterator();

                        while (iter.hasNext())
                        {
                            STAXSTAFQueueListener listener = 
                                (STAXSTAFQueueListener)iter.next();

                            listener.handleQueueMessage(msg, fJob);
                            listenersFound++;
                        }
                    }
                }

                if (listenersFound == 0)
                {   // Log a message in the job log
                    String logMsg = "STAXJob.notifyListeners: " +
                        "No listener found for message:\n" +
                        STAFMarshallingContext.unmarshall(
                            msg.getResult()).toString();
                    fJob.log(STAXJob.JOB_LOG, "warning", logMsg);
                }
            }
        }

        STAXJob fJob;

    } // end STAFQueueMonitor
}
