/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import com.ibm.staf.*;
import com.ibm.staf.service.*;
import com.ibm.staf.STAFUtil;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Vector;
import java.util.List;
import java.util.ArrayList;
import org.python.core.Py;

/**
 * The representation of the action to take when a sub-job, &lt;job&gt;,
 * element is encountered.
 * <p>
 * This is produced by the STAXJobActionFactory.  The resulting STAXJobAction
 * object describes a &lt;job> element.  It contains the job-file or job-data
 * to be executed and other optional data, as well as a name for the sub-job
 * to be executed.
 *
 * @see STAXJobActionFactory
 */
public class STAXJobAction implements STAXAction,
                                      STAXJobCompleteListener,
                                      STAXThreadCompleteListener
{
    // Initial state of the action
    static final int INIT = 0;

    // State where the STAX EXECUTE request command has been submitted, and
    // now it's waiting for it to complete
    static final int WAIT_SUBJOB = 1;

    // State where the action has been notified that the STAX EXECUTE command
    // is complete
    static final int SUBJOB_COMPLETE = 2;

    // State where the job-action has completed.
    static final int SUBJOB_COMPLETE_WAIT_ACTION = 3;

    // Completion state of the action
    static final int COMPLETE = 4;

    static final String INIT_STRING = "INIT";
    static final String WAIT_SUBJOB_STRING = "WAIT_SUBJOB";
    static final String SUBJOB_COMPLETE_STRING = "SUBJOB_COMPLETE";
    static final String SUBJOB_COMPLETE_WAIT_ACTION_STRING =
        "SUBJOB_COMPLETE_WAIT_ACTION";
    static final String COMPLETE_STRING = "COMPLETE";
    static final String STATE_UNKNOWN_STRING = "UNKNOWN";

    /**
      * Creates a new STAXJobAction instance.
      */
    public STAXJobAction()
    {
    }

    /**
     * Gets the factory for the action.
     * @return an instance of the action's factory
     */
    public STAXJobActionFactory getActionFactory() { return fFactory; }

    /**
     * Sets the factory for the action.
     * @param  factory  an instance of the action's factory
     */
    public void setActionFactory(STAXJobActionFactory factory)
    {
        fFactory = factory;
    }

    /**
     * Gets the name specified for the sub-job.
     * @return the name of the sub-job
     */
    public String getName() { return fName; }

    /**
     * Sets a name used to identify the sub-job.  It defaults to
     * the value of the function name called to start the job.
     * @param  name  the name used to identify the sub-job
     */
    public void setName(String name)
    {
        fName = name;
        fUnevalName = name;
    }

    /**
     * Gets the clear logs option specified for the sub-job.
     * @return the clear log option specified for the sub-job
     */
    public String getClearlogs() { return fClearlogs; }

    /**
     * Sets the clearlogs option which is used to determine whether
     * to delete the STAX Job and Job User logs before the job is
     * executed to ensure that only one job's contents are in the logs.
     * It defaults to the same value specified for its parent job.
     * @param  clearlogs  the clearlog option for the sub-job
     */
    public void setClearlogs(String clearlogs)
    {
        fClearlogs = clearlogs;
        fUnevalClearlogs = clearlogs;
    }

    /**
     * Gets the monitor option specified for the sub-job.
     * @return the monitor option specified for the sub-job
     */
    public String getMonitor() { return fMonitor; }

    /**
     * Sets the monitor option which is used to determine whether
     * to automatically monitor the sub-job
     * @param  monitor  the monitor option for the sub-job
     */
    public void setMonitor(String monitor)
    {
        fMonitor = monitor;
        fUnevalMonitor = monitor;
    }

    /**
     * Gets the Log TC Elapsed Time option specified for the sub-job.
     * @return the Log TC Elapsed Time option specified for the sub-job
     */
    public String getLogTCElapsedTime() { return fLogTCElapsedTime; }

    /**
     * Sets the Log TC Elapsed Time option which is used to determine whether
     * to log the elapsed time for testcases in summary records in the STAX
     * Job log and for a LIST TESTCASES request.
     * It defaults to the same value specified for its parent job.
     * @param  logTCElapsedTime  the logTCElapsedTime option for the sub-job
     */
    public void setLogTCElapsedTime(String logTCElapsedTime)
    {
        fLogTCElapsedTime = logTCElapsedTime;
        fUnevalLogTCElapsedTime = logTCElapsedTime;
    }

    /**
     * Gets the Log TC Num Starts option specified for the sub-job.
     * @return the Log TC Num Starts option specified for the sub-job
     */
    public String getLogTCNumStarts() { return fLogTCNumStarts; }

    /**
     * Sets the Log TC Num Starts option which is used to determine whether
     * to log the number of starts for testcases in summary records in the
     * STAX * Job log and for a LIST TESTCASES request.
     * It defaults to the same value specified for its parent job.
     * @param  logTCNumStarts  the logTCNumStarts option for the sub-job
     */
    public void setLogTCNumStarts(String logTCNumStarts)
    {
        fLogTCNumStarts = logTCNumStarts;
        fUnevalLogTCNumStarts = logTCNumStarts;
    }


    /**
     * Gets the Log TC Start/Stop option specified for the sub-job.
     * @return the Log TC Start/Stop option specified for the sub-job
     */
    public String getLogTCStartStop() { return fLogTCStartStop; }

    /**
     * Sets the Log TC Start/Stop option which is used to determine whether
     * to log start and stop status records for testcases in the STAX Job log.
     * It defaults to the same value specified for its parent job.
     * @param  logTCStartStop  the logTCStartStop option for the sub-job
     */
    public void setLogTCStartStop(String logTCStartStop)
    {
        fLogTCStartStop = logTCStartStop;
        fUnevalLogTCStartStop = logTCStartStop;
    }

    /**
     * Gets the fully-qualified name of a file containing the XML document
     * for the STAX job to be executed.
     * @return the job file name
     */
    public String getJobFile() { return fJobFile; }

    /**
     * Gets the name of the machine where the XML job file is located
     * @return the job file machine name
     */
    public String getJobFileMachine() { return fJobFileMachine; }

    /**
     * Sets the fully-qualified name of a file containing the XML document
     * for the STAX job to be executed.
     * @param  jobXMLFile  a string containing the fully-qualified name of the
     *                     file containing the XML document for the STAX job
     *                     to be executed
     * @param  machine     a string containing the name of the machine where
     *                     the XML file is located
     */
    public void setJobFile(String jobFile, String machine)
    {
        fJobFile = jobFile;
        fUnevalJobFile = jobFile;
        fJobFileMachine = machine;
        fUnevalJobFileMachine = machine;
    }

    /**
     * Gets the string containing the XML document for the STAX job to be
     * executed.
     * @return the job data
     */
    public String getJobData() { return fJobData; }

    /**
     * Gets the indicator of whether the XML data is to be evaluated by Python
     * in the parent job.
     * @return the job data eval indicator
     */
    public String getJobDataEval() { return fJobDataEval; }

    /**
     * Sets a string containing the XML document for the STAX job to be
     * executed.
     * @param  jobData  a string containing the XML document for the
     *                  STAX job to be executed
     * @param  eval     indicates the XML data is to be evaluated by Python
     *                  in the parent job if it evaluates to true
     *      */
    public void setJobData(String jobData, String eval)
    {
        fJobData = jobData;
        fUnevalJobData = jobData;
        fJobDataEval = eval;
    }

    /**
     * Gets the name of function to call to start the STAX job.
     * @return the starting function name for the STAX job
     */
    public String getFunction() { return fFunction; }

    /**
     * Sets the name of the function to call to start the STAX job.
     * @param  function    the name of the starting function
     * @param  functionIf  indicates to ignore the function specified if it
     *                     evaluates to false
     */
    public void setFunction(String function, String functionIf)
    {
        fFunction = function;
        fUnevalFunction = function;
        fFunctionIf = functionIf;
    }

    /**
     * Gets the arguments to pass to the function called to start the STAX job.
     * @return the starting function arguments for the STAX job
     */
    public String getFunctionArgs() { return fFunctionArgs; }

    /**
     * Sets the arguments to pass to the function called to start the STAX job.
     * @param  args    the arguments passed to the starting function
     * @param  eval    indicates the arguments are to be evaluated by Python
     *                 in the parent job if it evaluates to true
     * @param  argsIf  indicates to ignore the function args specified if it
     *                 evaluates to false
     */
    public void setFunctionArgs(String args, String eval, String argsIf)
    {
        fFunctionArgs = args;
        fUnevalFunctionArgs = args;
        fFunctionArgsEval = eval;
        fFunctionArgsIf = argsIf;
    }

    /**
     * Gets a vector containing the scripts (Python code) specified for the job.
     * @return a vector containing the scripts specified for the STAX job
     */
    public Vector getScripts() { return fScripts; }

    /**
     * Adds a script (Python code) to the scripts vector.
     * @param  script    a script consist of Python code
     * @param  eval      indicates the script is to be evaluated by Python
     *                   in the parent job if it evaluates to true
     * @param  scriptIf  indicates to ignore the script if it evaluates to
     *                   false
     */
    public void setScripts(String script, String eval, String scriptIf)
    {
        fUnevalScripts.add(script);
        fScriptsEval.add(eval);
        fScriptsIf.add(scriptIf);
    }

    /**
     * Gets a vector containing the script file names specified for the job.
     * @return a vector containing the names of the script files
     */
    public Vector getScriptFiles() { return fScriptFiles; }

    /**
     * Sets the name(s) of the script file(s) containing Python code to be
     * executed in the STAX job.
     * @param  scriptFiles    the name(s) of the script file(s)
     * @param  machine        the name of the machine where the script file(s)
     *                        are located
     * @param  scriptFilesIf  indicates to ignore the scriptfiles if it
     *                        evaluates to false
     */
    public void setScriptFiles(String scriptFiles, String machine,
                               String scriptFilesIf)
    {
        fUnevalScriptFiles = scriptFiles;
        fScriptFilesMachine = machine;
        fUnevalScriptFilesMachine = machine;
        fScriptFilesIf = scriptFilesIf;
    }

    /**
     * Gets the STAXAction object to be executed after the sub-job has started.
     * @return the STAXAction object for the STAX job
     */
    public STAXAction getJobAction() { return fJobAction; }

    /**
     * Sets the name of the function to call to start the STAX job.
     * @param  function    the name of the STAF service
     * @param  functionIf  indicates to ignore the function specified if it
     *                     evaluates to false
     */
    public void setJobAction(STAXAction jobAction, String jobActionIf)
    {
        fJobAction = jobAction;
        fJobActionIf = jobActionIf;
    }

    /**
     * Gets the request string to submit to the STAF service.
     * @return the request string
     */
    public String getRequest() { return fRequest; }

    /**
     * Gets the STAX-Thread instance where this action is being executed.
     * @return the STAX-Thread instance where this action is being executed
     */
    public STAXThread getThread() { return fThread; }

    /**
     * Gets the STAX-Thread instance where this action is being executed.
     * @return the STAX-Thread instance where this action is being executed
     */
    public String getCurrentBlockName() { return fCurrentBlockName; }

    /**
     * Gets the timestamp for when this action was started.
     * @return the timestamp for when this action was started
     */
    public STAXTimestamp getStartTimestamp() { return fStartTimestamp; }

    /**
     * Gets the job ID for the sub-job that is submitted.
     * @return the job ID for the sub-job
     */
    public int getJobID() { return fJobID; }

    /**
     * Gets the name of the function used to start the STAX job
     * @return the starting function name for the STAX job
     */
    public String getStartFunction() { return fStartFunction; }

    /**
     * Sets the name of the function to call to start the STAX job.
     * @parm function    the name of starting function
     */
    public void setStartFunction(String function)
    {
        fStartFunction = function;
    }

    /**
     * Gets the arguments to pass to the function called to start the STAX job.
     * @return the starting function arguments for the STAX job
     */
    public String getStartFuncArgs() { return fStartFunctionArgs; }

    /**
     * Sets the arguments to pass to the function called to start the STAX job.
     * @param  args    the arguments passed to the starting function
     */
    public void setStartFuncArgs(String args)
    {
        fStartFunctionArgs = args;
    }

    /**
     * Gets a string identifying the state of this action.  It could be
     * "INIT", "WAIT_REQUEST", "SUBJOB_COMPLETE", "SUBJOB_COMPLETE_WAIT_ACTION",
     * "COMPLETE", or "UNKNOWN".
     * @return a string identifying the state of this action.
     */
    public String getStateAsString()
    {
        switch (fState)
        {
            case INIT:
                return INIT_STRING;
            case WAIT_SUBJOB:
                return WAIT_SUBJOB_STRING;
            case SUBJOB_COMPLETE:
                return SUBJOB_COMPLETE_STRING;
            case SUBJOB_COMPLETE_WAIT_ACTION:
                return SUBJOB_COMPLETE_WAIT_ACTION_STRING;
            case COMPLETE:
                return COMPLETE_STRING;
            default:
                return STATE_UNKNOWN_STRING;
        }
    }

    public String getInfo()
    {
        return fName;
    }

    public String getDetails()
    {
        return "JobName:" + fName +
               ";JobID:" + fJobID +
               ";Request:" + fRequest +
               ";State:" + getStateAsString() +
               ";BlockName:" + fCurrentBlockName +
               ";StartTimestamp:" + fStartTimestamp +
               ";HoldThreadCondition:" + fHoldCondition;
    }

    // Submits an EXECUTE request to the STAX service by doing an asynchronous
    // submit of the request to the local STAX machine, adds a hold thread
    // condition, and waits for the submitted job to complete.
    //
    // If in its INIT state, it does the following:
    // - Evaluates (using Python) values as needed. If a Python evaluation
    //   exception occurs, it raises a STAXPythonEvaluationError signal and
    //   pops itself off the action stack.
    // - Submits the STAX EXECUTE request asyncronously.  If an error occurs
    //   submitting the request, it sets RC and STAFResult and pops itself
    //   off the action stack.
    // - Adds a hold thread condition while waiting for the job to complete
    //   so that another thread can become active.
    // - Adds the running job to the subJobMap so that it can be listed.
    // - Generates an event is generated to indicate that the sub-job has
    //   been started.
    //
    // If in its SUBJOB_COMPLETE state, it does the following:
    // - Removes its entry from the subJobMap so that it no longer will show
    //   up in the list of sub-jobs.
    // - Pops itself off the action stack since it is now complete.
    //
    // Note that this entire method is synchronized since its state can be
    // changed on another thread (e.g. via the jobComplete method).

    public synchronized void execute(STAXThread thread)
    {
        if (fState == INIT)
        {
            fThread = thread;
            String request = "execute";

            String evalElem = "";
            String evalAttr = "";

            try
            {
                // Set RC initially so if an error occurs (that does not
                // terminate the job) before a STAF request is
                // submitted, <if expr="RC != 0"> won't fail.
                fThread.pySetVar("RC", new Integer(-1));

                // Set STAFResult and STAXResult initially to Py.None
                fThread.pySetVar("STAFResult", Py.None);
                fThread.pySetVar("STAXResult", Py.None);

                // Set STAXSubJobID initially to 0
                fThread.pySetVar("STAXSubJobID", new Integer(0));

                if (!fUnevalJobFile.equals(""))
                {
                    evalElem = "job-file";
                    evalAttr = "";

                    fJobFile = thread.pyStringEval(fJobFile);

                    request += " file " + STAFUtil.wrapData(fJobFile);

                    if (!fJobFileMachine.equals(""))
                    {
                        evalAttr = "machine";

                        fJobFileMachine =
                            thread.pyStringEval(fUnevalJobFileMachine);

                        request += " machine " +
                                   STAFUtil.wrapData(fJobFileMachine);
                    }
                }
                else if (!fUnevalJobData.equals(""))
                {
                    evalElem = "job-data";
                    evalAttr = "eval";

                    if (thread.pyBoolEval(fJobDataEval))
                    {
                        evalAttr = "";

                        fJobData = thread.pyStringEval(fUnevalJobData);
                    }

                    request += " data " + STAFUtil.wrapData(fJobData);
                }

                if (!fUnevalName.equals(""))
                {
                    evalElem = "job";
                    evalAttr = "name";

                    fName = thread.pyStringEval(fUnevalName);

                    request += " jobname " + STAFUtil.wrapData(fName);
                }

                // Evaluate the clearlogs attribute

                evalElem = "job";
                evalAttr = "clearlogs";

                if (fUnevalClearlogs.equals(""))
                {
                    // Default to parent job's option if not specified
                    fClearlogs = thread.getJob().getClearLogsAsString();
                }
                else
                {
                    fClearlogs = thread.pyStringEval(fUnevalClearlogs);

                    if (fClearlogs.equalsIgnoreCase("parent"))
                    {
                        fClearlogs = thread.getJob().getClearLogsAsString();
                    }
                    else if (fClearlogs.equalsIgnoreCase("default"))
                    {
                        fClearlogs = thread.getJob().getSTAX().
                                            getClearLogsAsString();
                    }
                    else if (fClearlogs.equalsIgnoreCase("Enabled") ||
                             fClearlogs.equalsIgnoreCase("Disabled"))
                    {
                        // Do nothing - already set to Enabled or Disabled
                    }
                    else if (thread.pyBoolEval(fUnevalClearlogs))
                    {
                        fClearlogs = "Enabled";
                    }
                    else
                    {
                        fClearlogs = "Disabled";
                    }
                }

                request += " clearlogs " + fClearlogs;

                // Evaluate the monitor attribute

                evalElem = "job";
                evalAttr = "monitor";

                if (fUnevalMonitor.equals(""))
                {
                    fMonitor = "false";
                }
                else if (thread.pyBoolEval(fUnevalMonitor))
                {
                    fMonitor = "true";
                }
                else
                {
                    fMonitor = "false";
                }

                // Evaluate the logtcelapsedtime attribute

                evalElem = "job";
                evalAttr = "logtcelapsedtime";

                if (fUnevalLogTCElapsedTime.equals(""))
                {
                    // Default to parent job's option if not specified
                    fLogTCElapsedTime =
                        thread.getJob().getLogTCElapsedTimeAsString();
                }
                else
                {
                    fLogTCElapsedTime =
                        thread.pyStringEval(fUnevalLogTCElapsedTime);

                    if (fLogTCElapsedTime.equalsIgnoreCase("parent"))
                    {
                        fLogTCElapsedTime =
                            thread.getJob().getLogTCElapsedTimeAsString();
                    }
                    else if (fLogTCElapsedTime.equalsIgnoreCase("default"))
                    {
                        fLogTCElapsedTime = thread.getJob().getSTAX().
                                            getLogTCElapsedTimeAsString();
                    }
                    else if (fLogTCElapsedTime.equalsIgnoreCase("Enabled") ||
                             fLogTCElapsedTime.equalsIgnoreCase("Disabled"))
                    {
                        // Do nothing - already set to Enabled or Disabled
                    }
                    else if (thread.pyBoolEval(fUnevalLogTCElapsedTime))
                    {
                        fLogTCElapsedTime = "Enabled";
                    }
                    else
                    {
                        fLogTCElapsedTime = "Disabled";
                    }
                }

                request += " logtcelapsedtime " + fLogTCElapsedTime;

                // Evaluate the logtcnumstarts attribute

                evalElem = "job";
                evalAttr = "logtcnumstarts";

                if (fUnevalLogTCNumStarts.equals(""))
                {
                    // Default to parent job's option if not specified
                    fLogTCNumStarts =
                        thread.getJob().getLogTCNumStartsAsString();
                }
                else
                {
                    fLogTCNumStarts =
                        thread.pyStringEval(fUnevalLogTCNumStarts);

                    if (fLogTCNumStarts.equalsIgnoreCase("parent"))
                    {
                        fLogTCNumStarts =
                            thread.getJob().getLogTCNumStartsAsString();
                    }
                    else if (fLogTCNumStarts.equalsIgnoreCase("default"))
                    {
                        fLogTCNumStarts = thread.getJob().getSTAX().
                                          getLogTCNumStartsAsString();
                    }
                    else if (fLogTCNumStarts.equalsIgnoreCase("Enabled") ||
                             fLogTCNumStarts.equalsIgnoreCase("Disabled"))
                    {
                        // Do nothing - already set to Enabled or Disabled
                    }
                    else if (thread.pyBoolEval(fUnevalLogTCNumStarts))
                    {
                        fLogTCNumStarts = "Enabled";
                    }
                    else
                    {
                        fLogTCNumStarts = "Disabled";
                    }
                }

                request += " logtcnumstarts " + fLogTCNumStarts;

                // Evaluate the logtcstartstop attribute

                evalElem = "job";
                evalAttr = "logtcstartstop";

                if (fUnevalLogTCStartStop.equals(""))
                {
                    // Default to parent job's option if not specified
                    fLogTCStartStop =
                        thread.getJob().getLogTCStartStopAsString();
                }
                else
                {
                    fLogTCStartStop =
                        thread.pyStringEval(fUnevalLogTCStartStop);

                    if (fLogTCStartStop.equalsIgnoreCase("parent"))
                    {
                        fLogTCStartStop =
                            thread.getJob().getLogTCStartStopAsString();
                    }
                    else if (fLogTCStartStop.equalsIgnoreCase("default"))
                    {
                        fLogTCStartStop = thread.getJob().getSTAX().
                                          getLogTCStartStopAsString();
                    }
                    else if (fLogTCStartStop.equalsIgnoreCase("Enabled") ||
                             fLogTCStartStop.equalsIgnoreCase("Disabled"))
                    {
                        // Do nothing - already set to Enabled or Disabled
                    }
                    else if (thread.pyBoolEval(fUnevalLogTCStartStop))
                    {
                        fLogTCStartStop = "Enabled";
                    }
                    else
                    {
                        fLogTCStartStop = "Disabled";
                    }
                }

                request += " logtcstartstop " + fLogTCStartStop;

                // Evaluate the job-function element

                if (!fUnevalFunction.equals(""))
                {
                    evalElem = "job-function";
                    evalAttr = "if";

                    if (thread.pyBoolEval(fFunctionIf))
                    {
                        fFunction = thread.pyStringEval(fUnevalFunction);

                        request += " function " + STAFUtil.wrapData(fFunction);
                    }
                }

                // Evaluate the job-function-args element

                if (!fUnevalFunctionArgs.equals(""))
                {
                    evalElem = "job-function-args";
                    evalAttr = "if";

                    if (thread.pyBoolEval(fFunctionArgsIf))
                    {
                        if (thread.pyBoolEval(fFunctionArgsEval))
                        {
                            evalAttr = "";

                            fFunctionArgs =
                                thread.pyStringEval(fUnevalFunctionArgs);
                        }

                        request += " args " + STAFUtil.wrapData(fFunctionArgs);
                    }
                }

                // Evaluate the job-script elements

                evalElem = "job-script";

                for (int i = 0; i < fUnevalScripts.size(); i++)
                {
                    evalAttr = "if";
                    String scriptsIf = thread.pyStringEval(
                        (String)fScriptsIf.elementAt(i));

                    if (thread.pyBoolEval(scriptsIf))
                    {
                        evalAttr = "eval";
                        String script = new String();

                        if (thread.pyBoolEval(
                            (String)fScriptsEval.elementAt(i)))
                        {
                            evalAttr = "";

                            script = thread.pyStringEval(
                                (String)fUnevalScripts.elementAt(i));

                            fScripts.add(script);
                        }
                        else
                        {
                            script = (String)fUnevalScripts.elementAt(i);

                            fScripts.add(script);
                        }

                        request += " script " + STAFUtil.wrapData(script);
                    }
                }

                // Evaluate the job-scriptfile elements

                if (!fUnevalScriptFiles.equals(""))
                {
                    evalElem = "job-scriptfile";
                    evalAttr = "if";

                    if (thread.pyBoolEval(fScriptFilesIf))
                    {
                        evalAttr = "";

                        // Evaluate as a list; each item in the list is a
                        // single scriptfile.

                        List scriptFileList =
                            thread.pyListEval(fUnevalScriptFiles);

                        for (int j = 0; j < scriptFileList.size(); j++)
                        {
                            String scriptFile =
                                ((Object)scriptFileList.get(j)).toString();
                            fScriptFiles.add(scriptFile);
                            request += " scriptfile " +
                                       STAFUtil.wrapData(scriptFile);
                        }

                        if (!fUnevalScriptFilesMachine.equals(""))
                        {
                            evalAttr = "machine";

                            fScriptFilesMachine =
                                thread.pyStringEval(fUnevalScriptFilesMachine);

                            request += " scriptfilemachine " +
                                       STAFUtil.wrapData(fScriptFilesMachine);
                        }
                    }
                }

                // Evaluate the job-action element

                if (fJobAction != null)
                {
                    evalElem = "job-action";
                    evalAttr = "if";
                    fJobActionFlag = thread.pyBoolEval(fJobActionIf);
                }
            }
            catch (STAXPythonEvaluationException e)
            {
                fState = COMPLETE;
                fThread.popAction();

                String msg = "<job";

                if (!fUnevalName.equals(""))
                    msg += " name=\"" + fUnevalName + "\"> with ";
                else
                    msg += "> with ";

                if (!fUnevalJobFile.equals(""))
                    msg += "job-file " + fUnevalJobFile + "\n\n";
                else
                    msg += "job-data " + fUnevalJobData + "\n\n";

                if (evalElem == "job")
                {
                    if (evalAttr.equals("name"))
                        msg += "Job 'name' attribute in error";
                    else if (evalAttr.equals("clearlogs"))
                        msg += "Job 'clearlogs' attribute in error";
                    else if (evalAttr.equals("monitor"))
                        msg += "Job 'monitor' attribute in error";
                    else if (evalAttr.equals("logtcelapsedtime"))
                        msg += "Job 'logtcelapsedtime' attribute in error";
                    else if (evalAttr.equals("logtcnumstarts"))
                        msg += "Job 'logtcnumstarts' attribute in error";
                    else if (evalAttr.equals("logtcstartstop"))
                        msg += "Job 'logtcstartstop' attribute in error";
                }
                else
                {
                    if (!evalAttr.equals(""))
                        msg += "Attribute " + evalAttr +
                               " for job sub-element in error: <" + evalElem +
                               ">";
                    else
                        msg += "Job sub-element in error: <" + evalElem + ">";
                }
                
                fThread.setSignalMsgVar("STAXPythonEvalMsg", msg, e);
                fThread.raiseSignal("STAXPythonEvaluationError");

                return;
            }

            fRequest = request;

            // Create request info structure for STAX EXECUTE request

            String staxMachineName =
                fThread.getJob().getSTAX().getLocalMachineName();

            String staxServiceName =
                fThread.getJob().getSTAX().getServiceName();

            fState = WAIT_SUBJOB;

            // Get value for DiagEnabled needed for

            // Submit STAX EXECUTE Command

            STAXSubmitResult res = fThread.getJob().getSTAX().execute(
                new STAFServiceInterfaceLevel30.RequestInfo(
                    "",  // XXX: Default UUID to what?
                    staxMachineName, staxMachineName,
                    "STAF/Service/" + staxServiceName, 1, 5, true,
                    0,  // XXX: diagEnabled flag is not available
                    fRequest,
                    0,    // XXX: requestNumber field is not available
                    "none://anonymous",// Default user to "none://anonymous"
                    "local://local",  // Default endpoint to "local://local"
                    "local"           // Default physicalInterfaceID to "local"
                    ), this);

            fSubmitRC = res.getResult().rc;
            fSubmitResult = res.getResult().result;
            fSubJob = res.getJob();

            fThread.pySetVar("RC", new Integer(fSubmitRC));
            fThread.pySetVar("STAFResult", fSubmitResult);

            if (fSubmitRC == 0)
            {
                fJobID = (new Integer(fSubmitResult)).intValue();
                fThread.pySetVar("STAXSubJobID", fSubmitResult);

                // Set actual starting function and function args

                setStartFunction(fSubJob.getStartFunction());
                setStartFuncArgs(fSubJob.getStartFuncArgs());

                if ((fJobAction != null) && fJobActionFlag)
                {
                    // Start action specified by <job-action> element

                    synchronized (fThreadMap)
                    {
                        STAXThread childThread = fThread.createChildThread();

                        fThreadMap.put(new
                            Integer(childThread.getThreadNumber()),
                            childThread);

                        STAXAction childThreadAction = fJobAction;

                        childThread.addCompletionNotifiee(this);
                        childThread.pushAction(
                            childThreadAction.cloneAction());
                        childThread.schedule();
                    }
                }
            }
            else
            {
                // STAX EXECUTE Request failed

                fState = COMPLETE;
                fThread.popAction();

                String msg = "\n<job";

                if (!fUnevalName.equals(""))
                    msg += " name=\"" + fUnevalName + "\"> ";
                else
                    msg += "> ";

                msg += " request=" + fRequest + " rc=" + fSubmitRC +
                       " result=" + fSubmitResult;

                return;
            }

            // Set to the current date and time.
            fStartTimestamp = new STAXTimestamp();

            // Add the running STAX EXECUTE command to the subJobMap

            TreeMap subJobs = (TreeMap)fThread.getJob().
                               getData("subJobMap");
            synchronized (subJobs)
            {
                subJobs.put(String.valueOf(fJobID), this);
            }

            // Set Current Block Name
            try
            {
                fCurrentBlockName = fThread.pyStringEval("STAXCurrentBlock");
            }
            catch (STAXPythonEvaluationException e)
            {
                fCurrentBlockName = "";  //Shouldn't happen
            }

            // Generate a start event for a sub-job

            HashMap subJobMap = new HashMap();
            subJobMap.put("type", "subjob");
            subJobMap.put("block", fCurrentBlockName);
            subJobMap.put("status", "start");
            subJobMap.put("jobID", String.valueOf(fJobID));
            subJobMap.put("jobName", fName);

            if (!fJobFile.equals(""))
            {
                subJobMap.put("jobfile", fJobFile);
            }
            else
            {
                subJobMap.put("jobfile", STAX.INLINE_DATA);
            }

            subJobMap.put("jobfilemachine", fJobFileMachine);
            subJobMap.put("function", fStartFunction);
            subJobMap.put("functionargs", fStartFunctionArgs);
            subJobMap.put("clearlogs", fClearlogs);
            subJobMap.put("monitor", fMonitor);
            subJobMap.put("logtcelapsedtime", fLogTCElapsedTime);
            subJobMap.put("logtcnumstarts", fLogTCNumStarts);
            subJobMap.put("logtcstartstop", fLogTCStartStop);
            subJobMap.put("scriptfilesmachine", fScriptFilesMachine);

            // Convert the script file vector to a list

            List scriptFileList = new ArrayList();

            for (int i = 0; i < fScriptFiles.size(); i++)
            {
                scriptFileList.add((String)fScriptFiles.elementAt(i));
            }

            subJobMap.put("scriptFileList", STAFMarshallingContext.marshall(
                scriptFileList, new STAFMarshallingContext()));

            // Convert the script vector to a list

            List scriptList = new ArrayList();

            for (int i = 0; i < fScripts.size(); i++)
            {
                scriptList.add((String)fScripts.elementAt(i));
            }

            subJobMap.put("scriptList", STAFMarshallingContext.marshall(
                scriptList, new STAFMarshallingContext()));

            subJobMap.put("startdate",
                          fSubJob.getStartTimestamp().getDateString());
            subJobMap.put("starttime",
                          fSubJob.getStartTimestamp().getTimeString());

            fThread.getJob().generateEvent(
                STAXJobActionFactory.STAX_SUBJOB_EVENT,
                subJobMap);

            fThread.addCondition(fHoldCondition);
        }
        else if (fState == SUBJOB_COMPLETE)
        {
            synchronized(this)
            {
                fState = COMPLETE;
                fThread.popAction();

                // Remove from map of currently executing sub-jobs

                TreeMap subJobs = (TreeMap)fThread.getJob().
                    getData("subJobMap");
                 
                synchronized (subJobs)
                {
                    subJobs.remove(String.valueOf(fJobID));
                }
            }
        }
        else if (fState == COMPLETE)
        {
            // Note: We shouldn't be called in this state.
            fThread.popAction();
        }
    }

    // Note that this entire method is synchronized since the state of the
    // action can be changed on another thread (via the jobComplete method).

    public synchronized void handleCondition(STAXThread thread,
        STAXCondition cond)
    {
        // Terminate the job

        if (fState == WAIT_SUBJOB)
        {
            thread.getJob().submitAsyncForget(
                "local", fThread.getJob().getSTAX().getServiceName(),
                "TERMINATE JOB " + fJobID);
        }

        // Generate an event to indicate that the sub-job is stopped

        HashMap subJobMap = new HashMap();
        subJobMap.put("type", "subjob");
        subJobMap.put("block", fCurrentBlockName);
        subJobMap.put("status", "stop");
        subJobMap.put("jobID", String.valueOf(fJobID));
        subJobMap.put("result", fThread.getJob().getResult().toString());

        fThread.getJob().generateEvent(
            STAXJobActionFactory.STAX_SUBJOB_EVENT,
            subJobMap);

        // Remove from map of currently executing sub-jobs

        TreeMap subJobs = (TreeMap)fThread.getJob().
                           getData("subJobMap");
        synchronized (subJobs)
        {
            subJobs.remove(String.valueOf(fJobID));
        }

        fState = COMPLETE;
        thread.removeCondition(fHoldCondition);
        thread.popAction();
    }

    public STAXAction cloneAction()
    {
        STAXJobAction clone = new STAXJobAction();

        clone.fUnevalName = fUnevalName;
        clone.fUnevalClearlogs = fUnevalClearlogs;
        clone.fUnevalMonitor = fUnevalMonitor;
        clone.fUnevalLogTCElapsedTime = fUnevalLogTCElapsedTime;
        clone.fUnevalLogTCNumStarts = fUnevalLogTCNumStarts;
        clone.fUnevalLogTCStartStop = fUnevalLogTCStartStop;
        clone.fUnevalJobFile = fUnevalJobFile;
        clone.fUnevalJobFileMachine = fUnevalJobFileMachine;
        clone.fUnevalJobData = fUnevalJobData;
        clone.fUnevalFunction = fUnevalFunction;
        clone.fUnevalFunctionArgs = fUnevalFunctionArgs;
        clone.fUnevalScripts = fUnevalScripts;
        clone.fUnevalScriptFiles = fUnevalScriptFiles;
        clone.fUnevalScriptFilesMachine = fUnevalScriptFilesMachine;

        clone.fCurrentBlockName = fCurrentBlockName;
        clone.fFactory = fFactory;
        clone.fName = fName;
        clone.fClearlogs = fClearlogs;
        clone.fMonitor = fMonitor;
        clone.fJobFile = fJobFile;
        clone.fJobFileMachine = fJobFileMachine;
        clone.fJobData = fJobData;
        clone.fJobDataEval = fJobDataEval;
        clone.fFunction = fFunction;
        clone.fFunctionIf = fFunctionIf;
        clone.fFunctionArgs = fFunctionArgs;
        clone.fFunctionArgsIf = fFunctionArgsIf;
        clone.fFunctionArgsEval = fFunctionArgsEval;
        clone.fScripts = fScripts;
        clone.fScriptsIf = fScriptsIf;
        clone.fScriptsEval = fScriptsEval;
        clone.fScriptFiles = fScriptFiles;
        clone.fScriptFilesIf = fScriptFilesIf;
        clone.fScriptFilesMachine = fScriptFilesMachine;
        clone.fJobAction = fJobAction;
        clone.fJobActionIf = fJobActionIf;

        return clone;
    }

    // jobComplete is a STAXJobCompleteListener interface method
    //
    // Called when the submitted sub-job copmletes so it can set variables
    // RC and STAFResult with the result from the STAX EXECUTE request
    // and remove the hold thread condition and schedule the thread to run.
    // Generates an event to indicate that the submitted sub-job has completed.
    //
    // Note that this entire method is synchronized since its state can be
    // changed on another thread (e.g. via the execute and handleCondition
    // methods).

    public synchronized void jobComplete(STAXJob job)
    {
        if (fState == COMPLETE) return;

        // Assign STAXResult variable

        fThread.pySetVar("STAXResult", job.getResult());

        // Generate a stop event for the sub-job

        HashMap subJobMap = new HashMap();
        subJobMap.put("type", "subjob");
        subJobMap.put("block", fCurrentBlockName);
        subJobMap.put("status", "stop");
        subJobMap.put("jobID", String.valueOf(fJobID));
        subJobMap.put("result", job.getResult().toString());

        fThread.getJob().generateEvent(
            STAXJobActionFactory.STAX_SUBJOB_EVENT,
            subJobMap);

        if (fThreadMap.isEmpty())
        {
            fState = SUBJOB_COMPLETE;
            fThread.removeCondition(fHoldCondition);
            fThread.schedule();
        }
        else
        {
            fState = SUBJOB_COMPLETE_WAIT_ACTION;
        }
    }


    public synchronized void threadComplete(STAXThread thread, int endCode)
    {
        synchronized(fThreadMap)
        {
            fThreadMap.remove(new Integer(thread.getThreadNumber()));

            if (fThreadMap.isEmpty() &&
                (fState == SUBJOB_COMPLETE_WAIT_ACTION))
            {
                fState = SUBJOB_COMPLETE;
                fThread.removeCondition(fHoldCondition);
                fThread.schedule();
            }
        }
    }


    STAXThread fThread = null;
    int fState = INIT;
    private STAXHoldThreadCondition fHoldCondition = 
        new STAXHoldThreadCondition("Job");

    private String fName = new String();
    private String fClearlogs = new String();
    private String fMonitor = new String();
    private String fLogTCElapsedTime = new String();
    private String fLogTCNumStarts = new String();
    private String fLogTCStartStop = new String();
    private String fJobFile = new String();
    private String fJobFileMachine = new String();
    private String fJobData = new String();
    private String fJobDataEval = new String();
    private String fFunction = new String();
    private String fFunctionIf = new String();
    private String fFunctionArgs = new String();
    private String fFunctionArgsIf = new String();
    private String fFunctionArgsEval = new String();
    private Vector fScripts = new Vector();
    private Vector fScriptsIf = new Vector();
    private Vector fScriptsEval = new Vector();
    private Vector fScriptFiles = new Vector();
    private String fScriptFilesIf = new String();
    private String fScriptFilesMachine = new String();
    private STAXAction fJobAction = null;
    private String fJobActionIf = new String();
    private boolean fJobActionFlag = true;

    private String fUnevalName = new String();
    private String fUnevalClearlogs = new String();
    private String fUnevalMonitor = new String();
    private String fUnevalLogTCElapsedTime = new String();
    private String fUnevalLogTCNumStarts = new String();
    private String fUnevalLogTCStartStop = new String();
    private String fUnevalJobFile = new String();
    private String fUnevalJobFileMachine = new String();
    private String fUnevalJobData = new String();
    private String fUnevalFunction = new String();
    private String fUnevalFunctionArgs = new String();
    private Vector fUnevalScripts = new Vector();
    private String fUnevalScriptFiles = new String();
    private String fUnevalScriptFilesMachine = new String();
    private String fUnevalJobActionIf = new String();

    private String fStartFunction = new String();
    private String fStartFunctionArgs = new String();
    private String fRequest = new String();
    private int fRequestNumber = 0;
    private int fSubmitRC = 0;
    private String fSubmitResult = new String();
    private STAXJob fSubJob = null;
    private int fJobID = 0;
    private STAXTimestamp fStartTimestamp;
    private STAXJobActionFactory fFactory;
    private String fCurrentBlockName = new String();

    private HashMap fThreadMap = new HashMap();
}
