/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.util.Comparator;
import org.python.util.PythonInterpreter;
import org.python.core.*;
import com.ibm.staf.STAFMarshallingContext;

public class STAXThread implements STAXThreadCompleteListener
{
    // Thread states
    public static final int STATE_RUNNABLE = 0;
    public static final int STATE_RUNNING = 1;
    public static final int STATE_BLOCKED = 2;
    public static final int STATE_COMPLETE = 3;

    public static final String STATE_RUNNABLE_STRING = new String("Runnable");
    public static final String STATE_RUNNING_STRING = new String("Running");
    public static final String STATE_BLOCKED_STRING = new String("Blocked");
    public static final String STATE_COMPLETE_STRING = new String("Complete");
    public static final String STATE_UNKNOWN_STRING = new String("Unknown");

    // Completion codes
    public static final int THREAD_END_OK = 0;
    public static final int THREAD_END_DUPLICATE_SIGNAL = 1;
    public static final int THREAD_END_STOPPED_BY_PARENT = 2;

    // How many non-blocking actions may a thread execute before blocking
    public static final int MAX_NON_BLOCKING_ACTIONS = 100;

    // Turn on/off thread debugging
    private static final boolean STAX_DEBUG_THREAD = false;

    private STAXThread() { /* Do Nothing */ }

    // This creates another first-level thread for the job

    public STAXThread(STAXJob job)
    {
        fJob = job;
        fThreadNumber = job.getNextThreadNumber();

        fPythonInterpreter.set("STAXThreadID", new Integer(fThreadNumber));
        fPythonInterpreter.exec("STAXTestcaseStack = []");
        fPythonInterpreter.set("STAXCurrentTestcase", Py.None);
        fPythonInterpreter.exec("STAXBlockStack = []");
        fPythonInterpreter.set("STAXCurrentBlock", Py.None);

        // Import Java and Jython modules and classes and define a
        // cloneLocals function
        fPythonInterpreter.exec(
          "from com.ibm.staf import STAFResult as STAFRC \n" +
          "import STAFMarshalling \n" +
          "from STAFMarshalling import STAFMapClassDefinition, STAFMarshallingContext \n" +
          "import copy \n" +
          "import types \n" +
          "import org.python.core.PyFunction \n" +

          // Save the Python built-in function named type in a variable
          // named STAXBuiltinFunction_type in case a STAX xml file sets
          // the type variable to something else.  Instead of type, use
          // STAXBuiltinFunction_type in the STAXCloneGlobals function.

          "STAXBuiltinFunction_type = type \n" +

          "def STAXPythonFunction_CloneGlobals(STAXGlobals): \n" +
          "  STAXclone = copy.copy(STAXGlobals) \n" +
          "  STAXskipTypes = [types.ModuleType, types.FunctionType, \n" +
          "                   types.ClassType, \n" +
          "                   # since no StringMapType, use type(STAXclone) \n" +
          "                   STAXBuiltinFunction_type(STAXclone)] \n" +
          "  for STAXkey in STAXclone.keys(): \n" +
          "    if (STAXBuiltinFunction_type(STAXclone[STAXkey]) not in STAXskipTypes): \n" +
          "      try: \n" +
          "        STAXclone[STAXkey] = copy.deepcopy(STAXclone[STAXkey])\n" +
          "      except: \n" +
          "        pass  # ignore types that cannot be deepcopied \n" +

          // This section checks to see if the variable is a function definition
          // that was defined at the "global" scope, i.e., not within some other
          // Python module.  If so, then it replaces the function definition
          // with one that is identical except for the reference to the global
          // variable pool, which is redirected to look at the new global
          // variable pool being created.
          //
          // This was necessary, otherwise any variables created in the clone
          // are not accessable to functions defined before the clone.

          // Note, the "func_defaults or []" is necessary due to a bug in
          // Jython 2.1.  The PyFunction constructor doesn't handle a Null
          // parameter for func_defaults.  Note that it does handle func_closure
          // appropriately.

          "    elif ((STAXBuiltinFunction_type(STAXclone[STAXkey]) is types.FunctionType) and \n" +
          "          (STAXclone[STAXkey].func_globals is STAXGlobals)): \n" +
          "      STAXclone[STAXkey] = org.python.core.PyFunction(\n" +
          "          STAXclone,\n" +
          "          STAXclone[STAXkey].func_defaults or [],\n" +
          "          STAXclone[STAXkey].func_code,\n" +
          "          STAXclone[STAXkey].func_doc,\n" +
          "          STAXclone[STAXkey].func_closure)\n" +
          "  return STAXclone \n" +
          "\n" +
          "def STAXPythonFunction_FunctionExists(STAXfunction): \n" +         
          "  return STAXJob.functionExists(STAXfunction) \n" +
          "\n" +
          "class STAXUnique: \n" +
          "  def __init__(self, name): \n" +
          "    self.name = name \n" +
          "  def __str__(self): \n" +
          "    return str(self.name) \n" +
          "  def __repr__(self): \n" +
          "    return 'STAXUnique(%s)' % self.name \n" +
          "\n" +
          "STAXFunctionError = STAXUnique('STAXFunctionError') \n" +
          "\n" +
          "STAXNoResponseFromMachine = " +
          "STAXUnique('STAXNoResponseFromMachine') \n" +
          "\n" +
          "STAXFileCopyError = STAXUnique('STAXFileCopyError') \n" +
          "\n" +
          "STAXXMLParseError = STAXUnique('STAXXMLParseError') \n" +
          "\n" +
          "STAXImportModeError = STAXUnique('STAXImportModeError') \n" +
          "\n" +
          "class STAXGlobal: \n" +
          "\n\n" +
          "  # Basic customization \n" +
          "\n\n" +
          "  # constructor, optional value \n" +
          "  def __init__(self, value = None): \n" +
          "    self.data = value \n" +
          "\n" +
          "  def __del__(self): \n" +
          "    del self \n" +
          "\n" +
          "  # returns a string representation \n" +
          "  def __repr__(self): \n" +
          "    return repr(self.data) \n" +
          "\n" +
          "  def __str__(self): \n" +
          "    return str(self.data) \n" +
          "\n" +
          "  def __lt__(self, other): \n" +
          "    if self.data < other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __le__(self, other): \n" +
          "    if self.data <= other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __eq__(self, other): \n" +
          "    if self.data == other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __ne__(self, other): \n" +
          "    if self.data != other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __gt__(self, other): \n" +
          "    if self.data > other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __ge__(self, other): \n" +
          "    if self.data >= other: \n" +
          "      return true \n" +
          "    else: \n" +
          "      return false \n" +
          "\n" +
          "  def __cmp__(self, other): \n" +
          "    return self.data.cmp(other) \n" +
          "\n" +
          "  def __hash__(self): \n" +
          "    return hash(self.data) \n" +
          "\n" +
          "  # Not implementing __nonzero__ \n" +
          "\n" +
          "  # Customizing attribute access \n" +
          "\n" +
          "  def __getattr__(self, name): \n" +
          "    return getattr(self.data, name) \n" +
          "\n" +
          /* Not implementing __setattr__ and __delattr__ yet
          "  # Not sure if these are the correct implementations
          "  def __setattr__(self, name, value): \n" +
          "    self.__dict__[name] = value \n" +
          "\n" +
          "  def __delattr__(self, name): \n" +
          "    del self.__dict__[name] \n" +
          "\n" +
          */
          "  # Emulating callable objects \n" +
          "\n" +
          "  # Not implementing __call__ \n" +
          "\n" +
          "  # Emulating container types \n" +
          "\n" + 
          "  def __len__(self): \n" +
          "    return len(self.data) \n" +
          "\n" +
          "  def __getitem__(self, key): \n" +
          "    return self.data[key] \n" +
          "\n" + 
          "  def __setitem__(self, key, value): \n" +
          "    self.data[key] = value \n" +  
          "\n" +
          "  def __delitem__(self, key): \n" +
          "    del self.data[key] \n" +
          "\n" +
          "  # XXX: iter() is new in Python 2.2 so not supported yet" +
          "  # def __iter__(self): \n" +
          "  #   return iter(self.data) \n" +
          "\n" +
          "  def __contains__(self, item): \n" +
          "    return item in self.data \n" +
          "\n" +
          "  # Additional methods for emulation of sequence types \n" +
          "\n" +
          "  # Not implementing getslice, setslice, delslice - deprecated \n" +
          "\n" +
          "  # Emulating numeric types \n" +
          "  def __add__(self, other): \n" +
          "    return self.data + other \n" +
          "\n" +
          "  def __sub__(self, other): \n" +
          "    return self.data - other \n" +
          "\n" +
          "  def __mul__(self, other): \n" +
          "    return self.data * other \n" +
          "\n" +
          "  # Commented out - get SyntaxError when run \n" +
          "  # def __floordiv__(self, other): \n" +
          "  #   return self.data // other \n" +
          "\n" +
          "  def __mod__(self, other): \n" +
          "    return self.data % other \n" +
          "\n" +
          "  def __divmod__(self, other): \n" +
          "    return self.data.divmod(other) \n" +
          "\n" +
          "  def __pow__(self, other, modulo=None): \n" +
          "    return self.data.pow(other)      # ??? modulo \n" +
          "\n" +
          "  def __lshift__(self, other): \n" +
          "    return self.data << other \n" +
          "\n" +
          "  def __rshift__(self, other): \n" +
          "    return self.data >> other \n" +
          "\n" +
          "  def __and__(self, other): \n" +
          "    return self.data & other \n" +
          "\n" +
          "  def __xor__(self, other): \n" +
          "    return self.data ^ other \n" +
          "\n" +
          "  def __or__(self, other): \n" +
          "    return self.data | other \n" +
          "\n" +
          "  def __div__(self, other): \n" +
          "    return self.data / other \n" +
          "\n" +
          "  def __truediv__(self, other): \n" +
          "    return self.data / other \n" +
          "\n" + 
          "  def __radd__(self, other): \n" +
          "    return other + self.data \n" +
          "\n" + 
          "  def __rsub__(self, other): \n" +
          "    return other - self.data \n" +
          "\n" + 
          "  def __rmul__(self, other): \n" +
          "    return other * self.data \n" +
          "\n" + 
          "  def __rdiv__(self, other): \n" +
          "    return other / self.data \n" +
          "\n" + 
          "  def __rmod__(self, other): \n" +
          "    return other % self.data \n" +
          "\n" + 
          "  def __rdivmod__(self, other): \n" +
          "    return other.divmod(self.data) \n" +
          "\n" + 
          "  def __rpow__(self, other): \n" +
          "    return other.pow(self.data) \n" +
          "\n" + 
          "  def __rlshift__(self, other): \n" +
          "    return other << self.data \n" +
          "\n" + 
          "  def __rrshift__(self, other): \n" +
          "    return other >> self.data \n" +
          "\n" + 
          "  def __rand__(self, other): \n" +
          "    return other | self.data \n" +
          "\n" + 
          "  def __rxor__(self, other): \n" +
          "    return other ^ self.data \n" +
          "\n" + 
          "  def __ror__(self, other): \n" +
          "    return other | self.data \n" +
          "\n" + 
          "  def __iadd__(self, other): \n" +
          "    self.data += other \n" +
          "    return self \n" +
          "\n" + 
          "  def __isub__(self, other): \n" +
          "    self.data -= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __imul__(self, other): \n" +
          "    self.data *= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __idiv__(self, other): \n" +
          "    self.data /= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __imod__(self, other): \n" +
          "    self.data %= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __ipow__(self, other): \n" +
          "    self.data **= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __ilshift__(self, other): \n" +
          "    self.data <<= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __irshift__(self, other): \n" +
          "    self.data >>= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __iand__(self, other): \n" +
          "    self.data &= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __ixor__(self, other): \n" +
          "    self.data ^= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __ior__(self, other): \n" +
          "    self.data |= other \n" +
          "    return self \n" +
          "\n" + 
          "  def __neg__(self): \n" +
          "    self.data = -self.data \n" +
          "    return self \n" +
          "\n" + 
          "  def __pos__(self): \n" +
          "    self.data = +self.data \n" +
          "    return self \n" +
          "\n" + 
          "  def __abs__(self): \n" +
          "    self.data = abs(self.data) \n" +
          "    return self \n" +
          "\n" + 
          "  def __invert__(self): \n" +
          "    self.data = ~self.data \n" +
          "    return self \n" +
          "\n" + 
          "  def __complex__(self): \n" +
          "    return complex(self.data) \n" +
          "\n" + 
          "  def __int__(self): \n" +
          "    return int(self.data) \n" +
          "\n" + 
          "  def __long__(self): \n" +
          "    return long(self.data) \n" +
          "\n" + 
          "  def __float__(self): \n" +
          "    return float(self.data) \n" +
          "\n" + 
          "  def __oct__(self): \n" +
          "    return oct(self.data) \n" +
          "\n" + 
          "  def __hex__(self): \n" +
          "    return hex(self.data) \n" +
          "\n" + 
          "  # XXX: Currently not defining __coerce__(self, other) \n" +
          "\n" +
          "  # Providing set and get methods \n" +
          "\n" +
          "  def set(self, value): \n" +    
          "    self.data = value \n" +
          "\n" +
          "  def get(self): \n" +
          "    return self.data \n" +
          "\n" +
          "  # Implementing deepcopy so will be global \n" +
          "  def __deepcopy__(self): \n" +
          "    return self \n" +
          "\n" +
          "  # Providing an append method \n" +
          "  def append(self, other): \n" +
          "    self.data.append(other) \n" +
          "    return self \n" +
          "\n" +
          "  # Provide a stafMarshall method \n" +
          "  def stafMarshall(self, context): \n" +
          "    return STAFMarshalling.marshall(self.data, context) \n" +
          "");
        
        //Debugging:  Add following line after "  for STAXkey in ..." 
        //"      print '%s, %s' % (STAXkey, type(STAXclone[STAXkey]))\n" +

        // Get the current date and time and set as the starting date/time
        fStartTimestamp = new STAXTimestamp();
    }

    //
    // Thread functions
    //

    // This creates a child thread of this thread
    public STAXThread createChildThread()
    {
        STAXThread child = new STAXThread();

        child.fJob = fJob;
        child.fThreadNumber = fJob.getNextThreadNumber();
        child.fParent = this;
        child.fSignalHandlerMap = (HashMap)fSignalHandlerMap.clone();

        synchronized (fPythonInterpreter)
        {
            child.fPythonInterpreter = fPythonInterpreter.clonePyi();
        }

        child.fPythonInterpreter.set("STAXThreadID", 
                                     new Integer(child.fThreadNumber));

        // Get the current date and time and set as the starting date/time
        child.fStartTimestamp = new STAXTimestamp();

        child.addCompletionNotifiee(this);

        synchronized (fChildMap)
        {
            fChildMap.put(child.getThreadNumberAsInteger(), child);
        }

        return child;
    }

    // Note: When using the function below, do not try to modify the iterator
    //       passed to you, as it will be null

    public void visitChildren(STAXVisitor visitor)
    {
        synchronized (fChildMap)
        {
            Iterator iter = fChildMap.values().iterator();

            while (iter.hasNext())
                visitor.visit(iter.next(), null);
        }
    }

    public STAXThread getParentThread() { return fParent; }

    // XXX: Try to remove this function
    public TreeMap getChildrenThreads() { return fChildMap; }

    public STAXPythonInterpreter getPythonInterpreter()
    {
        synchronized(fPythonInterpreter)
        {
            return fPythonInterpreter;
        }
    }

    public void setPythonInterpreter(STAXPythonInterpreter pyi)
    {
        synchronized(fPythonInterpreter)
        {
            fPythonInterpreter = pyi;
        }
    }
    
    //
    // Miscellaneous information functions
    //

    public STAXJob getJob() { return fJob; }

    public int getThreadNumber() { return fThreadNumber; }

    public Integer getThreadNumberAsInteger()
    { return new Integer(fThreadNumber); }

    public STAXTimestamp getStartTimestamp() { return fStartTimestamp; }

    //
    // Completion notification function
    //

    public void addCompletionNotifiee(STAXThreadCompleteListener listener)
    {
        fCompletionNotifiees.addLast(listener);
    }

    //
    // Python Interpreter functions
    //

    // Set a Jython variable (e.g. RC, STAFResult) to a specified object.
 
    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                System.out.println("PySetVar PyException caught setting var=" +
                    varName + " to value=" + value + e.toString());
            } 
        }
    }

    // Execute Python code.
 
    public void pyExec(String value) throws STAXPythonEvaluationException
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                if (STAX.CACHE_PYTHON_CODE)
                {
                    PyCode codeObject = fJob.getCompiledPyCode(value, "exec");

                    synchronized(fPythonInterpreter)
                    {
                        fPythonInterpreter.exec(codeObject);
                    }
                }
                else
                {
                    fPythonInterpreter.exec(value);
                }
                
            }
            catch (PyException e)
            {
                throw new STAXPythonEvaluationException(e.toString());
            } 
        }
    }
    
    // Compile Python code (or get from cache if already compiled).
    // The combination of __builtin__.compile and the exec version, which
    // takes a code object, is used so that if Jython code is executed
    // more than once, the compiled code is cached, eliminating the need
    // to recompile the string each time it is executed.

    public PyObject pyCompile(String codeString)
    {
        if (STAX.CACHE_PYTHON_CODE)
        {
            PyCode codeObject = fJob.getCompiledPyCode(codeString, "eval");

            synchronized(fPythonInterpreter)
            {
                // XXX: Would prefer to do fPythonInterpreter.eval(code)
                // instead of the following 2 lines so that could remove 
                // assignment to STAXPyEvalResult in getCompiledPyCode,
                // but I don't know how to evaluate a PyCode object via
                // a Python Interpreter as eval only accepts a string,
                // not a PyCode object.
                fPythonInterpreter.exec(codeObject);
                return fPythonInterpreter.get("STAXPyEvalResult");
            }
        }
        else
        {
            synchronized(fPythonInterpreter)
            {
                return fPythonInterpreter.eval(codeString);
            }
        }
    }

    // Evaluate a value using Jython and return a PyObject result.
     
    public PyObject pyObjectEval(String value) 
        throws STAXPythonEvaluationException
    {
        try 
        {
            return pyCompile(value);
        }
        catch (PyException e)
        {
            throw new STAXPythonEvaluationException(
                "\nPython object evaluation failed for:\n" +
                value + "\n\n" + e.toString());
        }
    }

    // Evaluate a value using Jython and return a String result.
     
    public String pyStringEval(String value) 
        throws STAXPythonEvaluationException
    {
        try 
        {
            return pyCompile(value).__str__().toString();
        }
        catch (PyException e)
        {
            throw new STAXPythonEvaluationException(
                "\nPython string evaluation failed for:\n" +
                value + "\n\n" + e.toString());
        }
    }

    // Evaluate an expression using Jython and return a boolean result.
 
    public boolean pyBoolEval(String expression)
        throws STAXPythonEvaluationException
    {
        try
        {  
            return Py.py2boolean(pyCompile(expression));
        }
        catch (PyException e)
        {
            throw new STAXPythonEvaluationException(
                "\nPython boolean evaluation failed for:\n" +
                expression + "\n\n" + e.toString());
        }
    }

    // Evaluate a value using Jython and return an integer result.
 
    public int pyIntEval(String value) throws STAXPythonEvaluationException
    {
        try
        {  
            return Py.py2int(pyCompile(value));
        }
        catch (PyException e)
        {
            throw new STAXPythonEvaluationException(
                "\nPython integer evaluation failed for:\n" +
                value + "\n\n" + e.toString());
        }
    }

    // Evaluate a value (which evaluates to a list or tuple) using Jython, 
    // extract the Jython list/tuple into a Java List, and return a List.
    // Examples of possible values passed in:
    //   "machList", ['machA','machB'], ('machA','machB',machC'),
    //   machList[1:], machList[:-1], ['machA','machB'] + ['machC','machD']  

    public List pyListEval(String value)
        throws STAXPythonEvaluationException
    {
        List jList = new ArrayList();

        synchronized(fPythonInterpreter)
        {
            PyObject result = null;

            try
            {
                result = pyCompile(value);
            }
            catch (PyException e)
            {
                throw new STAXPythonEvaluationException(
                    "\nPython list evaluation failed for:\n" +
                    value + "\n" + e.toString());
            }
            
            // Check if value is a STAXGlobal object

            try 
            {
                if (pyBoolEval("isinstance(" + value + ", STAXGlobal)"))
                {
                    // Use the STAXGlobal class's get() method to retrieve
                    // it's contents
                    result = pyCompile(value + ".get()");
                }
            }
            catch (PyException e)
            {
                // Ignore error and assume not a STAXGlobal object
            }

            // Check if already is a Java List object

            Object javaObj = result.__tojava__(Object.class);

            if (javaObj instanceof java.util.List)
                return (List)javaObj;
            
            try
            {
                // Convert Python object to a Java ArrayList object

                fPythonInterpreter.set("STAXIterateList", result);

                if (Py.py2boolean(result))
                {
                    jList =  new ArrayList(Arrays.asList(
                        (Object[])fPythonInterpreter.get(
                            "STAXIterateList", Object[].class)));
                }
            }
            catch (PyException e)
            {
                try
                {
                    jList = new ArrayList();
                    jList.add(javaObj);
                }
                catch (PyException e2)
                {
                    throw new STAXPythonEvaluationException(
                        "\nPython list evaluation failed for:\n" +
                        value + "\n" + e.toString() + "\n" + e2.toString());
                }
            } 
        }

        return jList;
    }

    // Evaluate a value (which evaluates to a Python dictionary) using Jython, 
    // extract the contents of the PyDictionary into a Java Map, and return a
    // Map.  
    // Example of possible value that can be passed in:
    //   "{'server': 'machineA', 'testDir': 'C:/tests'}"

    public Map pyMapEval(String value) throws STAXPythonEvaluationException
    {
        // Extract a Python dictionary into a Java Map

        // Create an empty HashMap
        Map jMap = new HashMap(); 

        try
        {
            PyObject result = pyCompile(value);

            // If a null string, an empty string, or "None" is passed in,
            // return an empty map.

            if ((result != null) && !(result.toString().equals("None")) &&
                (result.toString() != ""))
            {
                if (result instanceof PyDictionary)
                { 
                    PyList pa = ((PyDictionary)result).items();

                    while (pa.__len__() != 0)
                    {
                        PyTuple po = (PyTuple)pa.pop();
                        Object first  = 
                            po.__finditem__(0).__tojava__(Object.class);
                        PyObject second = po.__finditem__(1);
                        jMap.put(first, second);
                    }
                }
                else
                {
                    throw new STAXPythonEvaluationException(
                        "\nThe following string does not evaluate " +
                        "to a PyDictionary:\n" + value);
                }
            }
        }
        catch (PyException e)
        {
            throw new STAXPythonEvaluationException(
                "\nPython dictionary evaluation failed for:\n" + value +
                "\n" + e.toString());
        }
        
        return jMap;
    }
    
    //
    // Signal functions
    //

    public void registerSignalHandler(String name,
                                      STAXSignalExecutionAction signalHandler)
    {
        fSignalHandlerMap.put(name, signalHandler);
    }

    public void raiseSignal(String name)
    {
        // Check for this signal already

        // XXX: Should also check if same signalhandler action
        if (fSignalStack.contains(name))
        {
            getJob().log(STAXJob.JOB_LOG, "error", "Duplicate signal " + name +
                         ".  Terminating job.");
            terminate(THREAD_END_DUPLICATE_SIGNAL);
            return;
        }

        STAXAction action = (STAXAction)fSignalHandlerMap.get(name);

        if (action == null)
        {
            if (name != "STAXNoSuchSignalHandler")
            {
                setSignalMsgVar("STAXNoSuchSignalHandlerMsg", 
                                "<raise signal=\"" + name + "\"/>");
                raiseSignal("STAXNoSuchSignalHandler");
            } 
            else
            {   // Should not happen - needed to break recursion
                getJob().log(STAXJob.JOB_LOG, "error", "No signal " +
                    "handler exists for signal STAXNoSuchSignalHandler");
            }
        }
        else
        {
            pushAction(action.cloneAction());
            fSignalStack.addLast(name);
        }
    }

    public void setSignalMsgVar(String signalMsgVarName, String elementInfo)
    {
        setSignalMsgVar(signalMsgVarName, elementInfo, null);
    }

    public void setSignalMsgVar(String signalMsgVarName,
                                String elementInfo,
                                STAXPythonEvaluationException pythonException)
    {
        // Generate a message containing information about why the signal
        // was raised using information provided.

        StringBuffer msg = new StringBuffer(
            "\n\n===== Element Information =====\n\n");
        msg.append(elementInfo);

        if (pythonException != null)
        {
            msg.append("\n\n===== Python Error Information =====\n\n");
            msg.append(pythonException);
            msg.append("\n");
        }
        else
        {
            msg.append("\n\n");
        }

        msg.append("===== Call Stack for STAX Thread ");
        msg.append(fThreadNumber);
        msg.append(" =====\n\n");
        msg.append(STAFMarshallingContext.formatObject(getCallStack()));

        List conditionStack = this.getConditionStack();

        if (conditionStack.size() > 0)
        {
            msg.append("\n\n==== Condition Stack for STAX Thread ");
            msg.append(fThreadNumber);
            msg.append(" =====\n\n");
            msg.append(STAFMarshallingContext.formatObject(
                getConditionStack()));
        }

        // Set the message value to the signal message variable name

        pySetVar(signalMsgVarName, msg.toString());
    }

    public void handledSignal(String name)
    {
        if ((fSignalStack.size() == 0) ||
            (!((String)fSignalStack.getLast()).equals(name)))
        {
            System.out.println("We aren't currently handling signal: " + name);
            // XXX: Not good
        }
        else fSignalStack.removeLast();
    }

    public List getCallStack()
    {
        // Generate the call stack

        List callStack = new ArrayList();

        this.visitActions(new STAXVisitorHelper(callStack)
        {
            public void visit(Object o, Iterator iter)
            {
                STAXAction action = (STAXAction)o;
                String className = action.getClass().getName();

                if (className.startsWith(STAX.PACKAGE_NAME))
                    className = className.substring(STAX.PACKAGE_NAME.length());

                String classSuffix = "Action";

                if (className.endsWith(classSuffix))
                {
                    className = className.substring(
                        0, className.length() - classSuffix.length());
                }

                List callStack = (ArrayList)fData;

                callStack.add(className + ": " + action.getInfo());
            }
        });

        return callStack;
    }

    public List getConditionStack()
    {
        // Generate the condition stack

        List conditionStack = new ArrayList();

        this.visitConditions(new STAXVisitorHelper(conditionStack)
        {
            public void visit(Object o, Iterator iter)
            {
                STAXCondition condition = (STAXCondition)o;
                String className = condition.getClass().getName();

                if (className.startsWith(STAX.PACKAGE_NAME))
                    className = className.substring(STAX.PACKAGE_NAME.length());

                String classSuffix = "Condition";

                if (className.endsWith(classSuffix))
                {
                    className = className.substring(
                        0, className.length() - classSuffix.length());
                }

                List conditionStack = (ArrayList)fData;

                conditionStack.add(
                    className + ": " + "Source=" + condition.getSource() +
                    ", Priority=" + condition.getPriority());
            }
        });

        return conditionStack;
    }
    
    //
    // State functions
    //

    public int getState()
    {
        return fState;
    }

    public String getStateAsString()
    {
        switch (fState)
        {
            case STATE_RUNNABLE:
                return STATE_RUNNABLE_STRING;
            case STATE_RUNNING:
                return STATE_RUNNING_STRING;
            case STATE_BLOCKED:
                return STATE_BLOCKED_STRING;
            case STATE_COMPLETE:
                return STATE_COMPLETE_STRING;
            default:
                return STATE_UNKNOWN_STRING;
        }
    }

    // XXX: Might we need an inheritCondition in addition to addCondition?
    //      Would this make the logic easier here and in the conditions?

    public boolean addCondition(STAXCondition condition)
    {
        synchronized (fConditionSet)
        { return fConditionSet.add(condition); }
    }

    public boolean removeCondition(STAXCondition condition)
    {
        synchronized (fConditionSet)
        { return fConditionSet.remove(condition); }
    }

    public void visitConditions(STAXVisitor visitor)
    {
        synchronized (fConditionSet)
        {
            Iterator iter = fConditionSet.iterator();

            while (iter.hasNext())
                visitor.visit(iter.next(), iter);
        }
    }

    //
    // Execution functions
    //

    public void pushAction(STAXAction action)
    {
        synchronized (fActionStack)
        { fActionStack.addLast(action); }
    }

    public void popAction()
    {
        synchronized (fActionStack)
        { fActionStack.removeLast(); }
    }

    public void visitActions(STAXVisitor visitor)
    {
        synchronized (fActionStack)
        {
            Iterator iter = fActionStack.iterator();

            while (iter.hasNext())
                visitor.visit(iter.next(), iter);
        }
    }

    public void schedule()
    {
        synchronized (fConditionSet)
        {
            if (fState == STATE_BLOCKED)
            {
                fState = STATE_RUNNABLE;
                fJob.getSTAX().getThreadQueue().add(this);
            }
        }
    }

    public void execute()
    {
        int numActions = 0;

        fState = STATE_RUNNING;

        while (true)
        {
            if (++numActions == MAX_NON_BLOCKING_ACTIONS)
            {
                synchronized (fConditionSet)
                {
                    fState = STATE_BLOCKED;
                    schedule();
                    return;
                }
            }

            if (fActionStack.size() == 0)
            {
                synchronized (fChildMap)
                {
                    synchronized (fConditionSet)
                    {
                        if (fChildMap.size() != 0)
                        {
                            fState = STATE_BLOCKED;
                            return;
                        }
                    }
                }

                fState = STATE_COMPLETE;

                synchronized (fConditionSet)
                {
                    boolean addedParentCondition = false;
                    Iterator iter = fConditionSet.iterator();

                    while (iter.hasNext())
                    {
                        STAXCondition cond = (STAXCondition)iter.next();

                        // XXX: We need to get the parent running again so
                        //      that it may kill any children threads if
                        //      necessary
                        // XXX: Is it safe to schedule the parent?
                        // XXX: Should addCondition call schedule()?

                        if ((fParent == null) && cond.isInheritable())
                        {
                            fJob.log(STAXJob.JOB_LOG, "error", "Unhandled " +
                                     "condition of type " +
                                     cond.getClass().getName() + " found at " +
                                     "end of job");
                        }
                        else if (cond.isInheritable())
                        {
                            fParent.addCondition(cond);
                            addedParentCondition = true;
                        }
                    }

                    if (addedParentCondition) fParent.schedule();
                }

                if (STAX_DEBUG_THREAD)
                {
                    System.out.println("Thread #" + fThreadNumber +
                                       ": complete");
                }

                while (!fCompletionNotifiees.isEmpty())
                {
                    STAXThreadCompleteListener listener =
                        (STAXThreadCompleteListener)
                        fCompletionNotifiees.removeFirst();

                    if (listener != null)
                        listener.threadComplete(this, fCompletionCode);
                }

                return;
            }

            STAXAction action = (STAXAction)fActionStack.getLast();
            STAXCondition currCondition = null;

            synchronized (fConditionSet)
            {
                if (fConditionSet.size() != 0)
                {
                    currCondition = (STAXCondition)fConditionSet.first();

                }

                if ((currCondition != null) &&
                    ((currCondition instanceof STAXHoldThreadCondition) ||
                     (currCondition instanceof STAXHardHoldThreadCondition)))
                {
                    if (STAX_DEBUG_THREAD)
                    {
                        String condName = currCondition.getClass().getName();
                        condName = condName.substring(
                                   condName.lastIndexOf(".") + 1);

                        System.out.println("Thread #" + fThreadNumber +
                                           ": blocking due to " + condName);
                    }

                    fState = STATE_BLOCKED;
                    return;
                }
            }

            if (currCondition == null)
            {
                if (STAX_DEBUG_THREAD)
                {
                    String actionName = action.getClass().getName();
                    actionName = actionName.substring(
                                 actionName.lastIndexOf(".") + 1);

                    System.out.println("Thread #" + fThreadNumber +
                                       ": calling " + actionName);
                }

                action.execute(this);
            }
            else
            {
                if (STAX_DEBUG_THREAD)
                {
                    String actionName = action.getClass().getName();
                    actionName = actionName.substring(
                                 actionName.lastIndexOf(".") + 1);
                    String condName = currCondition.getClass().getName();
                    condName = condName.substring(
                               condName.lastIndexOf(".") + 1);

                    System.out.println("Thread #" + fThreadNumber +
                                       ": calling " + actionName +
                                       ", handling " + condName);
                }

                action.handleCondition(this, currCondition);
            }

        }  // end while
    }

    public void terminate(int termCode)
    {
        synchronized (fConditionSet)
        {
            fCompletionCode = termCode;
            addCondition(new STAXTerminateThreadCondition("Thread"));
            schedule();
        }
    }

    // STAXThreadCompleteListener method

    public void threadComplete(STAXThread thread, int endCode)
    {
        synchronized(fChildMap)
        {
            fChildMap.remove(thread.getThreadNumberAsInteger());

            if (fChildMap.size() == 0)
                schedule();
        }
    }

    // This class is used to sort the conditions in the condition set

    class STAXConditionComparator implements Comparator
    {
        public int compare(Object o1, Object o2)
        {
            STAXCondition c1 = (STAXCondition)o1;
            STAXCondition c2 = (STAXCondition)o2;

            if (c1.getPriority() == c2.getPriority())
            {
                if (o1.hashCode() == o2.hashCode()) return 0;
                else if (o1.hashCode() < o2.hashCode()) return -1;
            }
            else if (c1.getPriority() < c2.getPriority()) return -1;

            return 1;
        }
    }

    private int fState = STATE_BLOCKED;
    private int fCompletionCode = THREAD_END_OK;
    private String fTerminationBlock = new String();
    private STAXJob fJob = null;
    private int fThreadNumber = 0;
    private STAXThread fParent = null;
    private HashMap fSignalHandlerMap = new HashMap();
    private STAXPythonInterpreter fPythonInterpreter =
                                  new STAXPythonInterpreter();
    private LinkedList fActionStack = new LinkedList();
    private LinkedList fSignalStack = new LinkedList();
    private TreeSet fHeldBlocksSet = new TreeSet();
    private TreeSet fConditionSet = new TreeSet(new STAXConditionComparator());
    private LinkedList fCompletionNotifiees = new LinkedList();
    private TreeMap fChildMap = new TreeMap();
    private STAXTimestamp fStartTimestamp;
}
