
/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFHandleManager.h"
#include "STAFProcess.h"
#include "STAFTrace.h"
#include "STAFServiceManager.h"
#include "STAFProcUtil.h"
#include "STAFHandleService.h"
#include "STAFDiagManager.h"
#include "STAFConnectionManager.h"
#include "STAFThreadManager.h"

unsigned int HandleMonitorThread(void *);

static STAFHandleManager *pHandleManager = 0;
static STAFThreadManager &getHandleThreadManager()
{
  static STAFThreadManager theThreadManager(1);
  return theThreadManager;
}

STAFHandleManager::STAFHandleManager() : fNextHandle(1)
{
    pHandleManager = this;

    getHandleThreadManager().dispatch(HandleMonitorThread, 0);
}


void STAFHandleManager::processTerminated(STAFProcessID_t pid,
                                           STAFProcessHandle_t procHandle,
                                           unsigned int rc, void *)
{
    pHandleManager->handleProcessTerminated(pid, procHandle);
}


void STAFHandleManager::handleProcessTerminated(STAFProcessID_t pid,
                                                STAFProcessHandle_t procHandle)
{
    // XXX Do we need a mutexsemlock here?
    HandleList::iterator handleIter;

    for(handleIter = fHandleList.begin();
        handleIter != fHandleList.end();
        ++handleIter)
    {
        if ((handleIter->second.pid == pid) &&
            (handleIter->second.procHandle == procHandle) &&
            (handleIter->second.state == kRegistered))
        {
            STAFHandleService::getInstance()->handleRemoved(handleIter->second.handle);
        }
    }

    // Basically, we loop through the list finding a registered handle with
    // the specified pid and procHandle.  We only have to worry about registered
    // handles here.  Pending handles are removed by the Process service, and we
    // never have static or in-process handles monitored by the ProcessManager.

    STAFMutexSemLock handleLock(fHandleListSem);

    for(handleIter = fHandleList.begin();
        (handleIter != fHandleList.end()) &&
            ((handleIter->second.pid != pid) ||
             (handleIter->second.procHandle != procHandle) ||
             (handleIter->second.state != kRegistered));
        ++handleIter)
    { /* Do Nothing */ }

    if (handleIter != fHandleList.end())
    {
        fHandleList.erase(handleIter);
    }
}


STAFString STAFHandleManager::name(STAFHandle_t handle)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        return fHandleList[handle].name;

    return STAFString();
}


STAFRC_t STAFHandleManager::variablePool(STAFHandle_t handle,
                                         STAFVariablePoolPtr &pool)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        pool = fHandleList[handle].variablePoolPtr;
    else rc = kSTAFHandleDoesNotExist;

    return rc;
}


STAFRC_t STAFHandleManager::handleQueue(STAFHandle_t handle,
                                        STAFHandleQueuePtr &queue)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        queue = fHandleList[handle].handleQueuePtr;
    else rc = kSTAFHandleDoesNotExist;
    
    return rc;
}


std::vector<STAFHandle_t> STAFHandleManager::handlesWithName(STAFString name)
{
    name.lowerCase();

    std::vector<STAFHandle_t> handles;
    STAFMutexSemLock handleLock(fHandleListSem);
    HandleList::iterator handleIter;

    // XXX: This should be able to be done with some form of copy_if

    for(handleIter = fHandleList.begin(); handleIter != fHandleList.end();
        ++handleIter)
    {
        if (name == handleIter->second.name.toLowerCase())
            handles.push_back(handleIter->second.handle);
    }

    return handles;
}

STAFRC_t STAFHandleManager::updateTimestamp(STAFHandle_t handle,
                                            STAFProcessID_t pid)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
        return kSTAFHandleDoesNotExist;

    HandleData &theHandle = fHandleList[handle];

    if ((theHandle.state != kStatic) && (theHandle.pid != pid))
        return kSTAFInvalidHandle;

    theHandle.lastUsedTimestamp = STAFTimestamp::now();

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetPendingHandle(STAFHandle_t &handle,
    STAFProcessID_t pid, STAFProcessHandle_t procHandle,
    STAFVariablePoolPtr pool)
{
    // Note: This method should only be used by STAFProcessService.  It will
    //       obtain a lock on our fHandleListSem before calling this method.
    //       Should anyone else need to call this method, they must also lock
    //       our fHandleListSem before calling this method.

    HandleData theHandle(fNextHandle++, pid, procHandle, kPending,
                         STAFTimestamp::now(), STAFString(), pool);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetStaticHandle(STAFHandle_t &handle,
                                                  const STAFString &name)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    HandleData theHandle(fNextHandle++, 0, 0, kStatic,
                         STAFTimestamp::now(), name);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetStaticHandle(STAFHandle_t &handle,
                                                  const STAFString &name,
                                                  STAFVariablePoolPtr pool)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    HandleData theHandle(fNextHandle++, 0, 0, kStatic,
                         STAFTimestamp::now(), name, pool);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}

STAFServiceResult STAFHandleManager::addNotification(
    STAFHandle_t handle, const STAFString machine, const STAFString uuid,
    const STAFString service, const STAFString key)
{
    STAFServiceResult serviceResult(kSTAFOk);

    if (isLocalMachine(machine, 1))
    {
        addNotificationEntry(handle, machine, uuid, service, key);
        return serviceResult;
    }
    else
    {
       // Remote machine

        STAFRC_t rc = kSTAFOk;
        STAFConnectionPtr connection;
        STAFString result;

        rc = gConnectionManagerPtr->makeConnection(
            machine, connection, result);

        if (rc)
        {
            serviceResult.fRC = rc;
            serviceResult.fResult = result;
            return serviceResult;
        }
        
        connection->writeUInt(kSTAFHandleTerminationRegistrationAPI);
        connection->writeUInt(0);    // Dummy level

        STAFRC_t ack = connection->readUInt();

        if (ack != kSTAFOk)
        {
            serviceResult.fRC = ack;
            serviceResult.fResult = "Error connecting "
                "kSTAFHandleTerminationRegistrationAPI";
            return serviceResult;
        }
        
        // min and max levels
        connection->writeUInt(1);
        connection->writeUInt(1);
        
        connection->writeUInt(handle);
        connection->writeString(gMachinePtr->toLowerCase());
        connection->writeString(uuid);
        connection->writeString(service);
        connection->writeString(key);
        
        addPollingEntry(handle, machine, uuid, service, key);

        return serviceResult;
    }
}


STAFRC_t STAFHandleManager::addNotificationEntry(STAFHandle_t &handle,
    const STAFString &machine, const STAFString &uuid,
    const STAFString &service, const STAFString &key)
{
    STAFMutexSemLock notificationLock(fNotificationListSem);
    
    // Make sure the handle/machine/service combination 
    // is not already registered
    
    STAFHandleManager::NotificationList::iterator iter;
    
    for (iter = fNotificationList.begin(); 
         iter != fNotificationList.end(); iter++)
    {
        STAFHandleManager::NotificationData notifiee(*iter);
            
        if ((notifiee.handle == handle) && 
            (notifiee.machine == machine) &&
            (notifiee.uuid == uuid) &&
            (notifiee.notifyService == service))
        {
            return kSTAFOk;
        }
    }

    NotificationData theNotification(handle, machine, uuid, key, service);
    
    fNotificationList.push_back(theNotification);

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::deleteNotification(STAFHandle_t handle,
                                               const STAFString machine,
                                               const STAFString uuid,
                                               const STAFString service)
{
    if (debug) // XXX
        cout << "Deleting notification for handle " << handle << "\n";
    STAFMutexSemLock notificationLock(fNotificationListSem);
    
    NotificationList::iterator notificationIter;
    
    for(notificationIter = fNotificationList.begin();
        (notificationIter != fNotificationList.end()) &&
        (notificationIter->handle != handle) &&
        (notificationIter->machine != machine) &&
        (notificationIter->uuid != uuid);
        ++notificationIter)
    { /* Do Nothing */}

    if (notificationIter != fNotificationList.end() &&
        (notificationIter->handle == handle) &&
        (notificationIter->machine == machine) &&
        (notificationIter->uuid == uuid))
    {
        fNotificationList.erase(notificationIter);
    }

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addPollingEntry(STAFHandle_t &handle,
    const STAFString &machine, const STAFString &uuid,
    const STAFString &service, const STAFString &key)
{
    STAFMutexSemLock pollingLock(fPollingDataListSem);
    
    // Make sure the handle/machine/service combination 
    // is not already registered
    
    STAFHandleManager::PollingDataList::iterator iter;
    
    for (iter = fPollingDataList.begin(); 
         iter != fPollingDataList.end(); iter++)
    {
        STAFHandleManager::PollingData notifiee(*iter);
            
        if ((notifiee.handle == handle) && 
            (notifiee.machine == machine) &&
            (notifiee.uuid == uuid) &&
            (notifiee.notifyService == service))
        {
            return kSTAFOk;
        }
    }

    PollingData theNotification(handle, machine, uuid, key, service);
    
    fPollingDataList.push_back(theNotification);

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::deletePolling(STAFHandle_t handle,
                                          const STAFString machine,
                                          const STAFString uuid,
                                          const STAFString service)
{
    if (debug) // XXX
        cout << "Deleting polling for handle " << handle << "\n";
    STAFMutexSemLock pollingLock(fPollingDataListSem);
    
    PollingDataList::iterator pollingIter;
    
    for (pollingIter = fPollingDataList.begin();
        (pollingIter != fPollingDataList.end()) &&
        (pollingIter->handle != handle) &&
        (pollingIter->machine != machine) &&
        (pollingIter->uuid != uuid);
        ++pollingIter)
    { /* Do Nothing */}

    if (pollingIter != fPollingDataList.end() &&
        (pollingIter->handle == handle) &&
        (pollingIter->machine == machine) &&
        (pollingIter->uuid == uuid))
    {
        fPollingDataList.erase(pollingIter);
    }

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::registerHandle(STAFHandle_t &handle,
                                           STAFProcessID_t pid,
                                           const STAFString &name)
{
    STAFString lowerName(name.toLowerCase());
    STAFMutexSemLock handleLock(fHandleListSem);
    HandleList::iterator handleIter;
    STAFProcessHandle_t procHandle = 0;

    for (handleIter = fHandleList.begin(); handleIter != fHandleList.end();
         ++handleIter)
    {

        if (handleIter->second.pid != pid) continue;

        if (pid == gSTAFProcPID)
        {
            if (lowerName == handleIter->second.lowerName)
            {
                // This is an existing kInProcess handle

                if (handleIter->second.state != kInProcess)
                {
                    STAFTrace::trace(kSTAFTraceError,
                                     "In STAFHandleManager::registerHandle(), "
                                     "found handle with STAFProc PID, but not "
                                     "state == kInProcess");
                }

                handle = handleIter->second.handle;

                return kSTAFOk;
            }
            else continue;
        }

        if (!STAFProcess::isRunning(handleIter->second.procHandle))
        {
            // This is a program that is no longer running, but which has not
            // been removed from the list yet.  Either our callback hasn't
            // been called for a regular kRegistered process, or the
            // ProcessManager has not told us to remove the kPending(Registered)
            // handle.  Either way, this handle is not a match.

            continue;
        }

        if (handleIter->second.state == kPending)
        {
            // This is an existing kPending handle, and the process is
            // just now registering.

            handle = handleIter->second.handle;
            handleIter->second.name = name;
            handleIter->second.state = kPendingRegistered;
            handleIter->second.lastUsedTimestamp = STAFTimestamp::now();

            return kSTAFOk;
        }

        procHandle = handleIter->second.procHandle;
    }

    // This is a valid new handle so register it

    if (procHandle == 0)
    {
        unsigned int osRC = 0;
        STAFRC_t rc = STAFProcess::getProcessHandleFromID(
                      pid, procHandle, osRC);
    }

    if (procHandle == 0)
    {
        STAFTrace::trace(kSTAFTraceError,
                         "In STAFHandleManager::registerHandle: received zero "
                         " process handle from getProcessHandleFromID()");
    }

    HandleData theHandle(fNextHandle++, pid, procHandle,
                         (pid == gSTAFProcPID) ? kInProcess : kRegistered,
                         STAFTimestamp::now(), name);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    if (pid == gSTAFProcPID) return kSTAFOk;

    STAFProcessEndCallbackLevel1 callback = { processTerminated, 0 };

    return STAFProcess::registerForProcessTermination(pid,
           theHandle.procHandle, callback);
}


STAFRC_t STAFHandleManager::unRegister(STAFHandle_t handle,
                                       STAFProcessID_t pid)
{
    STAFRC_t rc = kSTAFOk;
    STAFHandle_t unRegHandle;
    
    {
        STAFMutexSemLock handleLock(fHandleListSem);
    
        if (fHandleList.find(handle) == fHandleList.end())
        {
            rc = kSTAFHandleDoesNotExist;
            return rc;
        }
        else
        {
            HandleData &theHandle = fHandleList[handle];
            if (debug) // XXX
                cout << "Unregistering handle " << theHandle.handle << "\n";
            unRegHandle = theHandle.handle;

            if (theHandle.pid != pid)
            {
                rc = kSTAFInvalidHandle;
                return rc;
            }
            else if (theHandle.state == kPendingRegistered)
            {
                theHandle.state = kPending;
                return rc;
            }
            else if ((theHandle.state == kRegistered) ||
                     (theHandle.state == kInProcess))
                fHandleList.erase(handle);
        }
    }
    
    STAFHandleService::getInstance()->handleRemoved(unRegHandle);
    
    STAFMutexSemLock handleLock(fNotificationListSem);
    
    STAFHandleManager::NotificationList::iterator iter;
    STAFString result;
    
    bool moreHandles = true;
    if (debug) // XXX
        cout << "Removing notifications for handle " << unRegHandle << "\n";
    
    while (moreHandles)
    {
        for (iter = fNotificationList.begin(); 
             iter != fNotificationList.end()  &&
             (iter->handle != unRegHandle)
             ; iter++)
        {
        }
            
        if (iter != fNotificationList.end())
        {
            fNotificationList.erase(iter);
        }
        else
        {
            moreHandles = false;
        }
    }

    return rc;
}


STAFRC_t STAFHandleManager::removePendingHandle(STAFHandle_t handle,
                                                STAFProcessID_t pid)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        rc = kSTAFHandleDoesNotExist;
    }
    else
    {
        const HandleData &theHandle = fHandleList[handle];

        if (theHandle.pid != pid)
        {
            rc = kSTAFInvalidHandle;
        }
        else if ((theHandle.state == kPending) ||
                 (theHandle.state == kPendingRegistered))
        {
            fHandleList.erase(handle);
        }
        else
        {
            rc = kSTAFInvalidHandle;
        }
    }

    return rc;
}


STAFRC_t STAFHandleManager::removeStaticHandle(STAFHandle_t handle)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        rc = kSTAFHandleDoesNotExist;
    }
    else
    {
        const HandleData &theHandle = fHandleList[handle];

        if (theHandle.state == kStatic)
            fHandleList.erase(handle);
        else
            rc = kSTAFInvalidHandle;
    }

    return rc;
}


STAFHandleManager::HandleList STAFHandleManager::getHandleListCopy()
{
    STAFMutexSemLock handleLock(fHandleListSem);

    return fHandleList;
}


STAFHandleManager::NotificationList STAFHandleManager::getNotificationListCopy()
{
    STAFMutexSemLock handleLock(fNotificationListSem);

    return fNotificationList;
}


STAFHandleManager::PollingDataList STAFHandleManager::getPollingDataListCopy()
{
    STAFMutexSemLock handleLock(fPollingDataListSem);

    return fPollingDataList;
}


STAFMutexSem &STAFHandleManager::getHandleManagerSem()
{
    return fHandleListSem;
}


// STAFHandleManager::authenticate
//
// Perform Authentication for the handle.  If credentials are input, check if
// the handle is already authenticated, and if so, return an error.  If the
// specified Authenticator service is not registered, return an error if
// credentials are input; if authentication data is input, set authenticator to 
// "none" and set userIdentifier to "anonymous". If the specified Authenticator
// is found and the authenticationType is kData (indicating a remote system),
// check if there is cached authentication information for this remote
// machine's handle.  If so, and the cached authentication information is the
// same, then don't need to re-authenticate.  Otherwise, submit an
// AUTHENTICATE request to the Authenticator service, passing the specified 
// authentication data or credentials.  If the AUTHENTICATE request is
// successful and credentials were input, return authentication data in the 
// result.
//
// Parameters:
//   machine - machine name (Input)
//   handle  - handle (Input)
//   authenticator - name of authenticator service (Input/Output)
//   userIdentifier - name of the user (Input/Output)
//   authenticateType - authentication data type, kCredentials or kData (Input)
//   authenticationData - authentication credentials or data (Input)
//
// Returns:
//   rc 0 if authenticate was successful and if credentials passed
//   in via authenticationData, returns authentication data in result.
//   Non-zero rc if authenticate failed and any additional info about the
//   failure in the result.

STAFServiceResult STAFHandleManager::authenticate(
                                     const STAFString &machine,
                                     const STAFHandle_t handle,
                                     STAFString &authenticator,
                                     STAFString &userIdentifier,
                                     const AuthenticateType authenticateType,
                                     const STAFString &authenticationData)
{
    STAFRC_t rc = kSTAFOk;

    if (authenticateType == kCredentials)
    {
        // Check if handle is already authenticated.  If so, return an error.

        rc = isAuthenticated(handle);
        
        if (rc != kSTAFOk)
        {
            return STAFServiceResult(rc);
        }
    }

    // Check if the specified authenticator service is registered

    STAFServiceResult serviceResult(kSTAFOk);
    STAFServicePtr service;

    rc = gServiceManagerPtr->getAuthenticator(authenticator, service);

    if (rc != kSTAFOk)
    {
        // The authenticator service is not registered

        if (authenticateType == kCredentials)
        {
            // Return an error

            return STAFServiceResult(rc, authenticator);
        }
        else
        {   // Authentication data was passed in.
            // Don't return an error.  Set the authenticator to "none"
            // and the userIdentifier to "anonymous". 

            authenticator = gNoneString;
            userIdentifier = gAnonymousString;

            return serviceResult;
        }
    }

    if (authenticateType == kData)
    {
        // Get cached authentication information, if any, for the remote
        // machine's handle.

        STAFString theKey = machine.toLowerCase() + ";" + STAFString(handle);

        STAFMutexSemLock remoteHandle(fRemoteHandleMapSem);

        // If authentication information for remote machine/handle has been
        // cached, if it's the same, then there's no need to re-authenticate.

        if (fRemoteAuthHandleMap.find(theKey) != fRemoteAuthHandleMap.end())
        {
            AuthenticationInfo &authInfo = fRemoteAuthHandleMap[theKey];

            if ((authInfo.authenticator == authenticator.toUpperCase()) &&
                (authInfo.userIdentifier == userIdentifier) &&
                (authInfo.authenticationData == authenticationData))
            {   
                return serviceResult;
            }
        }
    }

    // Submit an Authenticate request to the HANDLE service

    STAFServiceRequest authRequestInfo;

    authRequestInfo.fMachine = "local"; // Must be issued from local system
    authRequestInfo.fMachineNickname = *gMachineNicknamePtr;
    authRequestInfo.fHandle = gSTAFProcHandle; // A handle on the local system
    authRequestInfo.fHandleName = gHandleManagerPtr->name(gSTAFProcHandle);
    authRequestInfo.fRequest = "AUTHENTICATE USER " + userIdentifier;
    authRequestInfo.fRequestNumber = 0;
    authRequestInfo.fDiagEnabled = gDiagManagerPtr->getEnabled();
    authRequestInfo.fTrustLevel = 5;    // Authenticate requires trust level 5
    authRequestInfo.fAuthenticator = gNoneString;
    authRequestInfo.fUserIdentifier = gAnonymousString;
    authRequestInfo.fUser = gUnauthenticatedUser;
    authRequestInfo.fInterface = "local";
    authRequestInfo.fLogicalInterfaceID = "local";
    authRequestInfo.fPhysicalInterfaceID = "local";
    authRequestInfo.fEndpoint = "local" + gSpecSeparator + "local";

    gHandleManagerPtr->variablePool(authRequestInfo.fHandle, 
                                    authRequestInfo.fRequestVarPool);

    authRequestInfo.fSourceSharedVarPool =
        STAFVariablePoolPtr(new STAFVariablePool, STAFVariablePoolPtr::INIT);
    authRequestInfo.fLocalSharedVarPool = *gSharedVariablePoolPtr;
    authRequestInfo.fLocalSystemVarPool = *gGlobalVariablePoolPtr;

    if (authenticateType == kCredentials)
        authRequestInfo.fRequest += " CREDENTIALS " + authenticationData;
    else
        authRequestInfo.fRequest += " DATA " + authenticationData;

    serviceResult = service->submitRequest(authRequestInfo);

    if (serviceResult.fRC == kSTAFOk)
    {
        // User was successfully authenticated

        STAFString authData = authenticationData;

        if (authenticateType == kCredentials)
        {
            // Assign authentication data returned by the Authenticator service
            authData = serviceResult.fResult;

            // Set handle variable STAF/Handle/User=authenticator://user

            STAFVariablePoolPtr handlePool;

            rc = variablePool(handle, handlePool);
            handlePool->set("STAF/Handle/User",
                            authenticator + gSpecSeparator + userIdentifier);
        }
        
        rc = gHandleManagerPtr->cacheAuthenticationInfo(machine, handle,
             authenticator, userIdentifier, authData);
    }

    return serviceResult;
}


// STAFHandleManager::unAuthenticate
//
// Unauthenticate the local handle
//
// Parameters:
//   handle  - handle (Input)
//
// Returns:
//   rc 0 if successfully un-authenticate the handle
//   Non-zero rc if could not un-authenticate the handle and the handle
//       in the result.

STAFServiceResult STAFHandleManager::unAuthenticate(const STAFHandle_t handle)
{
    {
        STAFMutexSemLock handleLock(fHandleListSem);

        if (fHandleList.find(handle) == fHandleList.end())
        {
            return STAFServiceResult(kSTAFHandleDoesNotExist, handle);
        }

        HandleData &theHandle = fHandleList[handle];

        theHandle.authenticator = gNoneString;
        theHandle.userIdentifier = gAnonymousString;
        theHandle.authenticationData = "";
    }

    // Set handle variable STAF/Handle/User=none://anonymous
    
    STAFVariablePoolPtr handlePool;

    STAFRC_t rc = variablePool(handle, handlePool);

    if (rc == kSTAFOk)
        handlePool->set("STAF/Handle/User", gUnauthenticatedUser);

    return STAFServiceResult(kSTAFOk);
}


// STAFHandleManager::cacheAuthenticationInfo
//
// Cache authentication information for the machine's handle.  If the
// machine is local, store the authentication information in the
// local Handle List.  If the machine is remote, store the authentication
// information in the Remote Authenticated Handle Map. 
//
// Parameters:
//   machine - machine name (Input)
//   handle  - handle (Input)
//   authenticator - name of authenticator service (Input)
//   userIdentifier - name of the user (Input)
//   authenticationData - authentication data (Input)
//
// Returns:
//   rc 0 if authentication information was cached successfully
//   Non-zero rc if authentication information was not cached successfully

STAFRC_t STAFHandleManager::cacheAuthenticationInfo(
                            const STAFString &machine,
                            const STAFHandle_t handle,
                            const STAFString &authenticator,
                            const STAFString &userIdentifier,
                            const STAFString &authenticationData)
{
    if (isLocalMachine(machine, 1))
    {
        STAFMutexSemLock handleLock(fHandleListSem);
        
        if (fHandleList.find(handle) == fHandleList.end())
        {
            // Should never happen
            return kSTAFHandleDoesNotExist;
        }

        HandleData &theHandle = fHandleList[handle];

        theHandle.authenticator = authenticator;
        theHandle.userIdentifier = userIdentifier;
        theHandle.authenticationData = authenticationData;
    }
    else
    {
        // Cache the authentication information for the remote machine/handle
        // in the fRemoteAuthHandleMap.

        STAFString theKey = machine.toLowerCase() + ";" + STAFString(handle);

        AuthenticationInfo authInfo(authenticator, userIdentifier,
                                    authenticationData);

        STAFMutexSemLock remoteHandleLock(fRemoteHandleMapSem);

        fRemoteAuthHandleMap[theKey] = authInfo;

        // XXX: Register for garbage collection so that when handle or machine
        // is unregistered, it can be removed from the fRemoteAuthHandleMap.
    }

    return kSTAFOk;
}


// STAFHandleManager::isAuthenticated
//
// Check if the local handle is authenticated.
//
// Parameters:
//   handle  - handle (Input)
//
// Returns:
//   0 if the local handle is not authenticated.
//   kSTAFHandleAlreadyAuthenticated if the handle is already authenticated.
//   kSTAFHandleDoesNotExist if the local handle does not exist.

STAFRC_t STAFHandleManager::isAuthenticated(
                            const STAFHandle_t handle)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        // Should never happen
        return kSTAFHandleDoesNotExist;
    }

    HandleData &theHandle = fHandleList[handle];

    if (theHandle.authenticator == gNoneString)
        return kSTAFOk;
    else
        return kSTAFHandleAlreadyAuthenticated;
}


// STAFHandleManager::getAuthenticationInfo
//
// Get the authentication information for a local handle.
//
// Parameters:
//   handle  - handle (Input)
//   authenticator - name of authenticator service (Output)
//   userIdentifier - name of the user (Output)
//   authenticationData - authentication data (Output)
//
// Returns:
//   rc 0 if authentication info for a local handle was retrieved successfully
//   Non-zero rc if could not get authentication info for a local handle

STAFRC_t STAFHandleManager::getAuthenticationInfo(
                            const STAFHandle_t handle,
                            STAFString &authenticator,
                            STAFString &userIdentifier,
                            STAFString &authenticationData)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        // Should never happen
        authenticator = gNoneString;
        userIdentifier = gAnonymousString;
        return kSTAFHandleDoesNotExist;
    }

    HandleData &theHandle = fHandleList[handle];

    authenticator = theHandle.authenticator;
    userIdentifier = theHandle.userIdentifier;
    authenticationData = theHandle.authenticationData;

    return kSTAFOk;
}

void STAFHandleManager::gcPolling()
{
    int interval = 60000; // XXX add an operational parameter to override

    gThreadManagerPtr->sleepCurrentThread(5000);
    
    try
    {
         while (1)
         {
             gGCPollingSem->wait(interval);
             
             if (!gContinueGCPolling)
                 return;
             
             PollingDataList pollingDataList = getPollingDataListCopy();

             PollingDataList::iterator iter;

             for (iter = pollingDataList.begin();
                  iter != pollingDataList.end(); iter++)
             {
                 PollingData notifiee(*iter);
                 STAFString machine = notifiee.machine;
                 
                 STAFResultPtr result = 
                     gSTAFProcHandlePtr->submit(machine,
                                                "PING",
                                                "PING");
                 if (result->rc != kSTAFOk)
                 {
                     STAFString request = "STAF_CALLBACK HANDLEDELETED HANDLE ";
                     request += notifiee.handle;
                     request += " MACHINE ";
                     request += machine;
                     request += " UUID ";
                     request += notifiee.uuid;
                     request += " KEY ";
                     request += notifiee.key;

                     STAFResultPtr result = 
                         gSTAFProcHandlePtr->submit("local",
                                                    notifiee.notifyService,
                                                    request,
                                                    kSTAFReqFireAndForget);
            
                     gHandleManagerPtr->deletePolling(notifiee.handle,
                                                      notifiee.machine,
                                                      notifiee.uuid,
                                                      notifiee.notifyService);
                 }
             }
        }
    }
    catch (STAFException &se)
    {
        se.trace("STAFHandleManager::gcPolling()");
    }
    catch (...)
    {
        STAFTrace::trace(kSTAFTraceError, "Caught unknown exception in "
                         "STAFHandleManager::gcPolling()");
    }
}

unsigned int HandleMonitorThread(void *data)
{
    pHandleManager->gcPolling();

    return kSTAFUnknownError;
}


